#include "core.h"
#include <iostream>
#include <cassert>
#include <QCoreApplication>
#include <QStringList>
#include <QTextStream>
#include <QDir>
#include <QThread>
#include <QMutex>
#include "configfile.h"
#include "iodata.h"
#include "iosnapshot.h"
#include "managepotential.h"
#include "massmodelaxi.h"
#include "odeint.h"
#include "optimization.h"
#include "orbitan.h"
#include "utils.h"
#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#ifdef _OPENMP
#include <omp.h>
#endif

namespace smile{

/** default configuration parameters **/
const CConfigCore defaultConfigCore = {      // internal smile core/gui parameters (not related to any libsmile modules)
    100.0,//double intTimeInPeriods;         // orbit integration time measured in units of long-axis-orbit periods
    0.0,  //double intTimeMaxAdaptive;       // upper limits for adaptive integration time, expressed in energy-dependent orbital time periods
    1.0,  //double adaptiveTimeThreshold;    // parameter in Pfenniger's method for adaptive integration time
    0.0,  //double intTimeMinAbs;            // minimum integration time in absolute units
    false,//bool usePS;                      // whether to plot points on Poincare section
    1, 0, //PSsectionCoord, PSdisplayCoord
    // data for frequency map
    192,  //int fm_numOrbitsStationary;      // number of orbits in stationary start-space
    192,  //int fm_numOrbitsPrincipalPlane;  // number of orbits in principal-plane start-space
    0,    //int fm_numOrbitsYalpha;          // number of orbits in Y-alpha start space
    0,    //int fm_numOrbitsRandom;          // number of orbits in random(ergodic) start-space
    // data for Schwarzschild modelling
    std::vector<unsigned int>(),             // densityComponents
    1000, //int sm_numOrbits;                // total number of orbits in model
    CBasicICGenerator::IC_VEL_EDDINGTON,     // initial velocity generator
    0.0,  //double sm_ICbeta;                // anisotropy for spherical Jeans IC generator
    0.0,  //double sm_ICbetaAxi;             // meridional anisotropy for axisymmetric Jeans IC generator
    0.0,  //double sm_ICkappaAxi;            // degree of rotation for axisymmetric Jeans IC generator
    0.0,  //double weightSkewFactorE;        // bias factor for energy sampling
    0.0,  //double weightSkewFactorL;        // same for ang.mom.
    CBasicSchwData::SD_DENS_GRIDCLASSIC,     // type of density constraints for Schwarzschild modelling
    20,   //int sm_numShells;                // grid size: number of radial shells
    2,    //int sm_linesPerSegment;          // grid size: number of lines dividing each of 3 segments of each shell into cells
    6,    //int sm_numAngularCoefs;          // order of angular spherical-harmonic expansion
    20,   //int sm_numRadialCoefs;           // number of radial coefs in BSE 
    15,   //int sm_numVerticalCoefs;         // number of vertical grid zones for Cylindrical model
    0.0,  //double sm_Alpha;                 // Alpha-parameter defining family of basis functions in BSE model, 0=autodetect
    0.0,  //double sm_innerShellMass;        // fraction of mass in the innermost shell, 0 for default value of 1/(numShells+1)
    0.0,  //double sm_outerShellMass;        // fraction of mass enclosed by the outermost shell, 0 for default value of 1-1/(numShells+1)
    1.0,  //double sm_grid_q;                // axis ratio (y/x) for ellipsoidal flattening of Schwarzschild grid
    1.0,  //double sm_grid_p;                // axis ratio (z/x) for Schwarzschild grid
    1e-3, //double chaoticMinFreqDiff        // FDR threshold
    0.0,  //double chaoticMinLambda          // Lyapunov exp. threshold
    0.0,  //double sm_chaoticPenalty         // penalty in cost function for using chaotic orbits
    0,    //int sm_numSamplingPoints;        // number of points sampled from trajectory, default NUM_SAMPLING_POINTS
    true, //bool sm_useBPMPD;                // whether to use BPMPD solver (if the executable is present), otherwise use GLPK
    0.0,  //sm_maxWeight;                    // upper bound on orbit weight
    0.1,  //sm_regularization;               // parameter controlling the regularization of orbit weights in the quadratic optimization
    1.0,  //sm_constraintPenaltyLin;         // penalty for constraint violation in the objective function in linear/quadratic optimization; 0 means infinity - must satisfy exactly
    1.0,  //sm_constraintPenaltyQuad;        // quadratic penalty term for constraint violation
    false,//sm_constrainBeta;                // whether to constrain anisotropy
    0, 0, //sm_betaIn, sm_betaOut;           // anisotropy coefficient requested (values for the inner and outer shells, linearly interpolated in between)
    false,//sm_constrainAngMom               // whether to constrain binned relative angular momentum distribution
    4,    //sm_numAngMomBins                 // number of bins in angular momentum distribution
};

CSmileCore::CSmileCore() :
  configCore(defaultConfigCore), 
  configOrbit(defaultConfigOrbit)
{
    configPotential.push_back(defaultConfigPotential);
    WorkDir = QDir::currentPath(); 
    AppDir  = getDirName(QCoreApplication::applicationFilePath());
    TempDir = "";
    potential=NULL;
    density=NULL;
    orbit=NULL;
    orbitlib=NULL;
    launcherOrbit=NULL;
    launcherLib=NULL;
    trajSampleCreator=NULL;
    orbitIntRunning=false;
    orbitLibTimer=0;
    schwModel=NULL;
    useICe=false;
    gsl_set_error_handler(&gsl_verbose_error_handler);
    maxNumThread=std::max<int>(1, QThread::idealThreadCount());  // auto select
}

CSmileCore::~CSmileCore()
{   // commented out because the objects may be destroyed while running multi-threaded orbit library integration, 
    // which results in a crash. Probably there is a workaround but since the program is gonna close anyway, who cares.
#if 0
    if(orbit!=NULL)
    {
        delete orbit->getPotential();
        delete orbit;
    }
    if(model!=NULL)
        delete model;
    if(orbitlib!=NULL)
    {
        delete orbitlib->getPotential();
        delete orbitlib;
    }
    if(potential!=NULL)
        delete potential;
#endif
}

void CSmileCore::initPotential()
{
    if(density!=potential && density!=NULL)
        delete density;  // need to delete density if it was initialized differently from potential
    density=NULL;
    if(potential)
        delete potential;
    potential=NULL;
    size_t numComp=configPotential.size();
    if(numComp>1 || configPotential[0].Mbh!=0) {  // composite potential
        std::vector<const CPotential*> components;
        for(size_t ind=0; ind<numComp; ind++) {
            const CPotential* pot=createPotential(&(configPotential[ind]));
            if(pot!=NULL)
                components.push_back(pot);
            else {
                my_error(FUNCNAME, "Warning, incorrect initialization of potential component "+convertToString(ind));
                components.push_back(new CPotentialMiyamotoNagai(0, 0, 1));  // zero-mass placeholder
            }
        }
        if(components.size()>0) {
            potential = new CPotentialComposite(components, configPotential[0].Mbh);
            for(size_t ind=0; ind<components.size(); ind++)
                delete components[ind];   // these were cloned by the composite potential and now are not needed anymore
        }
    }
    if(potential==NULL)  // not created at previous step -- means it is not composite
        potential=createPotential(&(configPotential[0]));
    if(potential==NULL) {  // still not created due to some error
        CConfigPotential cfg(defaultConfigPotential);
        cfg.PotentialType=CPotential::PT_HARMONIC;
        potential=createPotential(&cfg);
        if(configPotential[0].PotentialType==CPotential::PT_UNKNOWN) 
            configPotential[0].PotentialType=cfg.PotentialType;
    }
}

void CSmileCore::initDensity()
{
    assert(potential);
    if(density!=NULL && density!=potential)
        delete density;
    density=NULL;
    if(potential->PotentialType()==CPotential::PT_COMPOSITE) {
        std::vector<const CPotential*> components_d;
        const CPotentialComposite* pot_comp=static_cast<const CPotentialComposite*>(potential);
        size_t ncomp=pot_comp->size();
        for(size_t ic=0; ic<configCore.densityComponents.size(); ic++) {  // if list of components was provided
            if(configCore.densityComponents[ic]<ncomp) 
                components_d.push_back(pot_comp->component(configCore.densityComponents[ic]));
        }
        if(components_d.size()==0) {
            if(configCore.densityComponents.size()>0)  // list was provided but didn't work properly
                my_error(FUNCNAME, "Warning, invalid selection of potential components, using the entire potential instead");
            // fill the list with all components
            for(size_t ic=0; ic<ncomp; ic++)
                components_d.push_back(pot_comp->component(ic));
        }
        if(components_d.size()>1)
            density=new CPotentialComposite(components_d, 0);
        else 
            density=components_d[0]->clone();
    } else {  // not a composite potential
        density=potential;
    }
    assert(density);
}

/* Export/import operations */
bool CSmileCore::exportPotential(const QString& fileName, size_t indComp)
{
    if(potential==NULL) return false;
    vectord points[N_DIM];
    // export potential and its derivatives in a set of points: along three principal axes, three diagonals in each plane, and a x=y=z line
    const int Npoints=100;
    double rmin=0.001, rmax=1000;
    // in case of composite potential containing only one component, poten0 will refer to this component only 
    // (to correctly infer its type, and export potential coefficients if it is a BSE or Spline expansion),
    // otherwise it refers to the entire potential or its selected component, depending on indComp
    const CPotential* poten0 = potential;
    // poten1 refers to the selected component of composite potential. 
    // If indComp=0 then it refers to the entire composite potential, even if it has only one component 
    // (but it may contain a central black hole that should be accounted for in the potential to export)
    const CPotential* poten1 = potential;
    if(potential->PotentialType()==CPotential::PT_COMPOSITE) {
        size_t numComp=static_cast<const CPotentialComposite*>(potential)->size();
        if(numComp==1)
            poten0=static_cast<const CPotentialComposite*>(potential)->component(0); 
        if(indComp>numComp) indComp=numComp;
        if(indComp>0)
            poten0=poten1=static_cast<const CPotentialComposite*>(potential)->component(indComp-1);
    }
    if(poten0->PotentialType()==CPotential::PT_BSECOMPACT)
        rmax = static_cast<const CPotentialBSECompact*>(poten0)->getRmax();
    vectord radii;
    createNonuniformGrid(Npoints, rmin, rmax, false, &radii);
    double axiscoefs[7][3] = { {1,0,0}, {0,1,0}, {0,0,1}, {0.7,0.7,0}, {0,0.7,0.7}, {0.7,0,0.7}, {0.58,0.58,0.58} };
    for(int d=0; d<7; d++)
    {
        if((poten1->symmetry() & CPotential::ST_PLANESYM) == 0)   // no triaxial symmetry -- output both negative and positive radii
        {
            for(int i=Npoints-1; i>=0; i--)
            {
                double r=-radii[i];
                for(unsigned int c=0; c<N_DIM; c++)
                    points[c].push_back(r*axiscoefs[d][c]);
            }
        }
        for(int i=0; i<Npoints; i++)
        {
            double r=radii[i];
            for(unsigned int c=0; c<N_DIM; c++)
                points[c].push_back(r*axiscoefs[d][c]);
        }
    }
    if(!exportPotentialAtPoints(poten1, fileName, points, static_cast<int>(points[0].size())/7))
        return false;
    // export potential, forces and density at points defined as initial conditions for orbit library (if it exists)
    if(orbitlib && orbitlib->size()>0)
    {
        size_t count=orbitlib->size();
        for(unsigned int c=0; c<N_DIM; c++)
            points[c].assign(count, 0);
        for(size_t i=0; i<count; i++)
        {
            const CPosVelPoint<float>& pt=orbitlib->getOrbitDesc(i)->getInitCond();
            for(unsigned int c=0; c<N_DIM; c++)
                points[c][i]=pt.Pos[c];
        }
        if(!exportPotentialAtPoints(poten1, fileName+".points", points))
            return false;
    }
    // export coefficients if it makes sense
    if( poten0->PotentialType()==CDensity::PT_BSE || 
        poten0->PotentialType()==CDensity::PT_BSECOMPACT || 
        poten0->PotentialType()==CDensity::PT_SPLINE || 
        poten0->PotentialType()==CDensity::PT_SCALEFREESH ||
        poten0->PotentialType()==CDensity::PT_CYLSPLINE)
    {
        return writePotential(fileName.toStdString()+coefFileExtension(poten0->PotentialType()), poten0);
    }
    else
        return true;
}

bool CSmileCore::exportPotentialAtPoints(const CPotential* poten, const QString& fileName, const vectord points[N_DIM], int blockSize)
{
    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) 
        return false;
    QTextStream out(&file);
    out.setRealNumberPrecision(8);
    out << "x\ty\tz\tPhi\tFx\tFy\tFz\tRho\n";
    for(size_t p=0; p<points[0].size(); p++)
    {
        double coord[N_DIM]={points[0][p], points[1][p], points[2][p]};
        double force[N_DIM];
        double Phi = poten->Phi(coord[0], coord[1], coord[2]);
        poten->Force(coord, 0, force);
        out << coord[0] << "\t" << coord[1] << "\t" << coord[2] << "\t" << Phi << "\t" 
            <<-force[0] << "\t" <<-force[1] << "\t" <<-force[2] << "\t" 
            << poten->Rho(coord[0], coord[1], coord[2]) << "\n";
        out.flush();
        if(blockSize>0 && (p+1)%blockSize==0) out << "\n";  // separating line
    }
    return true;
}

bool CSmileCore::exportOrbit(const QString& fileName)
{
    if(orbit==NULL || orbit->getIntTime()==0) return false;
    CFileOutputText out(fileName.toStdString());
    out.writeOrbit(orbit);
    return out.ok();
}

bool CSmileCore::importOrbit(const QString &fileName)
{
    if(isOrbitIntRunning()) return false;
    CFileInputText in(fileName.toStdString());
    if(!in.ok()) return false;
    CPotential* newPot=potential->clone();
    COrbit* newOrb=in.readOrbit(&configOrbit, newPot);
    if(!newOrb)
    {
        delete newPot;
        return false;
    }
    if(orbit!=NULL)
    {
        delete orbit->getPotential();  // manually delete, since it was created by cloning externally to the object
        delete orbit;
    }
    if(newOrb->getIntTime()<=0 || newOrb->getTimeUnit()<=0)  // something wrong with the data
    {
        delete newOrb;
        delete newPot;
        return false;
    }
    orbit=newOrb;
    initCond = orbit->getInitCond();
    initCondE = totalEnergy(potential, initCond, configOrbit.Omega);
    timeUnit = orbit->getTimeUnit();
    configCore.intTimeInPeriods=orbit->getIntTime()/timeUnit; 
    return in.ok();
}

bool CSmileCore::exportOrbitLib(const QString &fileName, bool withModelData)
{
    if(orbitlib==NULL || orbitlib->size()==0) return false;
    CFileOutputText out(fileName.toStdString());
    out.writeOrbitLibrary(orbitlib, withModelData);
    return out.ok();
}

bool CSmileCore::importOrbitLib(const QString &fileName, bool withModelData)
{
    CFileInputText in(fileName.toStdString());
    if(!in.ok()) return false;
    COrbitLibrary* newOrbLib=in.readOrbitLibrary(withModelData);
    if(!newOrbLib)
        return false;
    if(orbitlib!=NULL)
        delete orbitlib;
    orbitlib=newOrbLib;
    if(launcherLib!=NULL) {
        delete launcherLib;   // this destroys previous copies of potential, config and creators
        trajSampleCreator=NULL;
    }
    // fill the launcher with the same runtime functions that were used for this orbit library
    // this is necessary if the user will call ExportNbody and there are some orbits that need reintegration --
    // thus we ensure that the reintegrated orbits get the same set of runtime functions as the original ones
    launcherLib = new COrbitLauncher(potential, configOrbit);
    launcherLib->addCreator(new COrbitRuntimeTrajectoryCreator());
    launcherLib->addCreator(new COrbitRuntimePericenterCreator());
    if(configOrbit.relaxRate>0)
        launcherLib->addCreator(new COrbitRuntimeVelPerturberCreator(createMassModel(potential), configOrbit.relaxRate));
    if(withModelData) {
        // don't initialize trajSampleCreator now -- if it will be needed in ExportNbody, it will be created in that routine
        initSchwData();
        // but we do need to set up the creators for Schw.data objects
        for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
        {
            const CBasicOrbitRuntimeFncCreator* crt=createSchwCreator(*iter);
            if(crt!=NULL) 
                launcherLib->addCreator(crt);
        }
        // create a temporary model to compute statistics
        vectorSchwData schwDataTmp;
        for(size_t i=0; i<schwData.size(); i++)
            if((schwData[i]->dataType() & CBasicSchwData::SD_DENS) == CBasicSchwData::SD_DENS)  // only use density constraints
                schwDataTmp.push_back(schwData[i]);
        CSchwModelQuadOpt model(orbitlib, schwDataTmp, NULL, NULL, 0, 0, 0, 0);
        my_message(FUNCNAME, convertToString(orbitlib->size())+" orbits loaded\n" + model.getStatistics());
    }
    else my_message(FUNCNAME, convertToString(orbitlib->size())+" orbits loaded");
    orbitlib->setLauncher(launcherLib);
    return true;
}

bool CSmileCore::exportSchwModelStats(const QString& fileName)
{
    if(orbitlib==NULL || schwData.size()==0)
        return false;
    std::ofstream strm(fileName.toStdString().c_str(), std::ios::out);
    if(!strm) return false;
    CSchwModelQuadOpt model(orbitlib, schwData, NULL, NULL, 0, 0, 0, 0);    // create temp.model for computing constraint deviations
    const CChaosOrbitFilteringFnc chaosWeight(1, configCore.chaoticMinLambda, configCore.chaoticMinFreqDiff);
    size_t ind_constr=0;
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
    {
        strm << "#Data_type: " << (*iter)->dataName() << "\n";
        const CSchwDataDensGrid*   model_cl=NULL;
        const CSchwDataDensSH*     model_sph=NULL;
        const CSchwDataKinemShell* model_kin=NULL;
        const CSchwDataAngMomDist* model_ang=NULL;
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_DENS_GRIDCLASSIC:
            case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL:
                model_cl=static_cast<const CSchwDataDensGrid*>(*iter);
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tChaos\tNumOrbits\tNumOrbitsNonzero\tradius\tx\ty\tz\n";
                break;
            case CBasicSchwData::SD_DENS_SHMESH:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tChaos\tNumOrbits\tNumOrbitsNonzero\tradius\tl\tm\n";
                break;
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                strm << "#index\tRequiredCoef\tActualCoef\tDifference\tNormFactor\tChaos\tNumOrbits\tNumOrbitsNonzero\tn\tl\tm\n";
                break;
            case CBasicSchwData::SD_KINEM_SHELL:
                strm << "#index\tradius\tM(r)\tsigma_r^2\tsigma_t^2\n";
                model_kin=static_cast<const CSchwDataKinemShell*>(*iter); 
                break;
            case CBasicSchwData::SD_KINEM_ANGMOMDIST:
                strm << "#index\tradius\tE(r)\tL2rel\tdistr.fun.\n";
                model_ang=static_cast<const CSchwDataAngMomDist*>(*iter); 
                break;
            default: ;
        }
        if(model_cl!=NULL || model_sph!=NULL) {
            for(size_t i=0; i<(*iter)->numConstraints(); i++)
            {
                double coefVal =(*iter)->getConstraint(i);
                double coefNorm=(*iter)->getConstraintNormFactor(i);
                double coefDev =gsl_nan();
                size_t norb=0, norbnz=0;
                if(model.isCorrect()) {
                    model.calcCoefDeviation(i+ind_constr, &coefDev, &norb, &norbnz);
                }
                double coefComp=model.calcCoefComposition(i+ind_constr, &chaosWeight);
                strm << i << "\t" << coefVal << "\t" << (coefDev+coefVal) << "\t" << coefDev << "\t" << 
                    coefNorm << "\t" << coefComp << "\t" << norb << "\t" << norbnz;
                if(model_cl)
                {
                    double X, Y, Z;
                    model_cl->cellCenter(i, &X, &Y, &Z);
                    strm << "\t" << sqrt(X*X+Y*Y+Z*Z) << "\t" << X << "\t" << Y << "\t" << Z;
                }
                else if(model_sph)
                {
                    double n;
                    int l, m;
                    model_sph->getCoefDecomposition(i, &n, &l, &m);
                    strm << "\t" << n << "\t" << l << "\t" << m;
                }
                strm << "\n";
            }
        }
        else if(model_kin!=NULL)
        {
            size_t numShellsKin=model_kin->numConstraints();
            vectord 
                shellVr(numShellsKin, 0), 
                shellVt(numShellsKin, 0), 
                shellTime(numShellsKin, 0);

            if(model.isCorrect())  for(size_t o=0; o<orbitlib->size(); o++)
            {
                // load kinematic information data from orbit desc
                for(size_t inf=0; inf<orbitlib->getOrbitDesc(o)->getInfoNum(); inf++)
                {
                    const CBasicInformation* info=orbitlib->getOrbitDesc(o)->getInfo(inf);
                    if(info->infoType()==CBasicInformation::IT_SCHW &&
                        static_cast<const CSchwInformation*>(info)->schwDataType()==CBasicSchwData::SD_KINEM_SHELL)
                    {
                        const CSchwInformation* infoKin=static_cast<const CSchwInformation*>(info);
                        double weight=orbitlib->getOrbitDesc(o)->getWeight();
                        for(size_t s=0; s<numShellsKin; s++)
                        {
                            double shTime = infoKin->getSchwData(s) * weight;
                            shellVr[s] += shTime * infoKin->getSchwData(s+1*numShellsKin);
                            shellVt[s] += shTime * infoKin->getSchwData(s+2*numShellsKin);
                            shellTime[s] += shTime;
                        }
                    }
                }
            }
            for(size_t s=0; s<numShellsKin; s++)
            {
                strm << s << "\t" << model_kin->getShellRadius(s) << "\t"
                    << potential->Mass(model_kin->getShellRadius(s)) << "\t"   // NB: potential may be different from the one in model!
                    << (shellVr[s]/shellTime[s]) << "\t" << (shellVt[s]/shellTime[s]) << "\n";
            }
        } else if(model_ang!=NULL)
        {
            size_t numBins=model_ang->getNumBins();
            size_t numShells=model_ang->numConstraints()/numBins;
            vectord dataAngMom(numShells*numBins);
            if(model.isCorrect())  for(size_t o=0; o<orbitlib->size(); o++)
            {
                // load angular momentum information data from orbit desc
                for(size_t inf=0; inf<orbitlib->getOrbitDesc(o)->getInfoNum(); inf++)
                {
                    const CBasicInformation* info=orbitlib->getOrbitDesc(o)->getInfo(inf);
                    if(info->infoType()==CBasicInformation::IT_SCHW &&
                        static_cast<const CSchwInformation*>(info)->schwDataType()==CBasicSchwData::SD_KINEM_ANGMOMDIST)
                    {
                        const CSchwInformation* infoAng=static_cast<const CSchwInformation*>(info);
                        double weight=orbitlib->getOrbitDesc(o)->getWeight();
                        double energy = infoAng->getSchwData(numBins);
                        size_t shell = model_ang->whichShellE(energy);
                        if(shell<numShells)
                            for(size_t b=0; b<numBins; b++)
                                dataAngMom[shell*numBins+b] += weight*infoAng->getSchwData(b);
                    }
                }
            }
            for(size_t s=0; s<numShells*numBins; s++)
            {
                strm << s << "\t" << model_ang->getShellEnergy(s/numBins) << "\t"
                    << ((pow_2(s%numBins)+pow_2(s%numBins+1))/2/pow_2(numBins)) << "\t" << dataAngMom[s] << "\n";
            }
        }
        strm << "\n";
    }
    return static_cast<bool>(strm);
}

/* ----------- orbit integration ------------ */
void CSmileCore::initIC()
{
    if(useICe)   // assign from initCondE
    {
        initCond=CPosVelPoint<double>();
        initCond.Pos[0]=longAxisRadius(potential, initCondE, configOrbit.Omega);
    }
    else
    {
        initCondE = totalEnergy(potential, initCond, configOrbit.Omega);
    }
    // calculate T_orb (for x-axis orbit)
    timeUnit = computeTimeUnit(potential, initCond/*, configOrbit.Omega*/);
}

void CSmileCore::startOrbit()
{
    if(timeUnit<=0 || configCore.intTimeInPeriods<=0 || configOrbit.samplingInterval<=0) return;  // error
    if(isOrbitIntRunning()) return;   // can't invoke this while another orbit is running
    if(orbit!=NULL)  // this one left from the previous orbitintegration (is not destroyed after it finishes)
        delete orbit;
    if(launcherOrbit!=NULL)     // in fact, these two conditions must be both true or false
        delete launcherOrbit;   // this destroys previous copies of potential, config and creators
    configOrbit.freqStrictOrder = configOrbit.Omega==0;  // can't reliably normalize frequencies in case of figure rotation 
    launcherOrbit = new COrbitLauncher(potential, configOrbit);
    launcherOrbit->addCreator(new COrbitRuntimeTrajectoryCreator());
    if(configCore.usePS) 
        launcherOrbit->addCreator(new COrbitRuntimePoincareCreator(configCore.PSdisplayCoord, configCore.PSsectionCoord));
    launcherOrbit->addCreator(new COrbitRuntimePericenterCreator());
    if(configOrbit.relaxRate>0)
        launcherOrbit->addCreator(new COrbitRuntimeVelPerturberCreator(createMassModel(potential), configOrbit.relaxRate));
    orbit = launcherOrbit->createOrbit(initCond, configCore.intTimeInPeriods*timeUnit, 0.0);
    // create a worker object for doing orbit integration
    COrbitIntWorker *wkr = new COrbitIntWorker(orbit);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's run method
    connect(thr, SIGNAL(started()), wkr, SLOT(run()));
    // when orbit integration finishes, notify about it
    connect(wkr, SIGNAL(calcFinished()), this, SLOT(coreOrbitFinished()));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(calcFinished()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread, no need to care about its proper deletion hereafter
    thr->start();
    orbitIntRunning=true;
}

void CSmileCore::coreOrbitFinished()
{
    orbitIntRunning=false;
    emit signalOrbitFinished();
}

/* ----------- Frequency map / Schwarzschild modelling ------------- */

void CSmileCore::startOrbitLibrary(const char* finishSignal)
{
    if(orbitlib==NULL || orbitlib->size()==0) {
        emit signalOrbitLibraryFinished();
        return;
    }
    for(int t=0; t<maxNumThread; t++)
    {
        // create a worker object for doing orbit integration in parallel
        COrbitLibWorker *wkr = new COrbitLibWorker(orbitlib);
        // create a separate thread to perform the job
        QThread* thr = new QThread();
        // move the worker object to the new thread
        wkr->moveToThread(thr);
        // when thread starts, launch the worker object's run method
        connect(thr, SIGNAL(started()), wkr, SLOT(run()));
        // if termination of orbit library integration is requested, inform the worker object about it
        connect(this, SIGNAL(signalStopOrbitLibrary()), wkr, SLOT(halt()), Qt::DirectConnection);
        // when the integration of entire orbit library finishes, notify about it with the provided signal
        connect(wkr, SIGNAL(calcFinished()), this, finishSignal);
        // when worker object finishes, the thread that hosted it should be deleted
        connect(wkr, SIGNAL(calcFinished()), thr, SLOT(quit()));
        // when the thread has finished, it will be deleted shortly thereafter
        connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
        // now start the thread, no need to care about its proper deletion hereafter
        thr->start(QThread::LowPriority);
    }
    my_message(FUNCNAME, "Starting integration..");
    orbitLibTimer = startTimer(2000);  // to refresh progress and to mark that orbit library integration is in progress
}

void CSmileCore::coreOrbitLibraryFinished()
{
    killTimer(orbitLibTimer);
    orbitLibTimer=0;  // stop showing progress indicator
    // delete non-used orbits (if integration threads were terminated prematurely, there exist some)
    orbitlib->removeUnused();
    // finally, send signal to the owner object
    emit signalOrbitLibraryFinished();
}

void CSmileCore::timerEvent(QTimerEvent *)
{
    if(orbitLibTimer==0)
        return;           // shouldn't happen - this routine is invoked only during orbit library integration
    emit signalTimer();   // triggered regularly while orbit library integration is running
}

void CSmileCore::buildFreqMapIC(double energy, bool useExistingIC)
{
    assert(potential);
    schwData.clear();   // no schw model is used in frequency map
    if(launcherLib!=NULL) {
        delete launcherLib;   // this destroys previous copies of potential, config and creators
        trajSampleCreator=NULL;
    }
    // fill the launcher with the set of runtime functions to be applied to each orbit
    configOrbit.freqStrictOrder = configOrbit.Omega==0;  // can't reliably normalize frequencies in case of figure rotation 
    launcherLib = new COrbitLauncher(potential, configOrbit);
    launcherLib->addCreator(new COrbitRuntimeTrajectoryCreator());
    launcherLib->addCreator(new COrbitRuntimePericenterCreator());
    if(configOrbit.relaxRate>0)
        launcherLib->addCreator(new COrbitRuntimeVelPerturberCreator(createMassModel(potential), configOrbit.relaxRate));
    if(configCore.usePS)
        launcherLib->addCreator(new COrbitRuntimePoincareCreator(configCore.PSdisplayCoord, configCore.PSsectionCoord));
    // prepare new set of initial conditions or use the existing one
    CPointMassSet<float>  startingPointsArr;
    CPointMassSet<float>* startingPoints=NULL;
    if(!useExistingIC || orbitlib==NULL || orbitlib->size()==0)
    {   // create new set of initial conditions
        my_message(FUNCNAME, "Preparing initial conditions..");
        startingPoints=&startingPointsArr;
        if(configCore.fm_numOrbitsStationary>0) {
            CICGeneratorStationary icg(potential, energy, configOrbit.N_dim, configOrbit.Omega);
            icg.createInitialConditions(configCore.fm_numOrbitsStationary, startingPoints);
        }
        if(configCore.fm_numOrbitsPrincipalPlane>0) {
            CICGeneratorPrincipalPlane icg(potential, energy, configOrbit.N_dim, configOrbit.Omega);
            icg.createInitialConditions(configCore.fm_numOrbitsPrincipalPlane, startingPoints);
        }
        if(configCore.fm_numOrbitsYalpha>0) {
            CICGeneratorYalpha icg(potential, energy, configOrbit.N_dim, configOrbit.Omega);
            icg.createInitialConditions(configCore.fm_numOrbitsYalpha, startingPoints);
        }
        if(configCore.fm_numOrbitsRandom>0) {
            CICGeneratorRandomE icg(potential, energy, configOrbit.N_dim, configOrbit.Omega);
            icg.createInitialConditions(configCore.fm_numOrbitsRandom, startingPoints);
        }
        if(orbitlib)
            delete orbitlib;
        orbitlib=new COrbitLibrary();
    }   // otherwise will modify existing initial conditions
    orbitlib->assignInitialConditions(startingPoints, launcherLib, 
        configCore.intTimeInPeriods, configCore.intTimeMaxAdaptive, configCore.intTimeMinAbs);
}

void CSmileCore::initSchwData()
{
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
        delete (*iter);
    schwData.clear();
    if(!density)
        initDensity();
    if(density->totalMass()<0) {
        my_error(FUNCNAME, "Error, density model has infinite mass");
        return;
    }
    my_message(FUNCNAME, "Creating Schwarzschild model..");
    CBasicSchwData* dataDens=NULL;
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Creating density model "+std::string(getSchwDataNameByType(configCore.sm_densDataType)));
#endif
    switch(configCore.sm_densDataType) {
        case CBasicSchwData::SD_DENS_GRIDCLASSIC:
            dataDens=new CSchwDataDensGridClassic(density, configCore.sm_numRadialCoefs, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, configCore.sm_linesPerSegment,
                configCore.sm_grid_q, configCore.sm_grid_p);
            break;
        case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL: {
            CSchwDataDensGridCylindrical* mod=new CSchwDataDensGridCylindrical(density, 
                configCore.sm_numRadialCoefs, configCore.sm_numVerticalCoefs, std::max<size_t>(configCore.sm_numAngularCoefs,1));
            dataDens=mod;
            break;
        }
        case CBasicSchwData::SD_DENS_SHMESH:
            dataDens=new CSchwDataDensSHMesh(density, configCore.sm_numRadialCoefs, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, configCore.sm_numAngularCoefs);
            break;
        case CBasicSchwData::SD_DENS_SHBSE:
            dataDens=new CSchwDataDensSHBSE(density, 
                configCore.sm_numAngularCoefs, configCore.sm_numRadialCoefs, configCore.sm_Alpha);
            break;
        default: my_error(FUNCNAME, "Unrecognized type of Schwarzschild density model"); dataDens=NULL;
    }
    if(dataDens!=NULL) schwData.push_back(dataDens);
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Creating kinematic model "+std::string(CSchwDataKinemShell::myName()));
#endif
    schwData.push_back(new CSchwDataKinemShell(density, configCore.sm_numShells, 
        configCore.sm_innerShellMass, configCore.sm_outerShellMass, 
        configCore.sm_betaIn, configCore.sm_betaOut));
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Creating kinematic model "+std::string(CSchwDataAngMomDist::myName()));
#endif 
    schwData.push_back(new CSchwDataAngMomDist(density, configCore.sm_numShells, 
        configCore.sm_innerShellMass, configCore.sm_outerShellMass, 
        configCore.sm_betaIn, configCore.sm_betaOut, potential, configCore.sm_numAngMomBins));
}

void CSmileCore::buildSchwIC(bool useExistingIC)
{
    assert(potential);
    if(configOrbit.N_dim!=N_DIM) {
        my_error(FUNCNAME, "Schwarzschild modelling is only possible in three dimensions");
        return;
    }
    initSchwData();
    if(schwData.size()==0) {  // indicates an error -- e.g. density model had infinite mass
        return;  // don't bug the user twice
    }
    // fill the list of runtime function creators
    if(launcherLib!=NULL) {
        delete launcherLib;   // this destroys previous copies of potential, config and creators
        trajSampleCreator=NULL;
    }
    configOrbit.freqStrictOrder = configOrbit.Omega==0;  // can't reliably normalize frequencies in case of figure rotation 
    launcherLib = new COrbitLauncher(potential, configOrbit);
    launcherLib->addCreator(new COrbitRuntimeTrajectoryCreator());
    launcherLib->addCreator(new COrbitRuntimePericenterCreator());
    if(configOrbit.relaxRate>0)
        launcherLib->addCreator(new COrbitRuntimeVelPerturberCreator(createMassModel(potential), configOrbit.relaxRate));
    if(configCore.sm_numSamplingPoints>0) 
    {   // special handling of this creator function: store the pointer to this object 
        // separately from the launcher, as it may need to be used in exportNbody
        trajSampleCreator=new COrbitRuntimeTrajSampleCreator(
            configCore.sm_numSamplingPoints, configCore.intTimeInPeriods);
        launcherLib->addCreator(trajSampleCreator);
    }
    for(vectorSchwData::iterator iter=schwData.begin(); iter!=schwData.end(); iter++)
    {
        const CBasicOrbitRuntimeFncCreator* crt=createSchwCreator(*iter);
        if(crt!=NULL) 
            launcherLib->addCreator(crt);
    }
    // Prepare initial conditions
    CPointMassSet<float>  startingPointsArr;
    CPointMassSet<float>* startingPoints=NULL;
    if(!useExistingIC || orbitlib==NULL || orbitlib->size()==0)
    {   // create new set of initial conditions
        my_message(FUNCNAME, "Preparing initial conditions..");
        startingPoints=&startingPointsArr;
        //CICGeneratorDensityEllipsoidal icd(density);  // obsolete, superceded by DensitySampler
        CICGeneratorDensitySampler icd(density);  // generator of initial positions
        CBasicICGenerator* icv=NULL;   // generator of initial velocities
        switch(configCore.sm_ICgenType) {
            case CBasicICGenerator::IC_VEL_JEANSSPH: 
                icv = new CICGeneratorVelocityJeansSph(potential, density==potential?NULL:density, configCore.sm_ICbeta);
                break;
            case CBasicICGenerator::IC_VEL_JEANSAXI: 
#ifdef HAVE_INTERP2D
                icv = new CICGeneratorVelocityJeansAxi(potential, density==potential?NULL:density, 
                    configCore.sm_ICbetaAxi, configCore.sm_ICkAxi);
                break;
#else
                my_error(FUNCNAME, "Axisymmetric Jeans models require interp2d");
#endif
            default:
                icv = new CICGeneratorVelocityEddington(potential, density==potential?NULL:density, configCore.weightSkewFactorL);
        }
        CICGeneratorSchw icg(&icd, icv, potential, configCore.weightSkewFactorE);  // combine positions and velocities and sort particles in energy
        icg.createInitialConditions(configCore.sm_numOrbits, startingPoints);
        delete icv;
        // debugging - unused
        //CIOSnapshotNemo snap("ic.nemo");
        //snap.writeSnapshot(startingPoints);
        if(orbitlib)
            delete orbitlib;
        orbitlib=new COrbitLibrary();
    }   // otherwise will modify existing initial conditions
    orbitlib->assignInitialConditions(startingPoints, launcherLib, 
        configCore.intTimeInPeriods, configCore.intTimeMaxAdaptive, configCore.intTimeMinAbs);
}

void CSmileCore::schwStartOptimization(CBasicSchwModel::MODELTYPE modeltype)
{
    if(isSchwOptRunning())
    {
        my_error(FUNCNAME, "Schwarzschild optimization is already running");
        return;      // do not allow to run two threads at once
    }
    if(!orbitlib || orbitlib->size()==0 || schwData.size()==0 || !density)
    {
        my_error(FUNCNAME, "No Schwarzschild model defined");
        return;
    }
    // check that orbit weight priors are set (they are used to inform the solver about total mass)
    double sumPriors=0;
    for(size_t o=0; o<orbitlib->size(); o++)
        sumPriors+=orbitlib->getOrbitDesc(o)->getWeightPrior();
    if(sumPriors==0) {  // shouldn't leave it like this; assign uniform prior
        double prior=density->totalMass()/orbitlib->size();
        for(size_t o=0; o<orbitlib->size(); o++)
            orbitlib->getOrbitDesc(o)->setWeightPrior(prior);
    }
    // create quad.solver object and orbit penalty fnc if necessary
    const CBasicQuadraticOptimizationSolver<smNumType>* solver = NULL;
    if(modeltype==CBasicSchwModel::SM_LINEAR || modeltype==CBasicSchwModel::SM_QUADRATIC) {
        if(configCore.sm_useBPMPD) 
            solver=new COptimizationSolverBPMPD<smNumType>(AppDir.toStdString()+"bpmpd", TempDir.toStdString()); 
        else {
#ifdef HAVE_GLPK
            if(modeltype==CBasicSchwModel::SM_LINEAR && solver==NULL)
                solver=new COptimizationSolverGLPK<smNumType>();
#endif
#ifdef HAVE_CVXOPT
            if(solver==NULL)
                solver=new COptimizationSolverCVXOPT<smNumType>(AppDir.toStdString());
#endif
        }
        if(solver==NULL) {
            my_error(FUNCNAME, 
                std::string(modeltype==CBasicSchwModel::SM_QUADRATIC ? "Quadratic" : "Linear") + 
                " optimization solver not available!");
            return;
        }
    }
    const CBasicOrbitFilteringFnc* orbitPenaltyFnc = 
        configCore.sm_chaoticPenalty!=0 ? 
        new CChaosOrbitFilteringFnc(configCore.sm_chaoticPenalty, configCore.chaoticMinLambda, configCore.chaoticMinFreqDiff) : NULL;
        //new CCentrophilicOrbitFilteringFnc(configCore.sm_chaoticPenalty)
        //new CAngMomOrbitFilteringFnc(configCore.sm_chaoticPenalty, configOrbit.Omega)
    // prepare velocity anisotropy constraints if necessary
    vectorSchwData schwDataNew;
    for(size_t i=0; i<schwData.size(); i++) {
        if((schwData[i]->dataType() & CBasicSchwData::SD_DENS) == CBasicSchwData::SD_DENS)
            schwDataNew.push_back(schwData[i]);
        if(schwData[i]->dataType() == CBasicSchwData::SD_KINEM_SHELL && configCore.sm_constrainBeta) {  // need to update 
            delete schwData[i];
            schwData[i]=new CSchwDataKinemShell(density, configCore.sm_numShells, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, 
                configCore.sm_betaIn, configCore.sm_betaOut);
            schwDataNew.push_back(schwData[i]);
        }
        if(schwData[i]->dataType() == CBasicSchwData::SD_KINEM_ANGMOMDIST && configCore.sm_constrainAngMomDist) {  // need to update 
            delete schwData[i];
            schwData[i]=new CSchwDataAngMomDist(density, configCore.sm_numShells, 
                configCore.sm_innerShellMass, configCore.sm_outerShellMass, 
                configCore.sm_betaIn, configCore.sm_betaOut, potential, configCore.sm_numAngMomBins);
            schwDataNew.push_back(schwData[i]);
        }
    }
    schwModel=new CSchwModelQuadOpt(
        orbitlib, schwDataNew, solver, orbitPenaltyFnc, 
        modeltype==CBasicSchwModel::SM_QUADRATIC ? configCore.sm_regularization : 0, 
        configCore.sm_constraintPenaltyLin, 
        modeltype==CBasicSchwModel::SM_QUADRATIC ? configCore.sm_constraintPenaltyQuad : 0,
        configCore.sm_maxWeight);
    if(schwModel==NULL) {
        my_error(FUNCNAME, "Error creating an instance of Schwarzschild model");
        if(solver!=NULL)
            delete solver;
        if(orbitPenaltyFnc!=NULL)
            delete orbitPenaltyFnc;
        return;
    } 
    // create a worker object for doing optimization
    CSchwarzschildWorker* wkr = new CSchwarzschildWorker(schwModel);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's doWork method
    connect(thr, SIGNAL(started()), wkr, SLOT(doWork()));
    // when optimization finishes, display results
    connect(wkr, SIGNAL(message(const QString&)), this, SLOT(coreSchwOptimizationFinished(const QString&)));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(workDone()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread, yeah!
    thr->start();
}

void CSmileCore::coreSchwOptimizationFinished(const QString& result)
{
    if(schwModel!=NULL) delete schwModel; // associated solver and penalty fnc will be deleted automatically
    schwModel=NULL;
    emit signalSchwOptimizationFinished(result);
}

void CSmileCore::schwExportNbody(int numPoints, const QString &fileName, const QString &fileFormat, int numBinsRefine)
{
    if(orbitlib==NULL || orbitlib->size()==0)
        return;  // nothing to share...
    if(fileName.isEmpty()) {
        my_error(FUNCNAME, "Filename cannot be empty!");
        return;
    }
    if(isExportNbodyRunning()) {
        my_error(FUNCNAME, "Error, export to Nbody is already in progress");
        return;
    }
    numPointsNbodyExport=numPoints;
    fileNameNbodyExport=fileName;     // non-empty filename means that export is in progress
    fileFormatNbodyExport=fileFormat;
    numBinsRefineNbodyExport=numBinsRefine;
    firstCallNbodyExport=true;
    coreNbodyExport();
}

void CSmileCore::coreNbodyExport()
{
    if(fileNameNbodyExport.isEmpty()) 
        return;  // incorrect call, shouldn't happen
    if(orbitLibTimer!=0) {  // called after reintegration was finished (or terminated by user)
        killTimer(orbitLibTimer); // to stop showing up progress bar after reintegration is complete
        orbitLibTimer=0;
        emit signalNbodyReintegrationFinished();
    }
    //const CRadiusMassRefinementFnc massRefineFnc(pow(2.0, numBinsRefineNbodyExport*1.0), 0.42);  // example of mass refinement within the given radius
    const CEnergyOrbitFilteringFnc rankFnc(configOrbit.Omega);
    const CBinnedMassRefinementFnc massRefineFnc(orbitlib, numBinsRefineNbodyExport, &rankFnc);
    CPointMassSet<float>* points = NULL;
    if(trajSampleCreator==NULL) 
    {   // if the previous orbit library integration did not use trajectory sampling, 
        // add the trajSampleCreator instance to the launcher, as it will be needed in exportNbody
        trajSampleCreator=new COrbitRuntimeTrajSampleCreator(
            std::max<size_t>(configCore.sm_numSamplingPoints, numPointsNbodyExport/orbitlib->size()), 
            configCore.intTimeInPeriods);
        launcherLib->addCreator(trajSampleCreator);
    }
    std::string result=orbitlib->exportNbody(
        numPointsNbodyExport, 
        numBinsRefineNbodyExport ? &massRefineFnc : NULL, 
        &points, trajSampleCreator);
    QString info = QString::fromStdString(result);
    if(points==NULL)
    {
        coreNbodyExportFinished("Error in creating Nbody model: "+info);
        return;
    }
    if(info.startsWith("Reintegrating"))
    {   // requires reintegration
        if(!firstCallNbodyExport)
        {   // for some reason the reintegration still did not create enough sampling points
            coreNbodyExportFinished("Export of Nbody model cancelled or was unsuccessful");
            delete points;
            return;
        }
        firstCallNbodyExport=false;
        startOrbitLibrary(SLOT(coreNbodyExport()));   // will be called once again
        my_message("", result);
        delete points;
        return;
    }
    my_message("", "Exporting Nbody model..         ");
    bool fileExportResult=true;
    CBasicIOSnapshot* snap = createIOSnapshotWrite(fileFormatNbodyExport.toStdString(), fileNameNbodyExport.toStdString(), 
        "SMILE version="SMILEVERSION" potential="+std::string(potential->PotentialName())+
        " schwmodel="+getSchwDataNameByType(configCore.sm_densDataType)+
        " numorbits="+QString::number(static_cast<unsigned int>(orbitlib->size())).toStdString());
    if(snap) { 
        fileExportResult &= snap->writeSnapshot(points);
        delete snap;
    } else {
        my_error(FUNCNAME, "Invalid export file format: "+fileFormatNbodyExport.toStdString());
        fileExportResult=false;
    }
    if(!fileExportResult) {
        coreNbodyExportFinished("Error writing file "+fileNameNbodyExport);
        delete points;
        return;
    }
    // start a helper thread with a worker object to compute virial ratio
    my_message("", result+"\nComputing virial ratio...");
    // if a potential was composite (i.e. contained some other component apart from 
    // the one that was just converted to an N-body snapshot (in particular, a central BH),
    // then we need to account for extra components in computing potential energy
    const CPotential* extraPotential=NULL;
    if(potential->PotentialType()==CPotential::PT_COMPOSITE) {
        const CPotentialComposite* pot_comp=static_cast<const CPotentialComposite*>(potential);
        std::vector<const CPotential*> components_ext;
        if(configCore.densityComponents.size()>0) { // list of density components was provided
            size_t ncomp=pot_comp->size();
            for(size_t ic=0; ic<ncomp; ic++) {
                bool used=false;
                for(size_t icden=0; icden<configCore.densityComponents.size(); icden++)
                    if(configCore.densityComponents[icden]==ic) used=true;
                if(!used)  // this component was not used in the density that is being exported, thus it counts as external potential
                    components_ext.push_back(pot_comp->component(ic));
            }
        }
        if(components_ext.size()>0)
            extraPotential=new CPotentialComposite(components_ext, 0 /*configPotential[0].Mbh*/);  // BH is already present as one of the exported particles
    }
    // points will be deleted when the worker object is destroyed
    CNbodyExportStatWorker* wkr = new CNbodyExportStatWorker(points, info, extraPotential);
    // create a separate thread to perform the job
    QThread* thr = new QThread();
    // move the worker object to the new thread
    wkr->moveToThread(thr);
    // when thread starts, launch the worker object's doWork method
    connect(thr, SIGNAL(started()), wkr, SLOT(doWork()));
    // to terminate the thread upon request
    connect(this, SIGNAL(signalStopExportNbody()), wkr, SLOT(halt()), Qt::DirectConnection);
    // when computation is done, display results and reset the filename
    connect(wkr, SIGNAL(exportFinished(const QString&)), this, SLOT(coreNbodyExportFinished(const QString&)));
    // when worker object finishes, the thread that hosted it should be deleted
    connect(wkr, SIGNAL(workDone()), thr, SLOT(quit()));
    // when the thread has finished, it will be deleted shortly thereafter
    connect(thr, SIGNAL(finished()), thr, SLOT(deleteLater()));
    // now start the thread
    thr->start();
}

void CSmileCore::coreNbodyExportFinished(const QString &result)
{
    fileNameNbodyExport.clear();   // clear internal flag that means the export is in progress
    emit signalNbodyExportFinished(result);
}

/* ----------- load/save settings ------------- */

bool CSmileCore::loadSettings(const std::string &fileNameSrc)
{
    std::string fileName;
    if(fileNameSrc=="") 
        fileName=AppDir.toStdString()+DEFAULT_SETTINGS_FILE_NAME;
    else if(!QFile::exists(QString::fromStdString(fileNameSrc)))
        return false;
    else fileName=fileNameSrc;
    CConfigFile settings(fileName);
    QString WorkDir1 = QString::fromStdString(settings.getValueString("Common/WorkDir"));
    if(WorkDir1!="")
    {
        QFileInfo fi(WorkDir1);
        if(fi.isRelative()) WorkDir = AppDir+WorkDir1; else WorkDir = WorkDir1;
    }
    QString TempDir1 = QString::fromStdString(settings.getValueString("Common/TempDir"));
    if(TempDir1!="")
    {
        QFileInfo fi(TempDir1);
        if(fi.isRelative()) TempDir = AppDir+TempDir1; else TempDir = TempDir1;
    } 
    else if(TempDir=="") TempDir=WorkDir;
    if(!WorkDir.endsWith('/')) WorkDir += '/';
    if(!TempDir.endsWith('/')) TempDir += '/';
    if(!QFile::exists(WorkDir)) { 
        my_error(FUNCNAME, "Warning, working directory "+WorkDir.toStdString()+" doesn't exist!");
        WorkDir=AppDir;
    }
    if(!QFile::exists(TempDir)) { 
        my_error(FUNCNAME, "Warning, temp directory "+TempDir.toStdString()+" doesn't exist!");
        TempDir=AppDir;
    }
    int newMaxNumThread=settings.getValueInt("Common/MaxThreads");
    if(newMaxNumThread>=1) maxNumThread=newMaxNumThread;   // if no value exists in ini file, leave current
    configCore.sm_useBPMPD = settings.getValueBool("Common/useBPMPD", configCore.sm_useBPMPD);
    if(!QFile::exists(AppDir+"bpmpd.exe")) configCore.sm_useBPMPD=false;  // if BPMPD.exe does not exist, resort to GLPK or CVXOPT
    // determine the number of potential sections
    std::vector<std::string> groups;
    settings.listGroups(&groups);
    int numPotGroups=1;
    for(size_t ig=0; ig<groups.size(); ig++) 
        if(groups[ig].find("Potential")==0) {
            int num=convertTo<int>(groups[ig].substr(9));
            if(num>=1)
                numPotGroups=std::max<int>(numPotGroups, num+1);
        }
    configPotential.resize(1);  // eliminate all but one
    for(int ig=1; ig<numPotGroups; ig++)
        configPotential.push_back(defaultConfigPotential);  // all but the first potentials are initialized by default
    // potential section (possibly several of them)
    for(int ig=0; ig<numPotGroups; ig++) 
        readConfigPotential(settings, &(configPotential[ig]), ig);
    // orbit integration section
    double icX = settings.getValueDouble("Orbit/x", 1);
    double icY = settings.getValueDouble("Orbit/y");
    double icZ = settings.getValueDouble("Orbit/z");
    double icVx = settings.getValueDouble("Orbit/vx");
    double icVy = settings.getValueDouble("Orbit/vy");
    double icVz = settings.getValueDouble("Orbit/vz");
    initCondE = settings.getValueDouble("Orbit/E");
    useICe = settings.getValueBool("Orbit/useE", false);
    initCond=CPosVelPoint<double>(icX, icY, icZ, icVx, icVy, icVz);
    configCore.intTimeInPeriods = settings.getValueDouble("Orbit/intTime", configCore.intTimeInPeriods);
    configCore.intTimeMaxAdaptive = settings.getValueDouble("Orbit/intTimeMax", configCore.intTimeMaxAdaptive);
    configCore.adaptiveTimeThreshold = settings.getValueDouble("Orbit/adaptiveTimeThreshold", configCore.adaptiveTimeThreshold);
    configCore.intTimeMinAbs = settings.getValueDouble("Orbit/intTimeMinAbs", configCore.intTimeMinAbs);
    configCore.usePS = settings.getValueBool("Orbit/usePS", configCore.usePS);
    readConfigOrbit(settings, &configOrbit);
    // frequency map
    configCore.fm_numOrbitsStationary = settings.getValueInt("Frequency_map/numOrbitsStationary", configCore.fm_numOrbitsStationary);
    configCore.fm_numOrbitsPrincipalPlane = settings.getValueInt("Frequency_map/numOrbitsPrincipalPlane", configCore.fm_numOrbitsPrincipalPlane);
    configCore.fm_numOrbitsYalpha = settings.getValueInt("Frequency_map/numOrbitsYalpha", configCore.fm_numOrbitsYalpha);
    configCore.fm_numOrbitsRandom = settings.getValueInt("Frequency_map/numOrbitsRandom", configCore.fm_numOrbitsRandom);
    // schwarzschild modelling
    QString comps=QString::fromStdString(settings.getValueString("Schwarzschild_model/densityComponents", "all"));
    configCore.densityComponents.clear();
    if(comps!="all") {
        QStringList SL=comps.split(QRegExp("[,\\s]+"),QString::SkipEmptyParts);
        for(int i=0; i<SL.size(); i++) {
            bool ok;
            int c=SL[i].toInt(&ok);
            if(ok) configCore.densityComponents.push_back(c);
        }
    }
    configCore.sm_numOrbits = settings.getValueInt("Schwarzschild_model/numOrbits", configCore.sm_numOrbits);
    configCore.sm_ICgenType = getICgenTypeByName(settings.getValueString("Schwarzschild_model/ICgenerator"));
    configCore.sm_ICbeta = settings.getValueDouble("Schwarzschild_model/ICbeta", configCore.sm_ICbeta);
    configCore.sm_ICbetaAxi = settings.getValueDouble("Schwarzschild_model/ICbetaAxi", configCore.sm_ICbetaAxi);
    configCore.sm_ICkAxi = settings.getValueDouble("Schwarzschild_model/ICkAxi", configCore.sm_ICkAxi);
    configCore.weightSkewFactorE = settings.getValueDouble("Schwarzschild_model/weightSkewFactorE", configCore.weightSkewFactorE);
    configCore.weightSkewFactorL = settings.getValueDouble("Schwarzschild_model/weightSkewFactorL", configCore.weightSkewFactorL);
    configCore.sm_densDataType = getSchwDataTypeByName(settings.getValueString("Schwarzschild_model/densityModelType", CSchwDataDensGridClassic::myName()));
    configCore.sm_linesPerSegment = settings.getValueInt("Schwarzschild_model/linesPerSegment", configCore.sm_linesPerSegment);
    configCore.sm_numAngularCoefs = settings.getValueInt("Schwarzschild_model/numAngularCoefs", configCore.sm_numAngularCoefs);
    configCore.sm_numRadialCoefs = settings.getValueInt("Schwarzschild_model/numRadialCoefs", configCore.sm_numRadialCoefs);
    configCore.sm_numVerticalCoefs = settings.getValueInt("Schwarzschild_model/numVerticalCoefs", configCore.sm_numVerticalCoefs);
    configCore.sm_grid_q = settings.getValueDouble("Schwarzschild_model/smGridYtoX", configCore.sm_grid_q);
    configCore.sm_grid_p = settings.getValueDouble("Schwarzschild_model/smGridZtoX", configCore.sm_grid_p);
    configCore.sm_Alpha = settings.getValueDouble("Schwarzschild_model/smAlpha", configCore.sm_Alpha);
    configCore.sm_numShells = settings.getValueInt("Schwarzschild_model/numShells", configCore.sm_numShells);
    configCore.sm_innerShellMass = std::max<double>(0, settings.getValueDouble("Schwarzschild_model/innerShellMass", configCore.sm_innerShellMass));
    configCore.sm_outerShellMass = std::max<double>(0, settings.getValueDouble("Schwarzschild_model/outerShellMass", configCore.sm_outerShellMass));
    configCore.sm_numSamplingPoints = std::max<int>(0, settings.getValueInt("Schwarzschild_model/numSamplingPoints", configCore.sm_numSamplingPoints));
    configCore.chaoticMinFreqDiff = settings.getValueDouble("Schwarzschild_model/chaoticMinFreqDiff", configCore.chaoticMinFreqDiff);
    configCore.chaoticMinLambda = settings.getValueDouble("Schwarzschild_model/chaoticMinLambda", configCore.chaoticMinLambda);
    configCore.sm_chaoticPenalty = settings.getValueDouble("Schwarzschild_model/chaoticPenalty", configCore.sm_chaoticPenalty);
    configCore.sm_maxWeight = std::max<double>(0, settings.getValueDouble("Schwarzschild_model/maxWeight", configCore.sm_maxWeight));
    configCore.sm_regularization = std::max<double>(0, settings.getValueDouble("Schwarzschild_model/regularization", configCore.sm_regularization));
    configCore.sm_constraintPenaltyLin  = settings.getValueDouble("Schwarzschild_model/constraintPenaltyLin", configCore.sm_constraintPenaltyLin);
    configCore.sm_constraintPenaltyQuad = std::max<double>(0, settings.getValueDouble("Schwarzschild_model/constraintPenaltyQuad", configCore.sm_constraintPenaltyQuad));
    configCore.sm_constrainBeta = settings.getValueBool("Schwarzschild_model/constrainBeta", configCore.sm_constrainBeta);
    configCore.sm_betaIn = settings.getValueDouble("Schwarzschild_model/betaIn", configCore.sm_betaIn);
    configCore.sm_betaOut= settings.getValueDouble("Schwarzschild_model/betaOut",configCore.sm_betaOut);
    configCore.sm_constrainAngMomDist = settings.getValueBool("Schwarzschild_model/constrainAngMomDist", configCore.sm_constrainAngMomDist);
    configCore.sm_numAngMomBins = std::max<int>(1, settings.getValueInt("Schwarzschild_model/numAngMomBins", configCore.sm_numAngMomBins));
    return true;
}

void CSmileCore::saveSettings(const std::string &fileName)
{
    CConfigFile settings(fileName.empty() ? AppDir.toStdString()+DEFAULT_SETTINGS_FILE_NAME : fileName);
    // potential section
    for(size_t ig=0; ig<configPotential.size(); ig++) 
        writeConfigPotential(settings, &(configPotential[ig]), ig);
    // orbit integration section
    settings.setValue("Orbit/x", initCond.Pos[0], 16);
    settings.setValue("Orbit/y", initCond.Pos[1], 16);
    settings.setValue("Orbit/z", initCond.Pos[2], 16);
    settings.setValue("Orbit/vx",initCond.Vel[0], 16);
    settings.setValue("Orbit/vy",initCond.Vel[1], 16);
    settings.setValue("Orbit/vz",initCond.Vel[2], 16);
    settings.setValue("Orbit/E", initCondE, 16);
    settings.setValue("Orbit/useE", useICe);
    settings.setValue("Orbit/intTime", configCore.intTimeInPeriods);
    settings.setValue("Orbit/intTimeMax", configCore.intTimeMaxAdaptive);
    settings.setValue("Orbit/adaptiveTimeThreshold", configCore.adaptiveTimeThreshold);
    settings.setValue("Orbit/intTimeMinAbs", configCore.intTimeMinAbs);
    settings.setValue("Orbit/usePS", configCore.usePS);
    writeConfigOrbit(settings, &configOrbit);
    // frequency map
    settings.setValue("Frequency_map/numOrbitsStationary", configCore.fm_numOrbitsStationary);
    settings.setValue("Frequency_map/numOrbitsPrincipalPlane", configCore.fm_numOrbitsPrincipalPlane);
    settings.setValue("Frequency_map/numOrbitsYalpha", configCore.fm_numOrbitsYalpha);
    settings.setValue("Frequency_map/numOrbitsRandom", configCore.fm_numOrbitsRandom);
    // schwarzschild modelling
    QString comps;
    for(size_t ic=0; ic<configCore.densityComponents.size(); ic++) {
        if(ic>0) comps+=",";
        comps += QString::number(configCore.densityComponents[ic]);
    }
    settings.setValue("Schwarzschild_model/densityComponents", configCore.densityComponents.size()==0 ? "all" : comps.toStdString());
    settings.setValue("Schwarzschild_model/numOrbits", configCore.sm_numOrbits);
    settings.setValue("Schwarzschild_model/ICgenerator", getICgenNameByType(configCore.sm_ICgenType));
    settings.setValue("Schwarzschild_model/ICbeta", configCore.sm_ICbeta);
    settings.setValue("Schwarzschild_model/ICbetaAxi", configCore.sm_ICbetaAxi);
    settings.setValue("Schwarzschild_model/ICkAxi", configCore.sm_ICkAxi);
    settings.setValue("Schwarzschild_model/weightSkewFactorE", configCore.weightSkewFactorE);
    settings.setValue("Schwarzschild_model/weightSkewFactorL", configCore.weightSkewFactorL);
    settings.setValue("Schwarzschild_model/densityModelType", getSchwDataNameByType(configCore.sm_densDataType));
    settings.setValue("Schwarzschild_model/linesPerSegment", configCore.sm_linesPerSegment);
    settings.setValue("Schwarzschild_model/numAngularCoefs", configCore.sm_numAngularCoefs);
    settings.setValue("Schwarzschild_model/numRadialCoefs", configCore.sm_numRadialCoefs);
    settings.setValue("Schwarzschild_model/numVerticalCoefs", configCore.sm_numVerticalCoefs);
    settings.setValue("Schwarzschild_model/smAlpha", configCore.sm_Alpha);
    settings.setValue("Schwarzschild_model/numShells", configCore.sm_numShells);
    settings.setValue("Schwarzschild_model/innerShellMass", configCore.sm_innerShellMass);
    settings.setValue("Schwarzschild_model/outerShellMass", configCore.sm_outerShellMass);
    settings.setValue("Schwarzschild_model/smGridYtoX", configCore.sm_grid_q);
    settings.setValue("Schwarzschild_model/smGridZtoX", configCore.sm_grid_p);
    settings.setValue("Schwarzschild_model/numSamplingPoints", configCore.sm_numSamplingPoints);
    settings.setValue("Schwarzschild_model/chaoticMinFreqDiff", configCore.chaoticMinFreqDiff);
    settings.setValue("Schwarzschild_model/chaoticMinLambda", configCore.chaoticMinLambda);
    settings.setValue("Schwarzschild_model/chaoticPenalty", configCore.sm_chaoticPenalty);
    settings.setValue("Schwarzschild_model/maxWeight", configCore.sm_maxWeight);
    settings.setValue("Schwarzschild_model/regularization", configCore.sm_regularization);
    settings.setValue("Schwarzschild_model/constraintPenaltyLin", configCore.sm_constraintPenaltyLin);
    settings.setValue("Schwarzschild_model/constraintPenaltyQuad", configCore.sm_constraintPenaltyQuad);
    settings.setValue("Schwarzschild_model/constrainBeta", configCore.sm_constrainBeta);
    settings.setValue("Schwarzschild_model/betaIn",  configCore.sm_betaIn);
    settings.setValue("Schwarzschild_model/betaOut", configCore.sm_betaOut);
    settings.setValue("Schwarzschild_model/constrainAngMomDist", configCore.sm_constrainAngMomDist);
    settings.setValue("Schwarzschild_model/numAngMomBins", configCore.sm_numAngMomBins);
}

QString getDirName(const QString &fileName)  // return file path ending with "/"
{
    QFileInfo fi(fileName);
    QString fn = fi.absolutePath();
    if(!fn.endsWith('/')) fn += '/';
    return fn;
}

/* ----------- Class for performing console/interactive scripting ------------ */ 

void CSmileConsole::runScript(const char* scriptFileName)
{
    my_message("", "SMILE console version "SMILEVERSION"  build "__DATE__);
    timeStartup=QDateTime::currentDateTime();
    core->loadSettings();
    core->initPotential();
    core->initIC();
    if(scriptFileName!=NULL)
    {
        isConsoleInput=false;
        input.open(scriptFileName);
        if(!input)
        {
            my_error(FUNCNAME, "Error! cannot read script file "+(scriptFileName[0]==0?"(empty)":std::string(scriptFileName)));
            emit scriptDone();
        }
    }
    else
    {
        isConsoleInput=true;
    }
    connect(this, SIGNAL(scriptNextLine()), SLOT(scriptProcessLine()));
    connect(core, SIGNAL(signalOrbitLibraryFinished()), SLOT(scriptOrbitLibraryFinished()));
    connect(core, SIGNAL(signalSchwOptimizationFinished(const QString&)), SLOT(scriptTaskFinished(const QString&)));
    connect(core, SIGNAL(signalNbodyExportFinished(const QString&)), SLOT(scriptTaskFinished(const QString&)));
    connect(core, SIGNAL(signalTimer()), SLOT(scriptTimerEvent()));
    emit scriptNextLine();
}

void CSmileConsole::scriptTimerEvent()
{
    if(!core->orbitlib) return;
    std::cout << (convertToString(core->orbitlib->numComplete())+" out of "+
        convertToString(core->orbitlib->size())+" orbits done\015") << std::flush;
}

void CSmileConsole::scriptOrbitLibraryFinished()
{
    int timeElapsed = timeOrbitLibStarted.msecsTo(QDateTime::currentDateTime());
    if(core->orbitlib)
        std::cout << (convertToString(core->orbitlib->numComplete())+" orbits total, "+
            convertToString(1000.0*core->orbitlib->numComplete()/timeElapsed)+" orbits/s") << std::endl;
    emit scriptNextLine();
}

void CSmileConsole::scriptTaskFinished(const QString& message)
{
    my_message("", message.toStdString());
    emit scriptNextLine();
}

void CSmileConsole::scriptProcessLine()
{
    /*std::cout << "> ";
    if(!input.atEnd())
        Line = input.readLine().trimmed();
    else
        Line = "Exit";*/
    std::string str;
    if(isConsoleInput) {
#ifdef HAVE_READLINE
        char* rline=readline("> ");
        if(rline) str=std::string(rline); 
        else str="Exit";
        add_history(rline);
#else
        std::cout << "> ";
        if(!std::getline(std::cin, str))  // eof
            str="Exit";
#endif
    } else {
        if(!std::getline(input, str) || !input || input.eof()) 
            str="Exit";
    }
    std::string::size_type indx=str.find_last_not_of("\r\n");
    if(indx!=std::string::npos) str.erase(indx+1);
    QString Line=QString::fromStdString(str);
    if(Line=="" || Line.startsWith("#"))
    {
        emit scriptNextLine();
        return;
    }
    if(Line.startsWith("Exit", Qt::CaseInsensitive) || Line.startsWith("Q", Qt::CaseInsensitive) || 
        Line.startsWith("Bye", Qt::CaseInsensitive))
    {
        std::cout << "Done!  Total runtime: " << 
            timeStartup.secsTo(QDateTime::currentDateTime()) << " seconds" << std::endl;
        emit scriptDone();
        return;
    }
    if(!isConsoleInput) 
        std::cout << ("> "+str) << std::endl;

    // parse input string and perform desired action
    if(Line.startsWith("ReadIni", Qt::CaseInsensitive) || Line.startsWith("ReadConfig", Qt::CaseInsensitive) || 
        Line.startsWith("LoadIni", Qt::CaseInsensitive) || Line.startsWith("LoadConfig", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = (i1>=2 && i2>i1+3) ? Line.mid(i1+2, i2-i1-3) : (i1==-1&&i2==-1 ? DEFAULT_SETTINGS_FILE_NAME : "");
        if(!FileName.isEmpty())
        {
            if(core->loadSettings(FileName.toStdString()))
            {
                std::cout << "Loaded config from " << FileName.toStdString() << std::endl;
                core->initPotential();
                core->initIC();
            }
            else
                std::cout << "Error! Cannot load config from " << FileName.toStdString() << std::endl;
        }
        else std::cout << "Error! No valid filename specified" << std::endl;
    }
    else if(Line.startsWith("ExportPotential", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = Line.mid(i1+2, i2-i1-3);
        if(i1>0 && i2>i1)
        {
            if(core->exportPotential(FileName))
                std::cout << "Potential exported to " << FileName.toStdString() << std::endl;
            else
                std::cout << "Error! Cannot export potential to " << FileName.toStdString() << std::endl;
        }
        else std::cout << "Error! No filename specified" << std::endl;
    }
    else if(Line.startsWith("BuildFreq", Qt::CaseInsensitive) || 
            Line.startsWith("BuildSchw", Qt::CaseInsensitive) || Line.startsWith("BuildModel", Qt::CaseInsensitive))
    {
        if(Line.startsWith("BuildFreq", Qt::CaseInsensitive))
            core->buildFreqMapIC(core->initCondE, Line.contains("exist", Qt::CaseInsensitive));
        else
            core->buildSchwIC(Line.contains("exist", Qt::CaseInsensitive));
        core->startOrbitLibrary(SLOT(coreOrbitLibraryFinished()));
        timeOrbitLibStarted=QDateTime::currentDateTime();
        return;
    }
    else if(Line.startsWith("SchwLinear", Qt::CaseInsensitive) || Line.startsWith("SchwQuad", Qt::CaseInsensitive))
    {
        if(core->schwData.size()>0 && core->orbitlib!=NULL && core->orbitlib->size()>0)
        {
            CBasicSchwModel::MODELTYPE modeltype = 
                Line.startsWith("SchwQuad", Qt::CaseInsensitive) ? CBasicSchwModel::SM_QUADRATIC : CBasicSchwModel::SM_LINEAR;
            core->schwStartOptimization(modeltype);
            return;
        }
        else
            std::cout << "No Schwarzschild model or orbit library defined" << std::endl;
    }
    else if(Line.startsWith("ExportOrbits", Qt::CaseInsensitive) || Line.startsWith("SaveOrbits", Qt::CaseInsensitive) || 
            Line.startsWith("ExportModel", Qt::CaseInsensitive)  || Line.startsWith("SaveModel", Qt::CaseInsensitive))
    {
        if(core->orbitlib!=NULL)
        {
            int i1=Line.indexOf("(""");
            int i2=Line.indexOf(""")");
            QString FileName = Line.mid(i1+2, i2-i1-3);
            if(i1>0 && i2>i1)
            {
                bool withmodel=Line.startsWith("ExportModel", Qt::CaseInsensitive) || Line.startsWith("SaveModel", Qt::CaseInsensitive);
                if(!core->exportOrbitLib(FileName, withmodel))
                    std::cout << "Error writing file: " << FileName.toStdString() << std::endl;
            }
            else std::cout << "No filename specified" << std::endl;
        }
        else
            std::cout << "No orbit library defined" << std::endl;
    }
    else if(Line.startsWith("ImportOrbits", Qt::CaseInsensitive) || Line.startsWith("LoadOrbits", Qt::CaseInsensitive) || 
            Line.startsWith("ImportModel", Qt::CaseInsensitive)  || Line.startsWith("LoadModel", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(""");
        int i2=Line.indexOf(""")");
        QString FileName = Line.mid(i1+2, i2-i1-3);
        if(i1>0 && i2>i1)
        {
            bool withmodel=Line.startsWith("ImportModel", Qt::CaseInsensitive) || Line.startsWith("LoadModel", Qt::CaseInsensitive);
            if(QFile::exists(FileName+".ini") && core->loadSettings(FileName.toStdString()+".ini"))
            {
                std::cout << "Loaded config from " << FileName.toStdString() << ".ini" << std::endl;
                core->initPotential();
                core->initIC();
            }
            if(!core->importOrbitLib(FileName, withmodel))
                std::cout << "Error reading file: " << FileName.toStdString() << std::endl;
            else {
                std::cout << core->orbitlib->numComplete() << " orbits loaded" << std::endl;
            }
        }
        else std::cout << "No filename specified" << std::endl;
    }
    else if(Line.startsWith("ExportStats", Qt::CaseInsensitive))
    {
        if(core->orbitlib!=NULL && core->schwData.size()>0)
        {
            int i1=Line.indexOf("(""");
            int i2=Line.indexOf(""")");
            QString FileName = Line.mid(i1+2, i2-i1-3);
            if(i1>0 && i2>i1)
            {
                if(!core->exportSchwModelStats(FileName))
                    std::cout << "Error writing file: " << FileName.toStdString() << std::endl;
            }
            else std::cout << "No filename specified" << std::endl;
        }
        else
            std::cout << "No orbit library or Schwarzschild model defined" << std::endl;
    }
    else if(Line.startsWith("ExportNbody", Qt::CaseInsensitive))
    {
        if(core->orbitlib)
        {
            int i1=Line.indexOf("(");
            int i2=Line.lastIndexOf(")");
            QString paramstr = Line.mid(i1+1, i2-i1-1);
            QStringList params=paramstr.split(QRegExp("[,\\s]+"));
            if(params.count()<2)
                std::cout << "Syntax: ExportNbody(NumBodies, \"FileName\" [, \"FileFormat\"=\"" << 
                formatsIOSnapshot.front() << "\" [, RefineFactor=0]])" << std::endl;
            else{
                int NPoints=static_cast<int>(params[0].toDouble());
                i1=params[1].indexOf("\"");
                i2=params[1].lastIndexOf("\"");
                QString fileName=params[1].mid(i1+1, i2-i1-1);
                std::string fileFormat=formatsIOSnapshot.front();
                if(params.count()>=3) {
                    i1=params[2].indexOf("\"");
                    i2=params[2].lastIndexOf("\"");
                    fileFormat=params[2].mid(i1+1, i2-i1-1).toStdString();
                    bool found=false;
                    for(size_t i=0; i<formatsIOSnapshot.size(); i++)
                        if(strings_equal(fileFormat, formatsIOSnapshot[i])) found=true;
                    if(!found) {
                        fileFormat=formatsIOSnapshot.front();
                        std::cout << "Warning, incorrect specification for file format, using \""<<fileFormat<<"\" instead" << std::endl;
                    }
                }
                int refineFactor=0;
                if(params.count()>=4)
                    refineFactor=params[3].toInt();
                if(NPoints==0)
                    std::cout << "Incorrect first parameter: NumBodies" << std::endl;
                else
                if(fileName!="")
                {
                    core->schwExportNbody(NPoints, fileName, QString(fileFormat.c_str()), refineFactor);
                    return;
                }
                else std::cout << "No filename specified" << std::endl;
            }
        }
        else
            std::cout << "No orbit library defined" << std::endl;
    }
    else if(Line.startsWith("RenderDensity", Qt::CaseInsensitive))
    {
        int i1=Line.indexOf("(");
        int i2=Line.lastIndexOf(")");
        QString paramstr = Line.mid(i1+1, i2-i1-1);
        QStringList params=paramstr.split(QRegExp("[,\\s]+"));
        if(params.count()<2)
            std::cout << "Syntax: RenderDensity(NumBodies, \"FileName\"[, \"FileFormat\"=\"" << 
            formatsIOSnapshot.front() << "\" [, RefineFactor=0]])" << std::endl;
        else {
            int NPoints=static_cast<int>(params[0].toDouble());
            i1=params[1].indexOf("\"");
            i2=params[1].lastIndexOf("\"");
            QString fileName=params[1].mid(i1+1, i2-i1-1);
            std::string fileFormat=formatsIOSnapshot.front();
            if(params.count()>=3) {
                i1=params[2].indexOf("\"");
                i2=params[2].lastIndexOf("\"");
                fileFormat=params[2].mid(i1+1, i2-i1-1).toStdString();
                bool found=false;
                for(size_t i=0; i<formatsIOSnapshot.size(); i++)
                    if(strings_equal(fileFormat, formatsIOSnapshot[i])) found=true;
                if(!found) {
                    fileFormat=formatsIOSnapshot.front();
                    std::cout << "Warning, incorrect specification for file format, using \""<<fileFormat<<"\" instead" << std::endl;
                }
            }
            if(NPoints==0)
                std::cout << "Incorrect first parameter: NumBodies" << std::endl;
            else if(fileName=="")
                std::cout << "No filename specified" << std::endl;
            else {
                if(!core->density)
                    core->initDensity();
                CPointMassSet<float> samplingPoints;
                CICGeneratorDensitySampler icd(core->density);
                icd.createInitialConditions(NPoints, &samplingPoints);
                CBasicIOSnapshot* snap = createIOSnapshotWrite(fileFormat, fileName.toStdString(), 
                    "SMILE version="SMILEVERSION" potential="+std::string(core->density->PotentialName()));
                if(snap) { 
                    snap->writeSnapshot(&samplingPoints);
                    delete snap;
                } else {
                    std::cout << "Invalid export file format: "+fileFormat;
                }
            }
        }
    }
    else 
        std::cout << "Invalid command" << std::endl;
    emit scriptNextLine();
}

/* -------- Helper threads that run various tasks in parallel to main program ------ */

// run a single orbit integration
void COrbitIntWorker::run()
{
    orbit->run();
    emit calcFinished();
    deleteLater();
}

// run orbit library integration in parallel threads
volatile int nthreads=0;
QMutex mutexL;
QMutex mutexC;

void COrbitLibWorker::run()
{
    mutexC.lock();
    ++nthreads;
    mutexC.unlock();
    finished=false;
    while(!finished)
    {
        mutexL.lock();
        int no = orbits->findUnstartedOrbit();
        mutexL.unlock();
        if(no>=0) {
            orbits->runOrbit(no);
            const COrbitDesc* last=orbits->getOrbitDesc(no);
            if(last!=NULL && last->getInfoNum()>=1 && 
                last->getInfo(0)->infoType()==CBasicInformation::IT_TRAJ_ANALYSIS &&
                static_cast<const COrbitInformation<float>* >(last->getInfo(0))->getDescription()=="Failed")   // indicates no memory
                    halt();   // need to stop entire orbit library integration
        }
        else finished=true;
    }
    mutexC.lock();
    --nthreads;
    if(nthreads==0)
        emit calcFinished();   // only once
    mutexC.unlock();
    deleteLater();
}

void CSchwarzschildWorker::doWork()
{
    std::string result;
    if(model==NULL) 
        result="Error, no Schw.model object is present!";
    else
        result=model->performModelling();
    emit message(QString::fromStdString(result));
    emit workDone();
    deleteLater();
};

void CNbodyExportStatWorker::doWork()
{
    handler = new CPointMassSetHandler<float>(points);
    double T=0, W=0, Wext=0;
    handler->computeVirialRatio(&T, &W);
    delete handler;
    handler=NULL;
    if(extraPotential!=NULL) {  // add contribution from external potential
        for(size_t i=0; i<points->size(); i++) {
            double pos[N_DIM]={points->at(i).first.Pos[0], points->at(i).first.Pos[1], points->at(i).first.Pos[2]};
            double force[N_DIM];
            extraPotential->Force(pos, 0, force, NULL);
            Wext -= points->at(i).second * (pos[0]*force[0]+pos[1]*force[1]+pos[2]*force[2]);
            //Wext -= points->at(i).second * extraPotential->Phi(
            //    points->at(i).first.Pos[0], points->at(i).first.Pos[1], points->at(i).first.Pos[2]);  // doesn't compute correctly
        }
    }
    delete points;   // deleting an object which was created externally and passed 
    // as a parameter to the constructor, but is no longer needed
    emit exportFinished(info + 
        "\nT="+QString::number(T, 'g', 5) + 
        ", W="+QString::number(W, 'g', 5) + 
        (extraPotential!=NULL ? ", Wext="+QString::number(Wext, 'g', 5) : "") +
        "\nVirial ratio="+QString::number(2*T/(W+Wext), 'g', 5));
    if(extraPotential!=NULL)
        delete extraPotential;
    emit workDone();
    deleteLater();
}

}
