#include "gui.h"

#include <cmath>
#include <vector>
#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <gsl/gsl_integration.h>

#include <QKeyEvent>
#include <QFileDialog>
#include <QMessageBox>
#include <QPrinter>
#include <QImage>
#include <QMap>
#include <QInputDialog>
#include <QMetaObject>

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qwt_text.h>
#include <qwt_symbol.h>
#include <qwt_plot_picker.h>
#include <qwt_plot_panner.h>
#include <qwt_plot_zoomer.h>
#include <qwt_scale_engine.h>
#include <qwt_plot_marker.h>

#include "core.h"
#include "potential.h"
#include "odeint.h"
#include "orbitan.h"
#include "orbitlib.h"
#include "schwarzschild.h"
#include "utils.h"
#include "iosnapshot.h"
#include "icgenerator.h"
#include "massmodelaxi.h"

#ifdef USE3DFACETS
// for triangulation thread
#include <QVector>
#include <QProcess>
#include <QTextStream>
#endif

#if QT_VERSION >= 0x050000
#define GETINTEGER getInt
#else
#define GETINTEGER getInteger
#endif

#if QWT_VERSION >= 0x060000
#include <qwt_plot_renderer.h>
#include <qwt_picker_machine.h>
typedef QwtLogScaleEngine QwtLog10ScaleEngine;
QwtSymbol* QWTSYM(QwtSymbol::Style style, int size, QColor color)
{
    QwtSymbol* sym = new QwtSymbol;
    sym->setStyle(style);
    sym->setSize(size);
    sym->setPen(QColor(Qt::black), 0, Qt::NoPen);
    sym->setBrush(color);
    return sym;
}
#define SETDATA setSamples
#define CONNECTPICKER(obj, slot) \
connect(obj, SIGNAL(selected(const QPointF &)), SLOT(slot(const QPointF &)));
#else
QwtSymbol QWTSYM(QwtSymbol::Style style, int size, QColor color)
{
    QwtSymbol sym;
    sym.setStyle(style);
    sym.setSize(size);
    sym.setPen(Qt::NoPen);
    sym.setBrush(color);
    return sym;
}
#define SETDATA setData
#define CONNECTPICKER(obj, slot) \
connect(obj, SIGNAL(selected(const QwtDoublePoint &)), SLOT(slot(const QwtDoublePoint &)));
#endif

namespace smile{

/// The global variable for the instance of GUI object is used to redirect text messages to it
CSmileGUI* GUI;

/// display the message string in the bottom-right text area of the main window
void my_message_gui(const std::string &origin, const std::string &message)
{
    QMetaObject::invokeMethod(GUI, "info", Q_ARG(QString, QString::fromStdString("{"+origin+"} "+message)));
}

/// pop up a message box with the error string
void my_error_gui(const std::string &origin, const std::string &message)
{
    my_stderr_show_message(origin, message);
    QMetaObject::invokeMethod(GUI, "error", Q_ARG(QString, QString::fromStdString(message)));
}

void CSmileGUI::info(const QString &message)
{
    std::string origin="";
    QString msg=message;
    int ind=message.indexOf('}');
    if(message.startsWith('{')&&ind>0) {
        origin=message.mid(1, ind-1).toStdString();
        msg=message.mid(ind+2);
    }
#ifndef DEBUGPRINT
    if(!origin.empty())  // if not DEBUGPRINT, then only dump messages to console if they were called by my_message_gui
#endif
    my_stderr_show_message(origin, msg.toStdString());
    ui.label_info->setPlainText(msg);
    ui.label_info->repaint();  // to ensure immediate displaying
}

void CSmileGUI::error(const QString &message)
{
    QMessageBox::warning(this, "Error", message);
}

CSmileGUI::CSmileGUI()
    : QMainWindow(NULL, 0)
{
    GUI=this;
    ui.setupUi(this);

    core=new CSmileCore();  // initialize the instance of "business logic" object
    my_error_ptr=&my_error_gui;
    my_message_ptr=&my_message_gui;

    // connect signals to the core
    connect(core, SIGNAL(signalOrbitFinished()), SLOT(CalcFinished()));
    connect(core, SIGNAL(signalOrbitLibraryFinished()), this, SLOT(OrbitLibraryFinished()));
    connect(core, SIGNAL(signalSchwOptimizationFinished(const QString&)), SLOT(SchwOptimizationFinished(const QString&)));
    connect(core, SIGNAL(signalNbodyReintegrationFinished()), SLOT(NbodyReintegrationFinished()));
    connect(core, SIGNAL(signalNbodyExportFinished(const QString&)), SLOT(NbodyExportFinished(const QString&)));
    connect(core, SIGNAL(signalTimer()), SLOT(timerEventCore()));
    connect(this, SIGNAL(stopOrbitLibrary()), core, SIGNAL(signalStopOrbitLibrary()));
    connect(this, SIGNAL(stopExportNbody()),  core, SIGNAL(signalStopExportNbody()));

    Eprev=Lz2prev=0;
    // fill potential, density, symmetry and SM types dropdown list
    size_t max_enum=64;
    ui.selectPotentialType->clear();
    ui.selectDensityType->clear();
    ui.selectSymmetryType->clear();
    for(size_t i=0; i<max_enum; i++) {
        const char* name=getPotentialNameByType(static_cast<CDensity::POTENTIALTYPE>(i));
        if(name && name[0]) 
            ui.selectPotentialType->insertItem(max_enum, name);
        name=getDensityNameByType(static_cast<CDensity::POTENTIALTYPE>(i));
        if(name && name[0]) 
            ui.selectDensityType->insertItem(max_enum, name);
        name=getSymmetryNameByType(static_cast<CDensity::SYMMETRYTYPE>(i));
        if(name && name[0]) 
            ui.selectSymmetryType->insertItem(max_enum, name);
    }
    ui.select_smType->clear();
    ui.select_smType->insertItem(max_enum, CSchwDataDensGridClassic::myName());
    ui.select_smType->insertItem(max_enum, CSchwDataDensGridCylindrical::myName());
    ui.select_smType->insertItem(max_enum, CSchwDataDensSHMesh::myName());
    ui.select_smType->insertItem(max_enum, CSchwDataDensSHBSE::myName());
    std::string integratorDefault=getIntegratorNameByType(static_cast<CBasicOdeIntegrator::STEPPERKIND>(max_enum));
    ui.selectIntegratorType->insertItem(max_enum, QString::fromStdString(integratorDefault));
    for(size_t i=0; i<max_enum; i++) {
        std::string integratorName=getIntegratorNameByType(static_cast<CBasicOdeIntegrator::STEPPERKIND>(i));
        if(integratorName!=integratorDefault)
            ui.selectIntegratorType->insertItem(max_enum, QString::fromStdString(integratorName));
    }

    // orbit plot
    PlotOrbit = new QwtPlot(QwtText("Orbit"), ui.frame_orbitarea);
    CurveOrbit= new QwtPlotCurve("");
    CurveOrbit->setPen(QColor(Qt::blue));
    CurveOrbit->attach(PlotOrbit);
    CurveEquipotential= new QwtPlotCurve("");
    CurveEquipotential->setPen(QColor(Qt::red));
    CurveEquipotential->attach(PlotOrbit);
    PlotOrbit->show();
    QVBoxLayout* LPlotOrbit = new QVBoxLayout(ui.frame_orbitarea);
    LPlotOrbit->addWidget(PlotOrbit);
    ZoomerOrbit= new QwtPlotZoomer2(QwtPicker::ActiveOnly, PlotOrbit->canvas());
    ZoomerOrbit->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerOrbit = new QwtPlotPanner(PlotOrbit->canvas());
    PannerOrbit->setMouseButton(Qt::MidButton);

    PlotOrbitParam = new QwtPlot(QwtText("Distribution of orbital parameters"), ui.frame_orbitarea);
    CurveOrbitParam= new QwtPlotCurve("");
    CurveOrbitParam->attach(PlotOrbitParam);
    PlotOrbitParam->hide();
    LPlotOrbit->addWidget(PlotOrbitParam);
    ZoomerOrbitParam= new QwtPlotZoomer2(QwtPicker::ActiveOnly, PlotOrbitParam->canvas());
    ZoomerOrbitParam->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerOrbitParam = new QwtPlotPanner(PlotOrbitParam->canvas());
    PannerOrbitParam->setMouseButton(Qt::MidButton);

#ifdef HAVE_QWT3D
    // 3d orbit
    Plot3d=new Qwt3D::SurfacePlot(ui.frame_orbitarea);
    Plot3d->setCoordinateStyle(Qwt3D::FRAME);
    Plot3d->setRotation(90,0,0);
    QColor bkg=ui.tabWidget->palette().color(QPalette::Background);
    Plot3d->setBackgroundColor(Qwt3D::RGBA(bkg.redF(), bkg.greenF(), bkg.blueF()));
    Plot3d->setMeshColor(Qwt3D::RGBA(0.2,0,0.4));
    for (unsigned i=0; i!=Plot3d->coordinates()->axes.size(); ++i)
    {
      Plot3d->coordinates()->axes[i].setMajors(7);
      Plot3d->coordinates()->axes[i].setMinors(4);
    }
    LPlotOrbit->addWidget(Plot3d);
#ifdef USE3DFACETS
    TriangThr = new CTriangThread(core->AppDir+"qdelaunay");
    connect(TriangThr, SIGNAL(finished(bool)), SLOT(redrawOrbit(bool)));
    Plot3d->setLightComponent(GL_DIFFUSE, 0.3);
    Plot3d->setMaterialComponent(GL_SPECULAR, 0.0);
    Plot3d->enableLighting();
    Plot3d->illuminate();
    Plot3d->updateGL();
#endif
#endif
    ui.progressBarOrbit->hide();
    ui.groupBox_2dplane->show();
    ui.groupBox_3dmesh->hide();

    // potential plot
    PlotPotential = new QwtPlot(QwtText(""), ui.frame_potentialplot);
    CurvePotential= new QwtPlotCurve("");
    CurvePotential->setPen(QColor(Qt::blue));
    CurvePotential->attach(PlotPotential);
    CurvePotential1= new QwtPlotCurve("");
    CurvePotential1->setPen(QColor(Qt::green));
    CurvePotential1->attach(PlotPotential);
    CurvePotential2= new QwtPlotCurve("");
    CurvePotential2->setPen(QColor(Qt::red));
    CurvePotential2->attach(PlotPotential);
    PlotPotential->show();
    ZoomerPotential= new QwtPlotZoomer2(QwtPicker::ActiveOnly, PlotPotential->canvas());
    ZoomerPotential->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerPotential = new QwtPlotPanner(PlotPotential->canvas());
    PannerPotential->setMouseButton(Qt::MidButton);
    QVBoxLayout* LPlotPotential = new QVBoxLayout(ui.frame_potentialplot);
    LPlotPotential->addWidget(PlotPotential);

    // poincare plot
    PlotPoincare = new QwtPlot(QwtText("Poincare section"), ui.frame_poincarearea);
    PlotPoincare->show();
    QVBoxLayout * LPlotPoincare = new QVBoxLayout(ui.frame_poincarearea);
    LPlotPoincare->addWidget(PlotPoincare);
    QwtPlotPicker *PickerPoincare = new QwtPlotPicker2(QwtPicker::AlwaysOn, PlotPoincare->canvas());
    //PickerPoincare->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    PickerPoincare->setMousePattern(QwtEventPattern::MouseSelect1, Qt::LeftButton, Qt::ControlModifier);
    CONNECTPICKER(PickerPoincare, PSclicked);
    ZoomerPoincare = new QwtPlotZoomer2(QwtPicker::AlwaysOn, PlotPoincare->canvas());
    ZoomerPoincare->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton, Qt::ControlModifier);
    ZoomerPoincare->setMousePattern(QwtEventPattern::MouseSelect2, Qt::RightButton, Qt::ControlModifier);
    ZoomerPoincare->setKeyPattern(QwtEventPattern::KeySelect1, Qt::Key_Space);
    QwtPlotPanner* PannerPoincare = new QwtPlotPanner(PlotPoincare->canvas());
    PannerPoincare->setMouseButton(Qt::MidButton);
    ui.comboBox_PSdisplayCoord->setCurrentIndex(0);
    ui.comboBox_PSsectionCoord->setCurrentIndex(1);

    // frequency plot
    PlotFreq = new QwtPlot(QwtText("Frequencies"), ui.frame_frequencies);
    CurveFreqX= new QwtPlotCurve("X");
    CurveFreqX->setPen(QColor(Qt::blue));
    CurveFreqX->attach(PlotFreq);
    CurveFreqY= new QwtPlotCurve("Y");
    CurveFreqY->setPen(QColor(Qt::darkGreen));
    CurveFreqY->attach(PlotFreq);
    CurveFreqZ= new QwtPlotCurve("Z");
    CurveFreqZ->setPen(QColor(Qt::darkRed));
    CurveFreqZ->attach(PlotFreq);
    CurveFreqPeaks= new QwtPlotCurve();
    CurveFreqPeaks->setStyle(QwtPlotCurve::Sticks);
    CurveFreqPeaks->attach(PlotFreq);
    CurveFreqPeaksP= new QwtPlotCurve();
    CurveFreqPeaksP->setStyle(QwtPlotCurve::Sticks);
    CurveFreqPeaksP->setPen(QColor(Qt::white));
    CurveFreqPeaksP->attach(PlotFreq);
    PlotFreq->show();
    QVBoxLayout* LPlotFreq = new QVBoxLayout(ui.frame_frequencies);
    LPlotFreq->addWidget(PlotFreq);
    ZoomerFreq = new QwtPlotZoomer2(QwtPicker::AlwaysOn, PlotFreq->canvas());
    ZoomerFreq->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton);
    QwtPlotPanner* PannerFreq = new QwtPlotPanner(PlotFreq->canvas());
    PannerFreq->setMouseButton(Qt::MidButton);

    // Lyapunov plot
    PlotLyapunov = new QwtPlot(QwtText("Lyapunov number"), ui.tab_l);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
    PlotLyapunov->setAxisScaleEngine(QwtPlot::yRight, new QwtLog10ScaleEngine);
    CurveDeviation = new QwtPlotCurve("");
    CurveDeviation->setPen(QColor(Qt::red));
    CurveDeviation->attach(PlotLyapunov);
    CurveDeviation->setYAxis(QwtPlot::yRight);
    PlotLyapunov->enableAxis(QwtPlot::yRight);
    CurveLyapunov = new QwtPlotCurve("");
    CurveLyapunov->setPen(QColor(Qt::blue));
    CurveLyapunov->attach(PlotLyapunov);
    PlotLyapunov->show();
    /*QwtPlotPicker *PickerLyapunov = */new QwtPlotPicker2(QwtPicker::AlwaysOn, PlotLyapunov->canvas());
    QVBoxLayout* LPlotLyapunov = new QVBoxLayout(ui.tab_l);
    LPlotLyapunov->addWidget(PlotLyapunov);

    // Frequency Map plot
    PlotFreqMap = new QwtPlot(QwtText("Frequency map"), ui.frame_fmarea);
    CurveFreqMapC = new QwtPlotCurve("");
    CurveFreqMapC->setSymbol(QWTSYM(QwtSymbol::Rect, 2, QColor(Qt::red)));
    CurveFreqMapC->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapC->attach(PlotFreqMap);
    CurveFreqMapR = new QwtPlotCurve("");
    CurveFreqMapR->setSymbol(QWTSYM(QwtSymbol::Rect, 2, QColor(Qt::blue)));
    CurveFreqMapR->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapR->attach(PlotFreqMap);
    drawFreqMapResonances();
    PlotFreqMap->show();
    QVBoxLayout* LPlotFreqMap = new QVBoxLayout(ui.frame_fmarea);
    LPlotFreqMap->addWidget(PlotFreqMap);
    QwtPlotPicker *PickerFreqMap = new QwtPlotPicker2(QwtPicker::AlwaysOn, PlotFreqMap->canvas());
    PickerFreqMap->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    CONNECTPICKER(PickerFreqMap, FMclicked);
    ZoomerFreqMap = new QwtPlotZoomer2(QwtPicker::AlwaysOn, PlotFreqMap->canvas());
    ZoomerFreqMap->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton, Qt::ControlModifier);
    ZoomerFreqMap->setMousePattern(QwtEventPattern::MouseSelect2, Qt::RightButton, Qt::ControlModifier);
    ZoomerFreqMap->setKeyPattern(QwtEventPattern::KeySelect1, Qt::Key_Space);
    QwtPlotPanner* PannerFreqMap = new QwtPlotPanner(PlotFreqMap->canvas());
    PannerFreqMap->setMouseButton(Qt::MidButton);

    PlotFreqMapHisto = new QwtPlot(QwtText("Histograms"), ui.frame_fmarea);
    PlotFreqMapHisto->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine);
    CurveFreqMapHisto = new QwtPlotCurve("");
    CurveFreqMapHisto->setStyle(QwtPlotCurve::Steps);
    CurveFreqMapHisto->attach(PlotFreqMapHisto);
    LPlotFreqMap->addWidget(PlotFreqMapHisto);
    /*QwtPlotPicker *PickerFreqMapHisto = */new QwtPlotPicker2(QwtPicker::AlwaysOn, PlotFreqMapHisto->canvas());
    PlotFreqMapHisto->hide();

    PlotFreqMapSS = new QwtPlot(QwtText("Start-space"), ui.frame_fmarea);
    CurveFreqMapSSR = new QwtPlotCurve("");
    CurveFreqMapSSR->setSymbol(QWTSYM(QwtSymbol::Rect, 4, QColor(Qt::blue)));
    CurveFreqMapSSR->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapSSR->attach(PlotFreqMapSS);
    CurveFreqMapSSC = new QwtPlotCurve("");
    CurveFreqMapSSC->setSymbol(QWTSYM(QwtSymbol::Rect, 4, QColor(Qt::red)));
    CurveFreqMapSSC->setStyle(QwtPlotCurve::Dots);
    CurveFreqMapSSC->attach(PlotFreqMapSS);
    QwtPlotPicker *PickerFreqMapSS = new QwtPlotPicker2(QwtPicker::ActiveOnly, PlotFreqMapSS->canvas());
    PickerFreqMapSS->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    CONNECTPICKER(PickerFreqMapSS, FMclicked);
    LPlotFreqMap->addWidget(PlotFreqMapSS);
    PlotFreqMapSS->hide();
    ui.progressBarFM->hide();

    // Schwarzschild plot
    QVBoxLayout* LPlotSchw = new QVBoxLayout(ui.frame_smarea);
    PlotSchw = new QwtPlot(QwtText("Schwarzschild model grid"), ui.frame_smarea);
    CurveSchwCellF = new QwtPlotCurve("");
    CurveSchwCellF->setSymbol(QWTSYM(QwtSymbol::Diamond, 8, QColor(Qt::blue)));
    CurveSchwCellF->setStyle(QwtPlotCurve::Dots);
    CurveSchwCellF->attach(PlotSchw);
    CurveSchwCellI= new QwtPlotCurve("");
    CurveSchwCellI->setSymbol(QWTSYM(QwtSymbol::Diamond, 8, QColor(Qt::red)));
    CurveSchwCellI->setStyle(QwtPlotCurve::Dots);
    CurveSchwCellI->attach(PlotSchw);
    QwtPlotPicker *PickerFreqMapSM = new QwtPlotPicker2(QwtPicker::ActiveOnly, PlotSchw->canvas());
    PickerFreqMapSM->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    CONNECTPICKER(PickerFreqMapSM, SMclicked);
    PlotSchw->show();
    LPlotSchw->addWidget(PlotSchw);
    ui.progressBarSM->hide();
    PlotSchwHisto = new QwtPlot(QwtText("Schwarzschild model orbit weights"), ui.frame_smarea);
    CurveSchwHistoR = new QwtPlotCurve("");
    CurveSchwHistoR->setStyle(QwtPlotCurve::Dots);
    CurveSchwHistoR->setSymbol(QWTSYM(QwtSymbol::Rect, 2, QColor(Qt::blue)));
    CurveSchwHistoR->attach(PlotSchwHisto);
    CurveSchwHistoC = new QwtPlotCurve("");
    CurveSchwHistoC->setStyle(QwtPlotCurve::Dots);
    CurveSchwHistoC->setSymbol(QWTSYM(QwtSymbol::Rect, 2, QColor(Qt::red)));
    CurveSchwHistoC->attach(PlotSchwHisto);
    LPlotSchw->addWidget(PlotSchwHisto);
    QwtPlotPicker *PickerSchwHisto = new QwtPlotPicker2(QwtPicker::AlwaysOn, PlotSchwHisto->canvas());
    PickerSchwHisto->setMousePattern(QwtEventPattern::MouseSelect1, Qt::RightButton);
    CONNECTPICKER(PickerSchwHisto, SMHistoClicked);
    PlotSchwHisto->hide();

    // connect various events to change corresponding parameters in core
    connect(ui.lineEdit_x, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_y, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_z, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vx, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vy, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_vz, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_Omega, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.lineEdit_E, SIGNAL(textChanged(const QString&)), SLOT(ICchanged()));
    connect(ui.selectPotentialIndex, SIGNAL(currentIndexChanged(int)), SLOT(potentialIndexChanged()));
    connect(ui.lineEdit_q, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_p, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_mass, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_scalerad, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_scalerad2, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Gamma, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_SersicIndex, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Ncoefs_radial, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Ncoefs_vertical, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Ncoefs_angular, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Alpha, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_Rmax, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_eps, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.lineEdit_tol, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.selectPotentialType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.selectDensityType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.selectSymmetryType, SIGNAL(currentIndexChanged(int)), SLOT(potentialChanged()));
    connect(ui.buttonChooseNbodyFile, SIGNAL(clicked()), SLOT(chooseNbodyFile()));
    connect(ui.lineEdit_NbodyFile, SIGNAL(returnPressed()), SLOT(typeinNbodyFile()));
    connect(ui.lineEdit_Mbh, SIGNAL(textChanged(const QString&)), SLOT(potentialChanged()));
    connect(ui.selectIntegratorType, SIGNAL(currentIndexChanged(int)), SLOT(integratorChanged()));
    connect(ui.lineEdit_tol_abs, SIGNAL(textChanged(const QString&)), SLOT(integratorChanged()));
    connect(ui.lineEdit_tol_rel, SIGNAL(textChanged(const QString&)), SLOT(integratorChanged()));
    connect(ui.lineEdit_tol_ias, SIGNAL(textChanged(const QString&)), SLOT(integratorChanged()));
    connect(ui.lineEdit_timestep_factor, SIGNAL(textChanged(const QString&)), SLOT(integratorChanged()));
    connect(ui.lineEdit_relax_rate, SIGNAL(textChanged(const QString&)), SLOT(integratorChanged()));
    connect(ui.radioButton_2dorbit, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_3dline, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_3dmesh, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opPericenter, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opLPeri, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opL, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opEnergy, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_opAngMom, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_xy, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_xz, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_2d_yz, SIGNAL(toggled(bool)), SLOT(redrawOrbit(bool)));
    connect(ui.radioButton_ptPotential, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_ptDensity  , SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_ptRotCurve , SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_ptFrequencies, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_ptSurfaceDensity, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_axis_x, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_axis_y, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.radioButton_axis_z, SIGNAL(toggled(bool)), SLOT(redrawPotentialB(bool)));
    connect(ui.comboBox_ptComponents, SIGNAL(currentIndexChanged(int)), SLOT(redrawPotential()));
    connect(ui.checkBox_ptLogScale, SIGNAL(toggled(bool)), SLOT(redrawPotential()));
    connect(ui.comboBox_PSdisplayCoord, SIGNAL(currentIndexChanged(int)), SLOT(PScoordChanged()));
    connect(ui.comboBox_PSsectionCoord, SIGNAL(currentIndexChanged(int)), SLOT(PScoordChanged()));
    connect(ui.radioButton_fmmapdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmmaplyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmhistodiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmhistolyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmssdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsslyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmspdiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsplyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsydiff, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_fmsylyapunov, SIGNAL(toggled(bool)), SLOT(redrawFreqMap(bool)));
    connect(ui.radioButton_smviewgrid, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.radioButton_smviewweights, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.radioButton_smviewhisto, SIGNAL(toggled(bool)), SLOT(redrawSchwModel(bool)));
    connect(ui.lineEdit_intTime, SIGNAL(textChanged(const QString&)), SLOT(intTime_changed()));
    connect(ui.lineEdit_timeStep,SIGNAL(textChanged(const QString&)), SLOT(intTime_changed()));
    connect(ui.lineEdit_fms, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmp, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmy, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.lineEdit_fmr, SIGNAL(textChanged(const QString&)), SLOT(fmparams_changed()));
    connect(ui.comboBox_smICGenerator, SIGNAL(currentIndexChanged(int)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgNumOrbits, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgSkewE, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgSkewL, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgBeta, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgBetaAxi, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_icgKappaAxi, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumSamplingPoints, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smDensityComponents, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.select_smType, SIGNAL(currentIndexChanged(int)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumLines, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumAngCoefs, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumRadCoefs, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumVerCoefs, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smAlpha, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smNumShells, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smInnerShellMass, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smOuterShellMass, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smConstraintPenaltyLin, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smConstraintPenaltyQuad, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smRegularization, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smChaoticPenalty, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.checkBox_smBeta, SIGNAL(toggled(bool)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smBetaIn, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smBetaOut, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_smMaxOrbitWeight, SIGNAL(textChanged(const QString&)), SLOT(smparams_changed()));
    connect(ui.lineEdit_MinFreqDiff, SIGNAL(textChanged(const QString&)), SLOT(ccparams_changed()));
    connect(ui.lineEdit_MinLambda, SIGNAL(textChanged(const QString&)), SLOT(ccparams_changed()));
    // pressing Return will start orbit integration if any of the following controls is under focus
    ui.lineEdit_x->installEventFilter(this);
    ui.lineEdit_y->installEventFilter(this);
    ui.lineEdit_z->installEventFilter(this);
    ui.lineEdit_vx->installEventFilter(this);
    ui.lineEdit_vy->installEventFilter(this);
    ui.lineEdit_vz->installEventFilter(this);
    ui.lineEdit_q->installEventFilter(this);
    ui.lineEdit_p->installEventFilter(this);
    ui.lineEdit_mass->installEventFilter(this);
    ui.lineEdit_scalerad->installEventFilter(this);
    ui.lineEdit_scalerad2->installEventFilter(this);
    ui.lineEdit_Mbh->installEventFilter(this);
    ui.lineEdit_Omega->installEventFilter(this);
    ui.lineEdit_intTime->installEventFilter(this);
    ui.lineEdit_timeStep->installEventFilter(this);
    ui.comboBox_Lyapunov->installEventFilter(this);
    ui.lineEdit_tol_abs->installEventFilter(this);
    ui.lineEdit_tol_rel->installEventFilter(this);
    ui.lineEdit_tol_ias->installEventFilter(this);
    ui.lineEdit_timestep_factor->installEventFilter(this);
    ui.lineEdit_relax_rate->installEventFilter(this);
    ui.radioButton_2d->installEventFilter(this);
    ui.radioButton_3d->installEventFilter(this);
    ui.radioButton_2dorbit->installEventFilter(this);
    ui.radioButton_3dline->installEventFilter(this);
    ui.radioButton_3dmesh->installEventFilter(this);
    ui.ButtonClearPS->installEventFilter(this);
    ui.ButtonRandom->installEventFilter(this);
    ui.tab_f->installEventFilter(this);
    ui.tab_p->installEventFilter(this);
    ui.tab_o->installEventFilter(this);
    ui.tabWidget->installEventFilter(this);
    PlotOrbit->installEventFilter(this);
    PlotPoincare->installEventFilter(this);
    PlotFreq->installEventFilter(this);
    PlotFreqMap->installEventFilter(this);

    // load and apply settings
    loadApplySettings();
}

CSmileGUI::~CSmileGUI()
{
    if(ui.checkBox_saveSettings->isChecked())
        core->saveSettings();
    delete core;
}

bool CSmileGUI::eventFilter(QObject* target, QEvent* Event)
{
    if(Event->type() == QEvent::KeyPress) 
    { 
        QKeyEvent* keyEvent = static_cast<QKeyEvent*>(Event);
        if((keyEvent->key() == Qt::Key_Return) || (keyEvent->key() == Qt::Key_Enter))
        {
            on_ButtonStart_clicked();
            return true;
        }
    }
    return QMainWindow::eventFilter(target, Event);
}

void CSmileGUI::timerEventCore()
{
    if(core->orbitlib==NULL) return;  // shouldn't happen
    // update both progress bars; only the one for the currently active operation is visible
    ui.progressBarFM->setValue(static_cast<int>(core->orbitlib->numComplete()));
    ui.progressBarSM->setValue(static_cast<int>(core->orbitlib->numComplete()));
    // also repaint frequency map if it is visible
    if(ui.tabWidget->currentWidget()==ui.tab_fm)
        redrawFreqMap(true);
}

void CSmileGUI::timerEvent(QTimerEvent * Event)
{
    if(Event->timerId()==myTimerOrbit && core->isOrbitIntRunning())
    {
        ui.progressBarOrbit->setValue(core->orbit->getIntTime()*100 / (core->configCore.intTimeInPeriods*core->timeUnit));
        ui.progressBarOrbit->show();
        redrawOrbit(true);
    }
}

/* ----------------------------  single orbit integration --------- */
void CSmileGUI::on_ButtonStart_clicked()
{
    if(core->isOrbitIntRunning()) { 
        core->orbit->halt();    // stop current computation
        return; 
    }
    if(core->timeUnit<=0 || core->configCore.intTimeInPeriods<=0 || core->configOrbit.samplingInterval<=0) 
    {
        error("Total energy is outside allowed range");
        return;  // error
    }
    ui.groupBox_N_dim->setEnabled(false);
    ui.groupBox_potential->setEnabled(false);
    ui.ButtonStart->setText("Stop!");
    ui.ButtonRandom->setEnabled(false);
    core->startOrbit();
    timeOI.start();
    myTimerOrbit = startTimer(2000);  // to refresh orbit plot
}

void CSmileGUI::CalcFinished()    // called from core->coreOrbitFinished()
{
    if(!core->isOrbitLibRunning() && !core->isExportNbodyRunning() && !core->isSchwOptRunning()) {
        ui.groupBox_N_dim->setEnabled(true);
        ui.groupBox_potential->setEnabled(true);
    }
    ui.ButtonStart->setText("Start");
    ui.ButtonRandom->setEnabled(true);
    timeOIelapsed=timeOI.elapsed();  // store to re-use in subsequent "refresh spectrum" events
    killTimer(myTimerOrbit);
    ui.progressBarOrbit->hide();
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    if(trj->FncType()!=COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS)
        return;  // shouldn't happen

    // Poincare section display (if it was used)
    CPoincareInformation* tps=NULL;
    for(size_t rf=0; rf<core->orbit->getInfoNum(); rf++)
        if(core->orbit->getRuntimeFnc(rf)->FncType() == CBasicOrbitRuntimeFnc::FT_POINCARE)
            tps = static_cast<CPoincareInformation*>(static_cast<const COrbitRuntimePoincare* >(core->orbit->getRuntimeFnc(rf))->getData());
    if(tps!=NULL && tps->size()>0)
    {
        addPoincareCurve(tps);
        PlotPoincare->replot();
        delete tps;
    }

#ifdef HAVE_QWT3D
#ifdef USE3DFACETS
    if(core->configOrbit.N_dim==3)
    {
        // calc max distance between adjacent points (for triangulation purposes)
        maxdist=trj->getMaxDist() * 1.01;
        ui.radioButton_3dmesh_full->setChecked(true);
        ui.lineEdit_3dmesh_maxdist->setText(QString::number(maxdist, 'g', 3));
        on_pushButton_refresh3dmesh_clicked();
    }
#endif
#endif

    // perform frequency graphs and orbit classification (also set label text and redraw orbit)
    if(ui.spinBox_intervalCount->value()!=1)
        ui.spinBox_intervalCount->setValue(1);  // refreshes frequency graph and calls for orbit classification
    else
        on_spinBox_intervalCount_valueChanged(1);
    // lyapunov
    size_t nActual=trj->getTrajSize();
    if(nActual<=100 || core->orbit->getConfig()->calcLyapunovMethod==0)
    {
        CurveLyapunov->SETDATA(NULL, NULL, 0);
        CurveDeviation->SETDATA(NULL, NULL, 0);
        PlotLyapunov->replot();
    }
    else
    {
        vectord timesteps(nActual);
        vectord lambdatmp(nActual);
        double devvecmin=1e4;
        for(size_t s=100; s<nActual; s++) 
        {
            timesteps[s] = s*trj->getTimeStep()/core->orbit->getTimeUnit();
            double lnw=trj->getLnDevVec(s);
            lambdatmp[s] = std::max<double>(lnw/timesteps[s], 1e-4);
            devvecmin = std::min<double>(lnw - log(timesteps[s]), devvecmin);
        }
        double devvecadd=(devvecmin<0.1?0.1-devvecmin:0);
        CurveLyapunov->SETDATA(&(timesteps.at(100)), &(lambdatmp.at(100)), static_cast<int>(nActual)-100);
        for(size_t s=100; s<nActual; s++) 
            lambdatmp[s] = trj->getLnDevVec(s) - log(timesteps[s]) + devvecadd;
        CurveDeviation->SETDATA(&(timesteps.at(100)), &(lambdatmp.at(100)), static_cast<int>(nActual)-100);
        //for(int i=0; i<PlotLyapunov->axisCnt; i++) PlotLyapunov->setAxisAutoScale(i); 
        PlotLyapunov->setAxisScale(QwtPlot::yLeft, 1e-4, 10);
        PlotLyapunov->setAxisScale(QwtPlot::yRight, 0.1, 1e4);
        PlotLyapunov->setAxisScale(QwtPlot::xBottom, 1, timesteps.back()*1.05);
        PlotLyapunov->replot();
    }
}

/* ----------------------------  orbit rendering --------- */
void CSmileGUI::redrawOrbit(bool show)
{
    if(!show || !core->orbit || !core->orbit->getIntTime()) return;

    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    if(trj->FncType()!=COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS) return;  // shoulnd't occur
    const CPotential* potential = core->orbit->getPotential();
    size_t nActual=trj->getTrajSize();
    size_t intNum = ui.spinBox_IntervalNum->value()-1;
    size_t intCount = ui.spinBox_intervalCount->value();
    size_t start=nActual * intNum / intCount;
    size_t count=nActual / intCount; 
    if(start+count>nActual) count=nActual-start;
    // default visibility of controls
    bool groupBox_2dplane=false;
    bool groupBox_3dmesh=false;
    bool plotOrbit=false;
    bool plotOrbitParam=false;
#ifdef HAVE_QWT3D
    bool plot3d=false;
#endif
    if(ui.radioButton_2dorbit->isChecked())
    {   // draw 2d orbit projection
        bool meridionalPlane = ui.radioButton_2d_xz->isChecked() &&
            (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC;
        groupBox_2dplane=true;
        vectord hor(count), ver(count);
        size_t indh=0, indv=0;
        if(core->configOrbit.N_dim==2) {
            indh=0; indv=1;
        } else {
            if(ui.radioButton_2d_yz->isChecked()) { indh=1; indv=2; }
            if(ui.radioButton_2d_xz->isChecked()) { indh=0; indv=2; }
            if(ui.radioButton_2d_xy->isChecked()) { indh=0; indv=1; }
        }
        if(meridionalPlane) {  // display orbit in meridional plane
            for(size_t i=0; i<count; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i+start);
                hor[i]=sqrt(pow_2(pt.Pos[0])+pow_2(pt.Pos[1]));
                ver[i]=pt.Pos[2];
            }
        } else {
            for(size_t i=0; i<count; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i+start);
                hor[i]=pt.Pos[indh];
                ver[i]=pt.Pos[indv];
            }
        }
        if(count>0)
            CurveOrbit->SETDATA(&(hor.front()), &(ver.front()), static_cast<int>(count));
        else
            CurveOrbit->SETDATA(NULL, NULL, 0);
        // equipotential surface
        double eqh[NUM_POINTS_PLOT_EQUIPOTENTIAL*4], eqv[NUM_POINTS_PLOT_EQUIPOTENTIAL*4];
        double E = core->initCondE;
        size_t num_points_plot_equipotential = 
            (potential->PotentialType()==CDensity::PT_NB) ? NUM_POINTS_PLOT_EQUIPOTENTIAL/5 : NUM_POINTS_PLOT_EQUIPOTENTIAL; 
        CDensity::SYMMETRYTYPE sym=potential->symmetry();
        for(size_t i=0; i<num_points_plot_equipotential; i++)
        {
            double angle=(i+0.5)*M_PI/2/num_points_plot_equipotential;
            double scrcoords[4][3] = { // four directions(quadrants) and three coords: horizontal, vertical and unused
                { cos(angle), sin(angle), 0}, 
                {-cos(angle), sin(angle), 0}, 
                {-cos(angle),-sin(angle), 0}, 
                { cos(angle),-sin(angle), 0} 
            };
            int coordindex[3]={0,0,0};  // X_k = scrcoords[quadrant][coordindex[k]];
            if(ui.radioButton_2d_xy->isChecked()) { coordindex[0]=0; coordindex[1]=1; coordindex[2]=2; }
            if(ui.radioButton_2d_yz->isChecked()) { coordindex[0]=2; coordindex[1]=0; coordindex[2]=1; }
            if(ui.radioButton_2d_xz->isChecked()) { coordindex[0]=0; coordindex[1]=2; coordindex[2]=1; }
            double r0=findIntersection(potential, E, scrcoords[0][coordindex[0]], scrcoords[0][coordindex[1]], scrcoords[0][coordindex[2]], core->configOrbit.Omega);
            double r1=r0, r2=r0, r3=r0;
            if((sym & CDensity::ST_TRIAXIAL) == 0)  // no symmetry w.r.t. changing sign of any coordinate
            {
                r1=findIntersection(potential, E, scrcoords[1][coordindex[0]], scrcoords[1][coordindex[1]], scrcoords[1][coordindex[2]], core->configOrbit.Omega);
                if((sym & CDensity::ST_REFLECTION) == 0)  // no symmetry w.r.t. reflection of all three coords
                {
                    r2=findIntersection(potential, E, scrcoords[2][coordindex[0]], scrcoords[2][coordindex[1]], scrcoords[2][coordindex[2]], core->configOrbit.Omega);
                    r3=findIntersection(potential, E, scrcoords[3][coordindex[0]], scrcoords[3][coordindex[1]], scrcoords[3][coordindex[2]], core->configOrbit.Omega);
                }
                else {r2=r0; r3=r1; }
            }
            if(r0<0) r0=gsl_nan();
            if(r1<0) r1=gsl_nan();
            if(r2<0) r2=gsl_nan();
            if(r3<0) r3=gsl_nan();
            eqh[i]=scrcoords[0][0]*r0; 
            eqv[i]=scrcoords[0][1]*r0;
            eqh[num_points_plot_equipotential*2-1-i]=scrcoords[1][0]*r1;
            eqv[num_points_plot_equipotential*2-1-i]=scrcoords[1][1]*r1;
            eqh[num_points_plot_equipotential*2+i]  =scrcoords[2][0]*r2;
            eqv[num_points_plot_equipotential*2+i]  =scrcoords[2][1]*r2;
            eqh[num_points_plot_equipotential*4-1-i]=scrcoords[3][0]*r3;
            eqv[num_points_plot_equipotential*4-1-i]=scrcoords[3][1]*r3;
        }
        // scan arrays and skip nan values
        double eqh1[NUM_POINTS_PLOT_EQUIPOTENTIAL*4], eqv1[NUM_POINTS_PLOT_EQUIPOTENTIAL*4];
        size_t num_pt=0;
        for(size_t i=0; i<4*num_points_plot_equipotential; i++) 
            if(gsl_finite(eqh[i])) { 
                eqh1[num_pt]=eqh[i];
                eqv1[num_pt]=eqv[i];
                num_pt++;
            }
        CurveEquipotential->SETDATA(eqh1, eqv1, num_pt);
        for(int i=0; i<PlotOrbit->axisCnt; i++) PlotOrbit->setAxisAutoScale(i); 
        PlotOrbit->replot();
        ZoomerOrbit->setZoomBase(false);
        plotOrbit=true;
    }
    else if(ui.radioButton_3dline->isChecked() || ui.radioButton_3dmesh->isChecked())
    {   // draw 3d orbit
#ifdef HAVE_QWT3D 
        if(ui.radioButton_3dline->isChecked())
        {
            Qwt3D::Cell trjcell;
            trjpos.clear();
            trjcells.clear();
            trjcell.clear();
            for(size_t i=0; i<count; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i+start);
                trjpos.push_back(Qwt3D::Triple(pt.Pos[0], pt.Pos[1], pt.Pos[2]));
                trjcell.push_back(static_cast<unsigned int>(i));
            }
            trjcells.push_back(trjcell);
            // draw trajectory
            Plot3d->loadFromData(trjpos, trjcells);
            Plot3d->setPlotStyle(Qwt3D::WIREFRAME);
        }
#ifdef USE3DFACETS
        if(ui.radioButton_3dmesh->isChecked())
        {
            // draw mesh
            if(TriangThr->restart || TriangThr->trjpos.size()==0 || TriangThr->facets.size()==0)
                Plot3d->loadFromData(Qwt3D::TripleField(), Qwt3D::CellField()); // triangulation is in process, so do not display anything
            else
                Plot3d->loadFromData(TriangThr->trjpos, TriangThr->facets);
            Plot3d->setPlotStyle(Qwt3D::FILLEDMESH);
            groupBox_3dmesh=true;
        }
#endif
        Plot3d->updateData();
        Plot3d->updateGL();
        plot3d=true;
#endif
    }
    else if(ui.radioButton_opLPeri->isChecked() || ui.radioButton_opPericenter->isChecked() 
        || ui.radioButton_opL->isChecked())
    {   // orbit params distribution
        vectord data;
        const COrbitRuntimePericenter* tpc=NULL;
        for(size_t rf=0; rf<core->orbit->getInfoNum(); rf++)
            if(core->orbit->getRuntimeFnc(rf)->FncType() == CBasicOrbitRuntimeFnc::FT_PERICENTER)
                tpc = static_cast<const COrbitRuntimePericenter* >(core->orbit->getRuntimeFnc(rf));
        if(ui.radioButton_opLPeri->isChecked())
        {
            if(tpc!=NULL)
                for(size_t i=0; i<tpc->size(); i++)
                    data.push_back(tpc->getl2(i));
        }
        else if(ui.radioButton_opPericenter->isChecked())
        {
            if(tpc!=NULL)
                for(size_t i=0; i<tpc->size(); i++)
                    data.push_back(tpc->getr(i));
        }
        else if(ui.radioButton_opL->isChecked())
        {
            data.resize(nActual);
            for(size_t i=0; i<nActual; i++)
            {
                const CPosVelPoint<double>& pt = trj->getTraj(i);
                data[i] = pow_2(pt.Pos[1]*pt.Vel[2] - pt.Pos[2]*pt.Vel[1]) +
                        + pow_2(pt.Pos[2]*pt.Vel[0] - pt.Pos[0]*pt.Vel[2]) +
                        + pow_2(pt.Pos[0]*pt.Vel[1] - pt.Pos[1]*pt.Vel[0]);
            }
        }
        size_t nump=data.size();
        if(nump>0)
        {
            std::sort(data.begin(), data.end());
            vectord frac(nump);
            for(size_t i=0; i<nump; i++)
                frac[i]=i*1.0/nump;
            CurveOrbitParam->SETDATA(&(frac.front()), &(data.front()), static_cast<int>(nump));
        }
        else CurveOrbitParam->SETDATA(NULL, NULL, 0);
        for(int i=0; i<PlotOrbitParam->axisCnt; i++) PlotOrbitParam->setAxisAutoScale(i); 
        //PlotOrbitParam->replot();
        ZoomerOrbitParam->setZoomBase(true);
        plotOrbitParam=true;
    } 
    else if(ui.radioButton_opEnergy->isChecked())
    {
        if(trj->enertime.empty())
            CurveOrbit->SETDATA(NULL, NULL, 0);
        else {
            size_t count=trj->enertime.size();
            double tlast=trj->enertime.at(count-1), tstart=tlast*intNum/intCount, tend=tlast*(intNum+1)/intCount;
            size_t istart=0; 
            while(istart<count && trj->enertime[istart]<tstart) istart++;
            size_t icount=0; 
            while(istart+icount<count && trj->enertime[istart+icount]<tend) icount++;
            if(icount>0)
                CurveOrbit->SETDATA(&(trj->enertime.at(istart)), &(trj->energy.at(istart)), static_cast<int>(icount));
            else
                CurveOrbit->SETDATA(NULL, NULL, 0);
        }
        CurveEquipotential->SETDATA(NULL, NULL, 0);
        for(int i=0; i<PlotOrbit->axisCnt; i++) PlotOrbit->setAxisAutoScale(i); 
        PlotOrbit->replot();
        ZoomerOrbit->setZoomBase(false);
        plotOrbit=true;
    }
    else if(ui.radioButton_opAngMom->isChecked())
    {
        vectord hor(count), ver(count);
        double timeStep=trj->getTimeStep();
        for(size_t i=0; i<count; i++)
        {
            hor[i]=(i+start)*timeStep;
            const CPosVelPoint<double>& pt = trj->getTraj(i+start);
            ver[i] = sqrt(pow_2(pt.Pos[1]*pt.Vel[2] - pt.Pos[2]*pt.Vel[1]) +
                        + pow_2(pt.Pos[2]*pt.Vel[0] - pt.Pos[0]*pt.Vel[2]) +
                        + pow_2(pt.Pos[0]*pt.Vel[1] - pt.Pos[1]*pt.Vel[0]) );
        }
        if(count>0)
            CurveOrbit->SETDATA(&(hor.front()), &(ver.front()), static_cast<int>(count));
        else
            CurveOrbit->SETDATA(NULL, NULL, 0);
        CurveEquipotential->SETDATA(NULL, NULL, 0);
        for(int i=0; i<PlotOrbit->axisCnt; i++) PlotOrbit->setAxisAutoScale(i); 
        PlotOrbit->replot();
        ZoomerOrbit->setZoomBase(false);
        plotOrbit=true;
    }
    else my_error(FUNCNAME, "Don't know what to display (program bug?)");  // shouldn't happen as long as the above alternatives are exhaustive
    ui.groupBox_2dplane->setVisible(groupBox_2dplane);
    ui.groupBox_3dmesh->setVisible(groupBox_3dmesh);
    if(plotOrbit) PlotOrbit->show(); else PlotOrbit->hide();
    if(plotOrbitParam) PlotOrbitParam->show(); else PlotOrbitParam->hide();
#ifdef HAVE_QWT3D
    if(plot3d) Plot3d->show(); else Plot3d->hide();
#endif
}

void CSmileGUI::on_pushButton_refresh3dmesh_clicked()
{
#ifdef USE3DFACETS
    bool ok;
    double md = ui.lineEdit_3dmesh_maxdist->text().toDouble(&ok); 
    if(ok) maxdist=md;
    int meshstyle=0;
    if(ui.radioButton_3dmesh_x->isChecked()) meshstyle=1; else
    if(ui.radioButton_3dmesh_y->isChecked()) meshstyle=2; else
    if(ui.radioButton_3dmesh_z->isChecked()) meshstyle=3; 
    std::vector<CPosPoint<double> > trajp;
    assert(core->orbit);
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    assert(trj->FncType()==COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS);
    size_t nActual=trj->getTrajSize();
    trajp.reserve(nActual);
    for(size_t i=0; i<nActual; i++)
        trajp.push_back(trj->getTraj(i));
    TriangThr->startTriangulation(trajp, maxdist, meshstyle);
#endif
}

/* ----------------------------  spectrum rendering --------- */
void CSmileGUI::on_spinBox_intervalCount_valueChanged(int count)
{
    if(InternalRecalc) return;
    bool doNotRefresh = (ui.spinBox_IntervalNum->value() > count);
    ui.spinBox_IntervalNum->setMaximum(count);
    if(doNotRefresh) return;   // setting maximum more than current value led to change of value of IntervalNum and hence the refresh function was already called
    if(ui.spinBox_IntervalNum->value()!=1)
        ui.spinBox_IntervalNum->setValue(1);  // refreshes frequency graph and calls for orbit classification
    else
        refreshOrbitAndSpectrum();
}

void CSmileGUI::on_spinBox_IntervalNum_valueChanged(int)
{
    refreshOrbitAndSpectrum();
}

void CSmileGUI::refreshOrbitAndSpectrum() 
{
    const COrbitRuntimeTrajectory* trj = static_cast<const COrbitRuntimeTrajectory*>(core->orbit->getRuntimeFnc(0));   // 0th runtime function is the orbit trajectory analyzer
    if(trj->FncType()!=COrbitRuntimeTrajectory::FT_TRAJ_ANALYSIS) return;
    size_t nActual=trj->getTrajSize();

    size_t intNum = ui.spinBox_IntervalNum->value() - 1;
    size_t intCount = ui.spinBox_intervalCount->value();
    if (intNum>=intCount) return;
    size_t start=nActual * intNum / intCount;
    size_t count=nActual / intCount; 
    if(start+count>nActual) count=nActual-start;

    timeOI.start();
    // obtain spectrum for given interval
    CSpectrum sl[N_DIM];
    std::vector<double> fa[N_DIM];
    COrbitInformation<double>* infoOrbit = trj->analyzeOrbit<double>(start, count, sl, fa);
    int nPointsSpectrum = static_cast<int>(fa[0].size()); 
    if(nPointsSpectrum<=0) return;  // ??? shouldn't happen but who knows..
    vectord freq(nPointsSpectrum);
    for(int i=0; i<nPointsSpectrum; i++)
        freq[i]=i/core->orbit->getIntTime()*core->orbit->getTimeUnit() * intCount;
    CurveFreqX->SETDATA(&freq.front(), &fa[0].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F*ui.lineEdit_timeStep->text().toDouble(),1));
    CurveFreqY->SETDATA(&freq.front(), &fa[1].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F*ui.lineEdit_timeStep->text().toDouble(),1));
    if(core->configOrbit.N_dim==3)
        CurveFreqZ->SETDATA(&freq.front(), &fa[2].front(), nPointsSpectrum*std::min<double>(2*PLOT_FREQ_MAX_F*ui.lineEdit_timeStep->text().toDouble(),1));
    else
        CurveFreqZ->SETDATA(NULL, NULL, 0);
    double sf[MAX_SPECTRAL_LINES*N_DIM], sa[MAX_SPECTRAL_LINES*N_DIM];
    double sfp[MAX_SPECTRAL_LINES*N_DIM], sap[MAX_SPECTRAL_LINES*N_DIM];
    int si=0, sip=0;
    for(unsigned int d=0; d<core->configOrbit.N_dim; d++)
      for(size_t s=0; s<sl[d].size() && s<MAX_SPECTRAL_LINES*N_DIM; s++) 
          if(sl[d][s].freq<PLOT_FREQ_MAX_F)
          {
            if(sl[d][s].method>0)
            {
                sfp[sip]=sl[d][s].freq;
                sap[sip]=sl[d][s].ampl;
                sip++;
            }
            else
            {
                sf[si]=sl[d][s].freq;
                sa[si]=sl[d][s].ampl;
                si++;
            }
          }
    CurveFreqPeaks->SETDATA(sf, sa, si);
    CurveFreqPeaksP->SETDATA(sfp, sap, sip);
    for(int i=0; i<PlotFreq->axisCnt; i++) PlotFreq->setAxisAutoScale(i); 
    PlotFreq->replot();
    ZoomerFreq->setZoomBase(true);

    // set label text
    std::string strInfo = infoOrbit->toString();   // obtain information from orbit analysis - not for entire trajectory but for a segment of it.
    delete infoOrbit;
    for(size_t rf=1; rf<core->orbit->getInfoNum(); rf++)    // obtain the rest (starting from 1 since 0th runtime function is the orbit analysis which was called manually above)
    {
        const CBasicInformation* info=core->orbit->getNewInfo(rf);
        if(info!=NULL) strInfo += info->toString();
        delete info;
    }
    strInfo+=core->orbit->toString(false);
    info(QString::fromStdString(strInfo)
        + "Time elapsed="+QString::number((timeOIelapsed+timeOI.elapsed())*0.001)+" s");
    redrawOrbit(true);  // redraw trajectory for given interval start--count
}

void CSmileGUI::on_checkBox_freqLines_stateChanged(int)
{
    if(ui.checkBox_freqLines->isChecked())
    {
        CurveFreqPeaksP->setStyle(QwtPlotCurve::Sticks);
        CurveFreqPeaks->setStyle(QwtPlotCurve::Sticks);
    }
    else
    {
        CurveFreqPeaksP->setStyle(QwtPlotCurve::NoCurve);
        CurveFreqPeaks->setStyle(QwtPlotCurve::NoCurve);
    }
    PlotFreq->replot(); 
}

/* ----------------------------  potential rendering --------- */
void CSmileGUI::redrawPotentialB(bool checked)
{
    if(checked) redrawPotential();
}

/// \cond INTERNAL_DOCS
struct CSurfaceDensityIntParam {
    const CDensity* density;
    double coord;
    int direction;
};

double surfaceDensityInt(double vscaled, void* param)
{
    double coord=static_cast<CSurfaceDensityIntParam*>(param)->coord;
    int dir=static_cast<CSurfaceDensityIntParam*>(param)->direction;
    double x=0,y=0,z=0,v=vscaled/(1-vscaled);
    switch(dir) {
        case 0: x=coord; z=v; break;
        case 1: y=coord; z=v; break;
        case 2: z=coord; x=v; break;
        default:;
    }
    double value=static_cast<CSurfaceDensityIntParam*>(param)->density->Rho(x,y,z) / pow_2(1-vscaled);
    if(dir==2) value*=v*2*M_PI;
    return value;
}
/// \endcond

void CSmileGUI::redrawPotential()
{
    const CPotential* potential=core->potential;
    if(potential->PotentialType()==CPotential::PT_COMPOSITE && ui.comboBox_ptComponents->currentIndex()>0) {
        int comp=ui.comboBox_ptComponents->currentIndex()-1;
        const CPotentialComposite* potcomp=static_cast<const CPotentialComposite*>(potential);
        if(comp<static_cast<int>(potcomp->size()))
            potential = potcomp->component(comp);
    }
    bool logscale=ui.checkBox_ptLogScale->isChecked();
    int dir=0;
    if(ui.radioButton_axis_x->isChecked()) dir=0; else
    if(ui.radioButton_axis_y->isChecked()) dir=1; else
    if(ui.radioButton_axis_z->isChecked()) dir=2;
    double direction[N_DIM]={0,0,0};
    direction[dir]=1.0;
    int num_points_plot = NUM_POINTS_PLOT_EQUIPOTENTIAL; 
    double cmax=findIntersection(core->potential, core->initCondE, direction[0], direction[1], direction[2], core->configOrbit.Omega);
    if(cmax<0)  // no maximum found -- set to initial radius
        cmax=sqrt(pow_2(core->initCond.Pos[0])+pow_2(core->initCond.Pos[1])+pow_2(core->initCond.Pos[2]));
    double cmin=logscale ? cmax*1e-3 : cmax/num_points_plot;
    if((potential->symmetry() & CDensity::ST_REFLECTION)==0 && !logscale) 
    {
        cmin=findIntersection(potential, core->initCondE, -direction[0], -direction[1], -direction[2], core->configOrbit.Omega);
        if(cmin<0) cmin=-cmax; else cmin*= -1;
    }
    vectord hor(num_points_plot+1), ver(num_points_plot+1), ver1, ver2;
    for(int i=0; i<=num_points_plot; i++)
    {
        if(logscale)
            hor[i]=cmin*exp(i*log(cmax/cmin)/num_points_plot);
        else
            hor[i]=cmin+i*(cmax-cmin)/num_points_plot;
    }
    CurvePotential1->SETDATA(NULL, NULL, 0);
    CurvePotential2->SETDATA(NULL, NULL, 0);
    if(ui.radioButton_ptPotential->isChecked()) {
        // plot potential as a function of coordinate
        for(int i=0; i<=num_points_plot; i++)
            ver[i]=potential->Phi(hor[i]*direction[0], hor[i]*direction[1], hor[i]*direction[2]) 
                - (direction[2]==0 ? pow_2(core->configOrbit.Omega*hor[i])*0.5 : 0);
    } else if(ui.radioButton_ptDensity->isChecked()) {
        // plot density as a function of coordinate
        for(int i=0; i<=num_points_plot; i++) {
            ver[i]=potential->Rho(hor[i]*direction[0], hor[i]*direction[1], hor[i]*direction[2]);
            if(logscale && ver[i]<=0) ver[i]=1e-15; // lower bound to avoid -inf
        }
    } else if(ui.radioButton_ptRotCurve->isChecked()) {
        // plot circular velocity as a function of coordinate
        for(int i=0; i<=num_points_plot; i++) {
            double coord[N_DIM]={hor[i]*direction[0], hor[i]*direction[1], hor[i]*direction[2]};
            double force[N_DIM+1]={0,0,0};
            potential->Force(coord, 0, force);
            ver[i]=sqrt(-hor[i]*force[dir]);
        }
    } else if(ui.radioButton_ptFrequencies->isChecked()) {
        ver1.assign(num_points_plot+1,0);
        ver2.assign(num_points_plot+1,0);
        for(int i=0; i<=num_points_plot; i++) {
            double coord[N_DIM]={hor[i]*direction[0], hor[i]*direction[1], hor[i]*direction[2]};
            double force[N_DIM+1]={0,0,0}, forceDeriv[N_DIM*2+1]={0,0,0,0,0,0,0};
            potential->Force(coord, 0, force, forceDeriv);
            // determine circular, epicyclic and vertical frequencies in the disk plane
            double Omega = -force[dir]/hor[i];
            double kappa = 3*Omega - forceDeriv[dir];
            double nu    = -forceDeriv[2];
            ver[i] = sqrt(std::max<double>(Omega,0));
            ver1[i]= sqrt(std::max<double>(kappa,0));
            ver2[i]= sqrt(std::max<double>(nu,0));
        }
        CurvePotential1->SETDATA(&(hor.front()), &(ver1.front()), static_cast<int>(num_points_plot+1));
        if(dir!=2) CurvePotential2->SETDATA(&(hor.front()), &(ver2.front()), static_cast<int>(num_points_plot+1));
    } else if(ui.radioButton_ptSurfaceDensity->isChecked()) {
        const CDensity* density = potential;
        if(dir==2)   // create temporary rotationally-symmetric density for computing mass in a slice (z,z+dz) integrated over R and phi
            density=new CPotentialAxisymmetrized(NULL, potential);  // (doesn't incur penalty if the original density is already axisymmetric)
#pragma omp parallel for
        for(int i=0; i<=num_points_plot; i++) {
            gsl_function Fnc;
            CSurfaceDensityIntParam param;
            Fnc.function=&surfaceDensityInt;
            Fnc.params=&param;
            param.density=density;
            param.direction=dir;
            param.coord=hor[i];
            double result, error;
            size_t neval;
            gsl_integration_qng(&Fnc, 0, 1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
            ver[i]=result;
            if(logscale && ver[i]<=0) ver[i]=1e-15; // lower bound to avoid -inf
        }
        if(dir==2) delete density;
    } else {
        my_error(FUNCNAME, "Undefined selection of function to plot");
        return;
    }
    if(logscale) {
        for(int i=0; i<=num_points_plot; i++)
            ver[i]=fabs(ver[i]);
        CurvePotential->SETDATA(&(hor.front()), &(ver.front()), static_cast<int>(num_points_plot+1));
        PlotPotential->setAxisScaleEngine(QwtPlot::xBottom, new QwtLog10ScaleEngine);
        PlotPotential->setAxisScaleEngine(QwtPlot::yLeft  , new QwtLog10ScaleEngine);
    } else {
        CurvePotential->SETDATA(&(hor.front()), &(ver.front()), static_cast<int>(num_points_plot+1));
        PlotPotential->setAxisScaleEngine(QwtPlot::xBottom, new QwtLinearScaleEngine);
        PlotPotential->setAxisScaleEngine(QwtPlot::yLeft  , new QwtLinearScaleEngine);
    }
    //setAxisOptions(QwtPlot::yLeft,   logscale ? QwtAutoScale::Logarithmic : QwtAutoScale::Linear);
    //setAxisOptions(QwtPlot::xBottom, logscale ? QwtAutoScale::Logarithmic : QwtAutoScale::Linear);
    for(int i=0; i<PlotOrbit->axisCnt; i++) PlotPotential->setAxisAutoScale(i); 
    PlotPotential->replot();
    ZoomerPotential->setZoomBase(false);
}

/* ----------------------------  edit parameters --------- */
bool CSmileGUI::loadApplySettings(QString fileName)
{
    bool result=core->loadSettings(fileName.toStdString());
    core->initPotential();
    InternalRecalc=true;
    ui.lineEdit_Omega->setText(QString::number(core->configOrbit.Omega));
    ui.selectIntegratorType->setCurrentIndex(ui.selectIntegratorType->findText(getIntegratorNameByType(core->configOrbit.integratorType)));
    ui.lineEdit_tol_abs->setText(QString::number(core->configOrbit.accuracyAbs));
    ui.lineEdit_tol_rel->setText(QString::number(core->configOrbit.accuracyRel));
    ui.lineEdit_tol_ias->setText(QString::number(core->configOrbit.accuracyIAS15));
    ui.lineEdit_timestep_factor->setText(QString::number(core->configOrbit.accuracyFixedTimestep));
    ui.lineEdit_relax_rate->setText(QString::number(core->configOrbit.relaxRate));
    ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    if(core->useICe) {
        ui.lineEdit_E->setEnabled(true);
        ui.groupBox_ICx->setEnabled(false);
    } else {
        ui.lineEdit_E->setEnabled(false);
        ui.groupBox_ICx->setEnabled(true);
    }
    ui.lineEdit_intTime->setText(QString::number(core->configCore.intTimeInPeriods));
    ui.lineEdit_timeStep->setText(QString::number(core->configOrbit.samplingInterval));
    ui.comboBox_Lyapunov->setCurrentIndex(core->configOrbit.calcLyapunovMethod);
    ui.checkBox_usePS->setChecked(core->configCore.usePS);
    if( core->configOrbit.N_dim==3 && core->configCore.usePS &&
        (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC) 
    { // set orbit display and Poincare section coords for axisymmetric system
        core->configCore.PSdisplayCoord=1;
        core->configCore.PSsectionCoord=2;
        ui.radioButton_2d_xz->setChecked(true);
    } else if(core->configOrbit.N_dim==2 && core->configCore.usePS) { // set Poincare section for 2d system
        core->configCore.PSdisplayCoord=0;
        core->configCore.PSsectionCoord=1;
    }
    ui.comboBox_PSdisplayCoord->setCurrentIndex(core->configCore.PSdisplayCoord);
    ui.comboBox_PSsectionCoord->setCurrentIndex(core->configCore.PSsectionCoord);
    ui.lineEdit_fms->setText(QString::number(core->configCore.fm_numOrbitsStationary));
    ui.lineEdit_fmp->setText(QString::number(core->configCore.fm_numOrbitsPrincipalPlane));
    ui.lineEdit_fmy->setText(QString::number(core->configCore.fm_numOrbitsYalpha));
    ui.lineEdit_fmr->setText(QString::number(core->configCore.fm_numOrbitsRandom));
    ui.lineEdit_icgNumOrbits->setText(QString::number(core->configCore.sm_numOrbits));
    int icgIndex =     // warning, assuming a given order of lines in comboboxz
        core->configCore.sm_ICgenType == CBasicICGenerator::IC_VEL_EDDINGTON ? 0 :
        core->configCore.sm_ICgenType == CBasicICGenerator::IC_VEL_JEANSSPH ? 1 :
        core->configCore.sm_ICgenType == CBasicICGenerator::IC_VEL_JEANSAXI ? 2 : 3;
    ui.comboBox_smICGenerator->setCurrentIndex(icgIndex);
    ui.lineEdit_icgSkewE->setText(QString::number(core->configCore.weightSkewFactorE));
    ui.lineEdit_icgSkewL->setText(QString::number(core->configCore.weightSkewFactorL));
    ui.lineEdit_icgBeta->setText(QString::number(core->configCore.sm_ICbeta));
    ui.lineEdit_icgBetaAxi->setText(QString::number(core->configCore.sm_ICbetaAxi));
    ui.lineEdit_icgKappaAxi->setText(QString::number(core->configCore.sm_ICkAxi));
    ui.select_smType->setCurrentIndex(ui.select_smType->findText(getSchwDataNameByType(core->configCore.sm_densDataType)));
    ui.lineEdit_smNumLines->setText(QString::number(core->configCore.sm_linesPerSegment));
    ui.lineEdit_smNumAngCoefs->setText(QString::number(core->configCore.sm_numAngularCoefs));
    ui.lineEdit_smNumRadCoefs->setText(QString::number(core->configCore.sm_numRadialCoefs));
    ui.lineEdit_smNumVerCoefs->setText(QString::number(core->configCore.sm_numVerticalCoefs));
    ui.lineEdit_smAlpha->setText(QString::number(core->configCore.sm_Alpha));
    ui.lineEdit_smNumShells->setText(QString::number(core->configCore.sm_numShells));
    ui.lineEdit_smInnerShellMass->setText(QString::number(core->configCore.sm_innerShellMass));
    ui.lineEdit_smOuterShellMass->setText(QString::number(core->configCore.sm_outerShellMass));
    ui.lineEdit_smNumSamplingPoints->setText(QString::number(core->configCore.sm_numSamplingPoints));
    QString comps;
    for(size_t ic=0; ic<core->configCore.densityComponents.size(); ic++) {
        if(ic>0) comps+=",";
        comps += QString::number(core->configCore.densityComponents[ic]);
    }
    ui.lineEdit_smDensityComponents->setText(comps);
    ui.selectEnergyLevel_sm->setMaximum(core->configCore.sm_numShells);
    ui.lineEdit_MinFreqDiff->setText(QString::number(core->configCore.chaoticMinFreqDiff));
    ui.lineEdit_MinLambda->setText(QString::number(core->configCore.chaoticMinLambda));
    ui.lineEdit_smConstraintPenaltyLin->setText(QString::number(core->configCore.sm_constraintPenaltyLin));
    ui.lineEdit_smConstraintPenaltyQuad->setText(QString::number(core->configCore.sm_constraintPenaltyQuad));
    ui.lineEdit_smRegularization->setText(QString::number(core->configCore.sm_regularization));
    ui.lineEdit_smChaoticPenalty->setText(QString::number(core->configCore.sm_chaoticPenalty));
    ui.lineEdit_smMaxOrbitWeight->setText(QString::number(core->configCore.sm_maxWeight));
    ui.lineEdit_smBetaIn->setText(QString::number(core->configCore.sm_betaIn));
    ui.lineEdit_smBetaOut->setText(QString::number(core->configCore.sm_betaOut));
    ui.checkBox_smBeta->setChecked(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaIn->setEnabled(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaOut->setEnabled(core->configCore.sm_constrainBeta);
#ifndef HAVE_CVXOPT
    ui.ButtonSchwQP->setVisible(core->configCore.sm_useBPMPD);  // disable QP button if we are not using BPMPD and the program was compiled without CVXOPT support
#endif
    if(core->useICe) 
        ui.radioButton_ICe->setChecked(true); 
    else 
        ui.radioButton_ICe->setChecked(false);
    if(core->configOrbit.N_dim==2) 
        ui.radioButton_2d->setChecked(true); 
    else 
        ui.radioButton_3d->setChecked(true);
    InternalRecalc=false;
    currentComponentIndex=0;
    refreshPotentialComponentList();
    updatePotentialSettings();
    ui.buttonInitPotential->hide();   // this button will be shown only when parameters changed
    integratorChanged();
    smparams_changed();
    ICchanged();
    return result;
}

void CSmileGUI::refreshPotentialComponentList()
{
    // refresh the components combobox in the potential plot tab
    ui.comboBox_ptComponents->clear();
    if(core->potential->PotentialType()==CPotential::PT_COMPOSITE) {
        ui.comboBox_ptComponents->setEnabled(true);
        ui.comboBox_ptComponents->insertItem(100, "all");
        const CPotentialComposite* pot=static_cast<const CPotentialComposite*>(core->potential);
        for(size_t ind=0; ind<pot->size(); ind++)
            ui.comboBox_ptComponents->insertItem(100, QString::number(ind)+(ind==1?"st":ind==2?"nd":ind==3?"rd":"th"));
        ui.comboBox_ptComponents->setCurrentIndex(1+currentComponentIndex);
    } else {
        ui.comboBox_ptComponents->setEnabled(false);
    }
    // refresh the combobox in the potential parameters panel
    InternalRecalc=true;
    size_t count=core->configPotential.size();
    ui.selectPotentialIndex->clear();
    for(size_t ind=0; ind<count; ind++)
        ui.selectPotentialIndex->insertItem(100, QString::number(ind)+(ind==1?"st":ind==2?"nd":ind==3?"rd":"th")+" component");
    ui.selectPotentialIndex->insertItem(100, "Add component");
    if(count>1) ui.selectPotentialIndex->insertItem(100, "Delete component");
    ui.selectPotentialIndex->setCurrentIndex(currentComponentIndex);
    InternalRecalc=false;
}

// called when the selected item in the drop-down list of potential components has changed. Add or delete component also leads here.
void CSmileGUI::potentialIndexChanged()
{
    if(InternalRecalc) return;
    InternalRecalc=true;    // temporary set flag so that changing the current selection does not call this function again
    size_t index=std::max<int>(ui.selectPotentialIndex->currentIndex(),0);
    bool reinitPot=false;
    if(index==core->configPotential.size()) {  // "Add new component"
        if(QMessageBox::question(this, "Composite potential", "Add another potential component?", 
            QMessageBox::Yes|QMessageBox::No) == QMessageBox::Yes) {
            core->configPotential.push_back(defaultConfigPotential);
            currentComponentIndex=index;
            updatePotentialSettings();
            reinitPot=true;
        } else {
            ui.selectPotentialIndex->setCurrentIndex(currentComponentIndex);
        }
    } else if(index==core->configPotential.size()+1 && core->configPotential.size()>1) {  // "Delete current component"
        if(QMessageBox::question(this, "Composite potential", "Delete current potential component?", 
            QMessageBox::Yes|QMessageBox::No) == QMessageBox::Yes) {
            core->configPotential.erase(core->configPotential.begin()+currentComponentIndex);
            if(currentComponentIndex>0) currentComponentIndex--;
            updatePotentialSettings();
            reinitPot=true;
        } else {
            ui.selectPotentialIndex->setCurrentIndex(currentComponentIndex);
        }
    } else if(index!=currentComponentIndex) {
        currentComponentIndex=std::min<size_t>(index,core->configPotential.size()-1);
        updatePotentialSettings();
    }
    InternalRecalc=false;
    if(reinitPot) {
        bool initNow=true;
        for(size_t i=0; i<core->configPotential.size(); i++)
            if( core->configPotential[i].PotentialType==CDensity::PT_BSE ||
                core->configPotential[i].PotentialType==CDensity::PT_BSECOMPACT ||
                core->configPotential[i].PotentialType==CDensity::PT_SPLINE || 
                core->configPotential[i].PotentialType==CDensity::PT_CYLSPLINE ||
                core->configPotential[i].PotentialType==CDensity::PT_NB)
                initNow=false;
        if(initNow)
            on_buttonInitPotential_clicked();  // do initialization for 'lightweight' potentials only
        else {
            ui.buttonInitPotential->show();    // let the user press this button manually
            refreshPotentialComponentList();
        }
    }
}

void CSmileGUI::updatePotentialSettings()
{
    InternalRecalc=true;
    ui.selectPotentialType->setCurrentIndex(ui.selectPotentialType->findText(
        getPotentialNameByType(core->configPotential[currentComponentIndex].PotentialType)));
    ui.selectDensityType  ->setCurrentIndex(ui.selectDensityType  ->findText(
        getDensityNameByType  (core->configPotential[currentComponentIndex].DensityType)));
    ui.selectSymmetryType ->setCurrentIndex(ui.selectSymmetryType ->findText(
        getSymmetryNameByType (core->configPotential[currentComponentIndex].SymmetryType)));
    ui.lineEdit_NbodyFile->setText(QString::fromStdString(core->configPotential[currentComponentIndex].NbodyFile));
    ui.lineEdit_Ncoefs_radial->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_radial));
    ui.lineEdit_Ncoefs_vertical->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_vertical));
    ui.lineEdit_Ncoefs_angular->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_angular));
    ui.lineEdit_Alpha->setText(QString::number(core->configPotential[currentComponentIndex].Alpha));
    ui.lineEdit_Rmax->setText(QString::number(core->configPotential[currentComponentIndex].Rmax));
    ui.lineEdit_eps->setText(QString::number(core->configPotential[currentComponentIndex].treecodeEps));
    ui.lineEdit_tol->setText(QString::number(core->configPotential[currentComponentIndex].treecodeTheta));
    ui.lineEdit_mass->setText(QString::number(core->configPotential[currentComponentIndex].mass));
    ui.lineEdit_q->setText(QString::number(core->configPotential[currentComponentIndex].q));
    ui.lineEdit_p->setText(QString::number(core->configPotential[currentComponentIndex].p));
    ui.lineEdit_scalerad->setText(QString::number(core->configPotential[currentComponentIndex].scalerad));
    ui.lineEdit_scalerad2->setText(QString::number(core->configPotential[currentComponentIndex].scalerad2));
    ui.lineEdit_Gamma->setText(QString::number(core->configPotential[currentComponentIndex].Gamma));
    ui.lineEdit_SersicIndex->setText(QString::number(core->configPotential[currentComponentIndex].sersicIndex));
    ui.lineEdit_Mbh->setText(QString::number(core->configPotential[0].Mbh));  // black hole mass is always in the first component
    InternalRecalc=false;
    showHidePotentialParams();
}

void CSmileGUI::showHidePotentialParams()
{
    CDensity::POTENTIALTYPE PotentialType = getPotentialTypeByName(ui.selectPotentialType->currentText().toStdString());
    CDensity::POTENTIALTYPE DensityType   = getDensityTypeByName  (ui.selectDensityType  ->currentText().toStdString());
    bool isPotExpansion =   // whether we are using any of the potential expansions which necessitate a separate density model
        PotentialType==CDensity::PT_BSE || PotentialType==CDensity::PT_BSECOMPACT || 
        PotentialType==CDensity::PT_SPLINE || PotentialType==CDensity::PT_CYLSPLINE;
    // specifies either the potential model or the underlying density model for the potential, if an expansion is used
    CDensity::POTENTIALTYPE ModelType = isPotExpansion ? DensityType : PotentialType;
    // show/hide various input elements based on the selected potential or density type
    bool showDensityType=false;
    bool showSymmetryType=false;
    bool showNcoefs_radial=false;
    bool showNcoefs_vertical=false;
    bool showNcoefs_angular=false;
    bool showAlpha=false;
    bool showRmax=false;
    bool showNbodyFile=false;
    bool showMass=false;
    bool showqp=false;
    bool showscalerad=false;
    bool showscalerad2=false;
    bool showGamma=false;
    bool showSersicIndex=false;

    if(isPotExpansion) {
        showDensityType=true;
        showNcoefs_radial=true;
        showNcoefs_angular=true;
    }
    if(PotentialType==CDensity::PT_CYLSPLINE)
        showNcoefs_vertical=true;
    if(PotentialType==CDensity::PT_BSE)
        showAlpha=true;
    if(PotentialType==CDensity::PT_BSECOMPACT)
        showRmax=true;
    if(PotentialType==CDensity::PT_NB || PotentialType==CDensity::PT_SPHERICAL)
        showNbodyFile=true;
    if(isPotExpansion && 
       (DensityType==CDensity::PT_NB  || DensityType==CDensity::PT_ELLIPSOIDAL || 
        DensityType==CDensity::PT_MGE || DensityType==CDensity::PT_COEFS) )
        showNbodyFile=true;
    if(isPotExpansion && DensityType==CDensity::PT_NB)
        showSymmetryType=true;
    if( ModelType==CDensity::PT_LOG || ModelType==CDensity::PT_HARMONIC || 
        ModelType==CDensity::PT_SCALEFREE || ModelType==CDensity::PT_SCALEFREESH || 
        ModelType==CDensity::PT_DEHNEN || 
        ModelType==CDensity::PT_MIYAMOTONAGAI || ModelType==CDensity::PT_FERRERS ||
        ModelType==CDensity::PT_PLUMMER || ModelType==CDensity::PT_ISOCHRONE ||
        ModelType==CDensity::PT_PERFECTELLIPSOID || ModelType==CDensity::PT_NFW || 
        ModelType==CDensity::PT_SERSIC || ModelType==CDensity::PT_EXPDISK)
        showMass=true;
    if( ModelType==CDensity::PT_LOG || ModelType==CDensity::PT_HARMONIC || 
        ModelType==CDensity::PT_SCALEFREE || ModelType==CDensity::PT_SCALEFREESH || 
        ModelType==CDensity::PT_DEHNEN || ModelType==CDensity::PT_FERRERS ||
        ModelType==CDensity::PT_PLUMMER || ModelType==CDensity::PT_ISOCHRONE ||
        ModelType==CDensity::PT_PERFECTELLIPSOID || ModelType==CDensity::PT_NFW || ModelType==CDensity::PT_SERSIC)
        showqp=true;
    if( ModelType==CDensity::PT_LOG || ModelType==CDensity::PT_DEHNEN || 
        ModelType==CDensity::PT_SCALEFREE || ModelType==CDensity::PT_SCALEFREESH || 
        ModelType==CDensity::PT_MIYAMOTONAGAI || ModelType==CDensity::PT_FERRERS ||
        ModelType==CDensity::PT_PLUMMER || ModelType==CDensity::PT_ISOCHRONE ||
        ModelType==CDensity::PT_PERFECTELLIPSOID || ModelType==CDensity::PT_NFW || 
        ModelType==CDensity::PT_SERSIC || ModelType==CDensity::PT_EXPDISK)
        showscalerad=true;
    if(ModelType==CDensity::PT_MIYAMOTONAGAI || ModelType==CDensity::PT_NFW || ModelType==CDensity::PT_EXPDISK) 
        showscalerad2=true;
    if(ModelType==CDensity::PT_DEHNEN || PotentialType==CDensity::PT_SCALEFREE || PotentialType==CDensity::PT_SCALEFREESH)
        showGamma=true;
    if(ModelType==CDensity::PT_SERSIC) 
        showSersicIndex=true;
    if(PotentialType==CDensity::PT_NB) {
        ui.label_eps->show();
        ui.label_tol->show();
        ui.lineEdit_eps->show();
        ui.lineEdit_tol->show();
        // disable Lyapunov exponent
        core->configOrbit.calcLyapunovMethod=0;
        ui.comboBox_Lyapunov->setCurrentIndex(0);
        ui.comboBox_Lyapunov->setEnabled(false);
        // set N_dim=3
        core->configOrbit.N_dim=N_DIM;
        on_radioButton_2d_toggled(false);
        ui.radioButton_3d->setChecked(true);
        ui.groupBox_N_dim->setEnabled(false);
    } else {
        ui.label_eps->hide();
        ui.label_tol->hide();
        ui.lineEdit_eps->hide();
        ui.lineEdit_tol->hide();
        ui.groupBox_N_dim->setEnabled(true);
        ui.comboBox_Lyapunov->setEnabled(true);
    }
    ui.lineEdit_Ncoefs_radial->setEnabled(isPotExpansion && DensityType!=CDensity::PT_COEFS);
    ui.lineEdit_Ncoefs_vertical->setEnabled(isPotExpansion && DensityType!=CDensity::PT_COEFS);
    ui.lineEdit_Ncoefs_angular->setEnabled(isPotExpansion && DensityType!=CDensity::PT_COEFS);
    ui.lineEdit_Alpha->setEnabled(isPotExpansion && DensityType!=CDensity::PT_COEFS);
    ui.lineEdit_Rmax->setEnabled(isPotExpansion && DensityType!=CDensity::PT_COEFS);
    if(PotentialType==CDensity::PT_SCALEFREESH) {
        showNcoefs_angular=true;
        DensityType=CDensity::PT_SCALEFREE;    // reset density type to scale-free (another possibility is PT_COEFS but it is not handled in GUI)
        ui.lineEdit_Ncoefs_angular->setEnabled(true);
    }
    if(showDensityType) {
        ui.selectDensityType->show();
        ui.label_densityType->show();
    } else {
        ui.selectDensityType->hide();
        ui.label_densityType->hide();
    }
    if(showSymmetryType) {
        ui.selectSymmetryType->show();
        ui.label_symmetryType->show();
    } else {
        ui.selectSymmetryType->hide();
        ui.label_symmetryType->hide();
    }
    if(showNcoefs_radial) {
        ui.lineEdit_Ncoefs_radial->show();
        ui.label_Ncoefs_radial->show();
    } else {
        ui.lineEdit_Ncoefs_radial->hide();
        ui.label_Ncoefs_radial->hide();
    }
    if(showNcoefs_vertical) {
        ui.lineEdit_Ncoefs_vertical->show();
        ui.label_Ncoefs_vertical->show();
    } else {
        ui.lineEdit_Ncoefs_vertical->hide();
        ui.label_Ncoefs_vertical->hide();
    }
    if(showNcoefs_angular) {
        ui.lineEdit_Ncoefs_angular->show();
        ui.label_Ncoefs_angular->show();
    } else {
        ui.lineEdit_Ncoefs_angular->hide();
        ui.label_Ncoefs_angular->hide();
    }
    if(showAlpha) {
        ui.lineEdit_Alpha->show();
        ui.label_Alpha->show();
    } else {
        ui.lineEdit_Alpha->hide();
        ui.label_Alpha->hide();
    }
    if(showRmax) {
        ui.lineEdit_Rmax->show();
        ui.label_Rmax->show();
    } else {
        ui.lineEdit_Rmax->hide();
        ui.label_Rmax->hide();
    }
    if(showRmax) {
        ui.lineEdit_Rmax->show();
        ui.label_Rmax->show();
    } else {
        ui.lineEdit_Rmax->hide();
        ui.label_Rmax->hide();
    }
    if(showNbodyFile) {
        ui.lineEdit_NbodyFile->show();
        ui.buttonChooseNbodyFile->show();
    } else {
        ui.lineEdit_NbodyFile->hide();
        ui.buttonChooseNbodyFile->hide();
    } 
    if(showMass) {
        ui.lineEdit_mass->show();
        ui.label_mass->show();
    } else {
        ui.lineEdit_mass->hide();
        ui.label_mass->hide();
    }
    if(showqp) {
        ui.lineEdit_q->show();
        ui.lineEdit_p->show();
        ui.label_q->show();
        ui.label_p->show();
    } else {
        ui.lineEdit_q->hide();
        ui.lineEdit_p->hide();
        ui.label_q->hide();
        ui.label_p->hide();
    } 
    if(showscalerad) {
        ui.lineEdit_scalerad->show();
        ui.label_scalerad->show();
    } else {
        ui.lineEdit_scalerad->hide();
        ui.label_scalerad->hide();
    }
    if(showscalerad2) {
        ui.lineEdit_scalerad2->show();
        ui.label_scalerad2->show();
    } else {
        ui.lineEdit_scalerad2->hide();
        ui.label_scalerad2->hide();
    } 
    if(showGamma) {
        ui.lineEdit_Gamma->show();
        ui.label_Gamma->show();
    } else {
        ui.lineEdit_Gamma->hide();
        ui.label_Gamma->hide();
    }
    if(showSersicIndex) {
        ui.lineEdit_SersicIndex->show();
        ui.label_SersicIndex->show();
    } else {
        ui.lineEdit_SersicIndex->hide();
        ui.label_SersicIndex->hide();
    }
    ui.frame_sidebar->repaint();
}

void CSmileGUI::potentialChanged()
{
    if(InternalRecalc) return;
    showHidePotentialParams();
    // check if any parameters were modified
    bool ok=true;
    CDensity::POTENTIALTYPE PotentialType = getPotentialTypeByName(ui.selectPotentialType->currentText().toStdString());
    CDensity::POTENTIALTYPE DensityType   = getDensityTypeByName  (ui.selectDensityType  ->currentText().toStdString());
    CDensity::SYMMETRYTYPE SymmetryType   = getSymmetryTypeByName (ui.selectSymmetryType ->currentText().toStdString());
    unsigned int newNcoefs_radial=ui.lineEdit_Ncoefs_radial->text().toInt(&ok);  if(!ok) return;
    unsigned int newNcoefs_vertical=ui.lineEdit_Ncoefs_vertical->text().toInt(&ok);  if(!ok) return;
    unsigned int newNcoefs_angular=ui.lineEdit_Ncoefs_angular->text().toInt(&ok);  if(!ok) return;
    double newAlpha=ui.lineEdit_Alpha->text().toDouble(&ok);  if (!ok) return;
    double newRmax=ui.lineEdit_Rmax->text().toDouble(&ok);  if (!ok) return;
    double newtreecodeEps=ui.lineEdit_eps->text().toDouble(&ok); if(!ok) return;
    double newtreecodeTheta=ui.lineEdit_tol->text().toDouble(&ok); if(!ok) return;
    double newmass=ui.lineEdit_mass->text().toDouble(&ok);  if(!ok) return;
    double newq=(ui.lineEdit_q->text().toDouble(&ok));  if(!ok || (newq==0)) return;
    double newp=(ui.lineEdit_p->text().toDouble(&ok));  if(!ok || (newp==0)) return;
    double newscalerad=ui.lineEdit_scalerad->text().toDouble(&ok);  if (!ok) return;
    double newscalerad2=ui.lineEdit_scalerad2->text().toDouble(&ok);  if (!ok) return;
    double newGamma=ui.lineEdit_Gamma->text().toDouble(&ok);  if (!ok) return;
    double newSersicIndex=ui.lineEdit_SersicIndex->text().toDouble(&ok);  if (!ok) return;
    double newMbh=ui.lineEdit_Mbh->text().toDouble(&ok);  if(!ok) return;
    if( PotentialType!=core->configPotential[currentComponentIndex].PotentialType || 
        newNcoefs_radial!=core->configPotential[currentComponentIndex].Ncoefs_radial || 
        newNcoefs_vertical!=core->configPotential[currentComponentIndex].Ncoefs_vertical || 
        newNcoefs_angular!=core->configPotential[currentComponentIndex].Ncoefs_angular || 
        newAlpha!=core->configPotential[currentComponentIndex].Alpha || 
        newRmax!=core->configPotential[currentComponentIndex].Rmax || 
        newtreecodeEps!=core->configPotential[currentComponentIndex].treecodeEps || 
        newtreecodeTheta!=core->configPotential[currentComponentIndex].treecodeTheta ||
        SymmetryType!=core->configPotential[currentComponentIndex].SymmetryType || 
        DensityType!=core->configPotential[currentComponentIndex].DensityType || 
        newmass!=core->configPotential[currentComponentIndex].mass || 
        newq!=core->configPotential[currentComponentIndex].q || 
        newp!=core->configPotential[currentComponentIndex].p || 
        newscalerad!=core->configPotential[currentComponentIndex].scalerad || 
        newscalerad2!=core->configPotential[currentComponentIndex].scalerad2 || 
        newGamma!=core->configPotential[currentComponentIndex].Gamma || 
        newSersicIndex!=core->configPotential[currentComponentIndex].sersicIndex || 
        newMbh!=core->configPotential[0].Mbh )
    {
        core->configPotential[currentComponentIndex].PotentialType=PotentialType;
        core->configPotential[currentComponentIndex].Ncoefs_radial=newNcoefs_radial; 
        core->configPotential[currentComponentIndex].Ncoefs_vertical=newNcoefs_vertical; 
        core->configPotential[currentComponentIndex].Ncoefs_angular=newNcoefs_angular; 
        core->configPotential[currentComponentIndex].Alpha=newAlpha;
        core->configPotential[currentComponentIndex].Rmax=newRmax;
        core->configPotential[currentComponentIndex].treecodeEps=newtreecodeEps;
        core->configPotential[currentComponentIndex].treecodeTheta=newtreecodeTheta;
        core->configPotential[currentComponentIndex].SymmetryType=SymmetryType;
        core->configPotential[currentComponentIndex].DensityType=DensityType;
        core->configPotential[currentComponentIndex].mass=newmass; 
        core->configPotential[currentComponentIndex].q=newq; 
        core->configPotential[currentComponentIndex].p=newp; 
        core->configPotential[currentComponentIndex].scalerad=newscalerad; 
        core->configPotential[currentComponentIndex].scalerad2=newscalerad2; 
        core->configPotential[currentComponentIndex].Gamma=newGamma; 
        core->configPotential[currentComponentIndex].sersicIndex=newSersicIndex; 
        core->configPotential[0].Mbh=newMbh;   // black hole is always in 0th component
        bool initNow=true;
        for(size_t i=0; i<core->configPotential.size(); i++)
            if( core->configPotential[i].PotentialType==CDensity::PT_BSE ||
                core->configPotential[i].PotentialType==CDensity::PT_BSECOMPACT ||
                core->configPotential[i].PotentialType==CDensity::PT_SPLINE || 
                core->configPotential[i].PotentialType==CDensity::PT_CYLSPLINE ||
                core->configPotential[i].PotentialType==CDensity::PT_NB)
                initNow=false;
        if(initNow)
            on_buttonInitPotential_clicked();  // do initialization for 'lightweight' potentials only
        else
            ui.buttonInitPotential->show();    // let the user press this button manually for 'heavy-weight' initialization of general-purpose potentials listed above
    }
}

void CSmileGUI::on_buttonInitPotential_clicked()
{
    core->initPotential();
    ICchanged();
    if( core->configPotential[currentComponentIndex].PotentialType==CDensity::PT_BSE || 
        core->configPotential[currentComponentIndex].PotentialType==CDensity::PT_BSECOMPACT || 
        core->configPotential[currentComponentIndex].PotentialType==CDensity::PT_SPLINE ||
        core->configPotential[currentComponentIndex].PotentialType==CDensity::PT_CYLSPLINE)
    {   // the symmetry type and parameters may have changed by potential constructor, especially if it loaded coefficients from file rather than initialized from Nbody file
        InternalRecalc=true;
        ui.selectDensityType-> setCurrentIndex(ui.selectDensityType-> findText(
            getDensityNameByType (core->configPotential[currentComponentIndex].DensityType)));
        ui.selectSymmetryType->setCurrentIndex(ui.selectSymmetryType->findText(
            getSymmetryNameByType(core->configPotential[currentComponentIndex].SymmetryType)));
        ui.lineEdit_Ncoefs_radial->setEnabled(core->configPotential[currentComponentIndex].DensityType!=CDensity::PT_COEFS);
        ui.lineEdit_Ncoefs_angular->setEnabled(core->configPotential[currentComponentIndex].DensityType!=CDensity::PT_COEFS);
        ui.lineEdit_Alpha->setEnabled(core->configPotential[currentComponentIndex].DensityType!=CDensity::PT_COEFS);
        // ensure that N_radial, N_angular and Alpha input lines are consistent with corresponding variables (which may have been updated after loading a coef file)
        if(core->configPotential[currentComponentIndex].Ncoefs_radial != static_cast<unsigned int>(ui.lineEdit_Ncoefs_radial->text().toInt()))
            ui.lineEdit_Ncoefs_radial->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_radial));
        if(core->configPotential[currentComponentIndex].Ncoefs_vertical != static_cast<unsigned int>(ui.lineEdit_Ncoefs_vertical->text().toInt()))
            ui.lineEdit_Ncoefs_vertical->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_vertical));
        if(core->configPotential[currentComponentIndex].Ncoefs_angular != static_cast<unsigned int>(ui.lineEdit_Ncoefs_angular->text().toInt())) 
            ui.lineEdit_Ncoefs_angular->setText(QString::number(core->configPotential[currentComponentIndex].Ncoefs_angular));
        if(core->configPotential[currentComponentIndex].Alpha != ui.lineEdit_Alpha->text().toDouble())
            ui.lineEdit_Alpha->setText(QString::number(core->configPotential[currentComponentIndex].Alpha));
        if(core->configPotential[currentComponentIndex].Rmax != ui.lineEdit_Rmax->text().toDouble())
            ui.lineEdit_Rmax->setText(QString::number(core->configPotential[currentComponentIndex].Rmax));
        if(core->configPotential[0].Mbh != ui.lineEdit_Mbh->text().toDouble())
            ui.lineEdit_Mbh->setText(QString::number(core->configPotential[0].Mbh));
        InternalRecalc=false;
    }
    ui.buttonInitPotential->hide();   // this button will be shown only when parameters changed
    if(core->configOrbit.Omega>0) {
        double LX, LY;
        findLagrangianPoints(core->potential, core->configOrbit.Omega, &LX, &LY);
        info("Lagrangian points: L1x="+QString::fromStdString(convertToString(LX))+
             ", L4y="+QString::fromStdString(convertToString(LY)));
    }
    ui.frame_sidebar->repaint();
    refreshPotentialComponentList();
    if(core->potential->PotentialType()==CPotential::PT_COMPOSITE && 
        static_cast<const CPotentialComposite*>(core->potential)->size()>1)
        ui.lineEdit_smDensityComponents->setEnabled(true);
    else {
        ui.lineEdit_smDensityComponents->setEnabled(false);
        ui.lineEdit_smDensityComponents->setText("");  // clear means use entire potential
    }
}

void CSmileGUI::chooseNbodyFile()
{
    assignNbodyFile(QFileDialog::getOpenFileName(this, "Open Nbody file", core->WorkDir));
}

void CSmileGUI::typeinNbodyFile()
{
    QString fileName = ui.lineEdit_NbodyFile->text();
    if(!QFile::exists(fileName))
        fileName = QFileDialog::getOpenFileName(this, "Open Nbody file", core->WorkDir);
    assignNbodyFile(fileName);
}

void CSmileGUI::assignNbodyFile(QString fileName)
{
    if(fileName.isEmpty() || !QFile::exists(fileName)) return;
    core->WorkDir = getDirName(fileName);
    core->configPotential[currentComponentIndex].NbodyFile=fileName.toStdString();
    ui.lineEdit_NbodyFile->setText(fileName);
    // set explicitly because it may have been reset to default if previous load was unsuccessful
    core->configPotential[currentComponentIndex].PotentialType = getPotentialTypeByName(ui.selectPotentialType->currentText().toStdString());
    on_buttonInitPotential_clicked();  // do initialization
}

void CSmileGUI::integratorChanged()
{
    if(InternalRecalc) return;
    InternalRecalc=true;    // temporary set flag so that changing options below does not call this function again
    CBasicOdeIntegrator::STEPPERKIND integratorType = getIntegratorTypeByName(ui.selectIntegratorType->currentText().toStdString());
    if( integratorType==CBasicOdeIntegrator::SK_DEFAULT ||
        integratorType==CBasicOdeIntegrator::SK_DOP853 ||
        integratorType==CBasicOdeIntegrator::SK_HERMITE ||
        integratorType==CBasicOdeIntegrator::SK_ODEINT_CK5 ||
        integratorType==CBasicOdeIntegrator::SK_ODEINT_DP5 ||
        integratorType==CBasicOdeIntegrator::SK_ODEINT_BS3 ||
        integratorType==CBasicOdeIntegrator::SK_ODEINT_BS )
    {
        ui.label_tol_rel->show();
        ui.label_tol_ias->hide();
        ui.label_timestep_factor->hide();
        ui.lineEdit_tol_rel->show();
        ui.lineEdit_tol_ias->hide();
        ui.lineEdit_timestep_factor->hide();
        if(integratorType==CBasicOdeIntegrator::SK_HERMITE) {
            ui.lineEdit_tol_abs->hide();
            ui.label_tol_abs->hide();
        } else {
            ui.lineEdit_tol_abs->show();
            ui.label_tol_abs->show();
        }
    } else if(integratorType==CBasicOdeIntegrator::SK_IAS15) {
        ui.label_tol_abs->hide();
        ui.label_tol_rel->hide();
        ui.label_tol_ias->show();
        ui.label_timestep_factor->hide();
        ui.lineEdit_tol_abs->hide();
        ui.lineEdit_tol_rel->hide();
        ui.lineEdit_tol_ias->show();
        ui.lineEdit_timestep_factor->hide();
    } else {
        ui.label_tol_abs->hide();
        ui.label_tol_rel->hide();
        ui.label_tol_ias->hide();
        ui.label_timestep_factor->show();
        ui.lineEdit_tol_abs->hide();
        ui.lineEdit_tol_rel->hide();
        ui.lineEdit_tol_ias->hide();
        ui.lineEdit_timestep_factor->show();
    }
    core->configOrbit.integratorType = integratorType;
    bool ok=true;
    double n=ui.lineEdit_tol_abs->text().toDouble(&ok);
    if(ok) core->configOrbit.accuracyAbs=n;
    n=ui.lineEdit_tol_rel->text().toDouble(&ok);
    if(ok) core->configOrbit.accuracyRel=n;
    n=ui.lineEdit_tol_ias->text().toDouble(&ok);
    if(ok) core->configOrbit.accuracyIAS15=n;
    n=ui.lineEdit_timestep_factor->text().toDouble(&ok);
    if(ok) core->configOrbit.accuracyFixedTimestep=n;
    n=ui.lineEdit_relax_rate->text().toDouble(&ok);
    if(ok) core->configOrbit.relaxRate=n;
    InternalRecalc=false;
}

/* ----------------------------  modification of various parameters by GUI lineEdits ----- */

void CSmileGUI::ICchanged()
{
    if(InternalRecalc) return;
    bool ok=true;
    double newOmega=ui.lineEdit_Omega->text().toDouble(&ok);  if (!ok) return;
    double newE=ui.lineEdit_E->text().toDouble(&ok);  if(!ok) return;
    double X=ui.lineEdit_x->text().toDouble(&ok);  if(!ok) return;
    double Y=ui.lineEdit_y->text().toDouble(&ok);  if(!ok) return;
    double Z=(core->configOrbit.N_dim==3)?ui.lineEdit_z->text().toDouble(&ok):0;  if(!ok) return;
    double Vx=ui.lineEdit_vx->text().toDouble(&ok);  if(!ok) return;
    double Vy=ui.lineEdit_vy->text().toDouble(&ok);  if(!ok) return;
    double Vz=(core->configOrbit.N_dim==3)?ui.lineEdit_vz->text().toDouble(&ok):0;  if(!ok) return;
    if(core->useICe)
    {
        core->initCondE=newE;
    }
    else
    {
        core->initCond=CPosVelPoint<double>(X, Y, Z, Vx, Vy, Vz);
    }
    bool omegaChanged=core->configOrbit.Omega!=newOmega;
    core->configOrbit.Omega=newOmega;
    InternalRecalc=true;
    core->initIC();
    if(core->useICe)
    {
        ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
        ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
        ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
        ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
        ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
        ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    }
    else
    {
        ui.lineEdit_E->setText(QString::number(core->initCondE));
    }
    ui.lineEdit_Torb->setText(QString::number(core->timeUnit));
    InternalRecalc=false;
    if((core->configCore.PSdisplayCoord==1 && core->configCore.PSsectionCoord==2 && 
        (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC && 
        fabs(pow_2(core->initCond.Pos[0]*core->initCond.Vel[1]-core->initCond.Pos[1]*core->initCond.Vel[0])-Lz2prev)>1e-4) ||
        fabs(core->initCondE-Eprev)> 1e-4)
        on_ButtonClearPS_clicked();
    redrawPotential();
    if(omegaChanged && core->configOrbit.Omega>0) {
        double LX, LY;
        findLagrangianPoints(core->potential, core->configOrbit.Omega, &LX, &LY);
        info("Lagrangian points: L1x="+QString::fromStdString(convertToString(LX))+
             ", L4y="+QString::fromStdString(convertToString(LY)));

    }
}

void CSmileGUI::intTime_changed()
{
    if(InternalRecalc) return;
    bool ok;
    double t = ui.lineEdit_intTime->text().toDouble(&ok);
    if(ok)  core->configCore.intTimeInPeriods=t;
    t = ui.lineEdit_timeStep->text().toDouble(&ok);
    if(ok)  core->configOrbit.samplingInterval=t;
}

void CSmileGUI::on_checkBox_fmExistingSS_stateChanged(int Checked)
{
    ui.lineEdit_fms->setEnabled(Checked==0);
    ui.lineEdit_fmp->setEnabled(Checked==0);
    ui.lineEdit_fmr->setEnabled(Checked==0);
    ui.lineEdit_fmy->setEnabled(Checked==0);
}

void CSmileGUI::fmparams_changed()
{
    if(InternalRecalc) return;
    bool ok; int t;
    t = ui.lineEdit_fms->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsStationary=t;
    t = ui.lineEdit_fmp->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsPrincipalPlane=t;
    t = ui.lineEdit_fmy->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsYalpha=t;
    t = ui.lineEdit_fmr->text().toInt(&ok);
    if(ok)  core->configCore.fm_numOrbitsRandom=t;
}

void CSmileGUI::smparams_changed()
{
    if(InternalRecalc) return;
    bool ok; int t; double d;
    // orbit library tab
    t = ui.lineEdit_icgNumOrbits->text().toInt(&ok);
    if(ok)  core->configCore.sm_numOrbits=t;
    t = ui.lineEdit_smNumSamplingPoints->text().toInt(&ok);
    if(ok) core->configCore.sm_numSamplingPoints=t;
    t = ui.comboBox_smICGenerator->currentIndex();
    switch(t) {
        case 0:
            core->configCore.sm_ICgenType=CBasicICGenerator::IC_VEL_EDDINGTON;
            ui.groupBox_smICG->show();
            break;
        case 1:
            core->configCore.sm_ICgenType=CBasicICGenerator::IC_VEL_JEANSSPH;
            ui.groupBox_smICG->show();
            break;
        case 2:
            core->configCore.sm_ICgenType=CBasicICGenerator::IC_VEL_JEANSAXI;
            ui.groupBox_smICG->show();
            break;
        default:
            ui.groupBox_smICG->hide();
            break;
    }
    d = ui.lineEdit_icgSkewE->text().toDouble(&ok);
    if(ok) core->configCore.weightSkewFactorE=d;
    if(core->configCore.sm_ICgenType==CBasicICGenerator::IC_VEL_EDDINGTON) {
        ui.lineEdit_icgSkewL->show();
        ui.label_icgSkewL->show();
        d = ui.lineEdit_icgSkewL->text().toDouble(&ok);
        if(ok) core->configCore.weightSkewFactorL=d;
    } else {
        ui.lineEdit_icgSkewL->hide();
        ui.label_icgSkewL->hide();
    }
    if(core->configCore.sm_ICgenType==CBasicICGenerator::IC_VEL_JEANSSPH) {
        ui.lineEdit_icgBeta->show();
        ui.label_icgBeta->show();
        d = ui.lineEdit_icgBeta->text().toDouble(&ok);
        if(ok) core->configCore.sm_ICbeta=d;
    } else {
        ui.lineEdit_icgBeta->hide();
        ui.label_icgBeta->hide();
    }
    if(core->configCore.sm_ICgenType==CBasicICGenerator::IC_VEL_JEANSAXI) {
        ui.lineEdit_icgBetaAxi->show();
        ui.label_icgBetaAxi->show();
        ui.lineEdit_icgKappaAxi->show();
        ui.label_icgKappaAxi->show();
        d = ui.lineEdit_icgBetaAxi->text().toDouble(&ok);
        if(ok) core->configCore.sm_ICbetaAxi=d;
        d = ui.lineEdit_icgKappaAxi->text().toDouble(&ok);
        if(ok) core->configCore.sm_ICkAxi=d;
    } else {
        ui.lineEdit_icgBetaAxi->hide();
        ui.label_icgBetaAxi->hide();
        ui.lineEdit_icgKappaAxi->hide();
        ui.label_icgKappaAxi->hide();
    }
    // model params tab
    CBasicSchwData::SCHWDATATYPE densDataType = getSchwDataTypeByName(ui.select_smType->currentText().toStdString());
    if(densDataType==CBasicSchwData::SD_DENS_GRIDCLASSIC)
    {
        ui.lineEdit_smNumLines->show();
        ui.label_smNumLines->show();
        ui.lineEdit_smNumAngCoefs->hide();
        ui.label_smNumAngCoefs->hide();
    }
    else
    {
        ui.lineEdit_smNumLines->hide();
        ui.label_smNumLines->hide();
        ui.lineEdit_smNumAngCoefs->show();
        ui.label_smNumAngCoefs->show();
    }
    if(densDataType==CBasicSchwData::SD_DENS_GRIDCYLINDRICAL)
    {
        ui.lineEdit_smNumVerCoefs->show();
        ui.label_smNumVerCoefs->show();
    }
    else
    {
        ui.lineEdit_smNumVerCoefs->hide();
        ui.label_smNumVerCoefs->hide();
    }
    if(densDataType==CBasicSchwData::SD_DENS_SHBSE)
    {
        ui.lineEdit_smAlpha->show();
        ui.label_smAlpha->show();
    }
    else
    {
        ui.lineEdit_smAlpha->hide();
        ui.label_smAlpha->hide();
    }
    if(densDataType!=core->configCore.sm_densDataType)
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_densDataType=densDataType;
    }
    // check density components
    QString comps=ui.lineEdit_smDensityComponents->text();
    std::vector<unsigned int> components;
    if(!comps.isEmpty() && comps!="all") {
        QStringList SL=comps.split(QRegExp("[,\\s]+"),QString::SkipEmptyParts);
        for(int i=0; i<SL.size(); i++) {
            bool ok;
            int c=SL[i].toInt(&ok);
            if(ok) components.push_back(c);
        }
    }
    if(components!=core->configCore.densityComponents)
    {
        ui.groupBox_smOptim->setEnabled(false);
        if(core->density!=NULL && core->density!=core->potential)  // invalidate an existing density object, it will be recreated when necessary
            delete core->density;
        core->density=NULL;
        core->configCore.densityComponents=components;
    }
    // check other density model params
    t = ui.lineEdit_smNumAngCoefs->text().toInt(&ok);
    if(ok&&(core->configCore.sm_numAngularCoefs!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_numAngularCoefs=t;
    }
    t = ui.lineEdit_smNumRadCoefs->text().toInt(&ok);
    if(ok&&(core->configCore.sm_numRadialCoefs!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_numRadialCoefs=t;
    }
    t = ui.lineEdit_smNumVerCoefs->text().toInt(&ok);
    if(ok&&(core->configCore.sm_numVerticalCoefs!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_numVerticalCoefs=t;
    }
    d = ui.lineEdit_smAlpha->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_Alpha!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_Alpha=d;
    }
    t = ui.lineEdit_smNumLines->text().toInt(&ok);
    if(ok&&(core->configCore.sm_linesPerSegment!=t))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_linesPerSegment=t;
    }
    t = ui.lineEdit_smNumShells->text().toInt(&ok);
    if(ok)
    {
        if(core->configCore.sm_numShells!=t) ui.groupBox_smOptim->setEnabled(false);
        ui.selectEnergyLevel_sm->setMaximum(t);
        ui.selectEnergyLevel_fm->setMaximum(t);
        core->configCore.sm_numShells=t;
    }
    d = ui.lineEdit_smInnerShellMass->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_innerShellMass!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_innerShellMass=d;
    }
    d = ui.lineEdit_smOuterShellMass->text().toDouble(&ok);
    if(ok&&(core->configCore.sm_outerShellMass!=d))
    {
        ui.groupBox_smOptim->setEnabled(false);
        core->configCore.sm_outerShellMass=d;
    }
    // optimization tab
    d = ui.lineEdit_smConstraintPenaltyLin->text().toDouble(&ok);
    if(ok) core->configCore.sm_constraintPenaltyLin = d;
    d = ui.lineEdit_smConstraintPenaltyQuad->text().toDouble(&ok);
    if(ok) core->configCore.sm_constraintPenaltyQuad = d;
    d = ui.lineEdit_smRegularization->text().toDouble(&ok);
    if(ok) core->configCore.sm_regularization = d;
    d = ui.lineEdit_smChaoticPenalty->text().toDouble(&ok);
    if(ok) core->configCore.sm_chaoticPenalty = d;
    d = ui.lineEdit_smMaxOrbitWeight->text().toDouble(&ok);
    if(ok) core->configCore.sm_maxWeight = d;
    core->configCore.sm_constrainBeta=ui.checkBox_smBeta->isChecked();
    ui.lineEdit_smBetaIn->setEnabled(core->configCore.sm_constrainBeta);
    ui.lineEdit_smBetaOut->setEnabled(core->configCore.sm_constrainBeta);
    d = ui.lineEdit_smBetaIn->text().toDouble(&ok);
    if(ok)  core->configCore.sm_betaIn = d;
    d = ui.lineEdit_smBetaOut->text().toDouble(&ok);
    if(ok)  core->configCore.sm_betaOut = d;
    // compute # of coefs
    int numCoefs=0;
    switch(densDataType)
    {
    case CBasicSchwData::SD_DENS_GRIDCLASSIC: 
        numCoefs=3*core->configCore.sm_linesPerSegment*core->configCore.sm_linesPerSegment*core->configCore.sm_numRadialCoefs;
        break;
    case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL:
        numCoefs=core->configCore.sm_numRadialCoefs*core->configCore.sm_numVerticalCoefs*std::max<size_t>(1,core->configCore.sm_numAngularCoefs);
        break;
    case CBasicSchwData::SD_DENS_SHMESH: 
        numCoefs=1+core->configCore.sm_numRadialCoefs*(core->configCore.sm_numAngularCoefs/2+1)*(core->configCore.sm_numAngularCoefs/2+2)/2;
        break;
    case CBasicSchwData::SD_DENS_SHBSE: 
        numCoefs=(1+core->configCore.sm_numRadialCoefs)*(core->configCore.sm_numAngularCoefs/2+1)*(core->configCore.sm_numAngularCoefs/2+2)/2;
        break;
    default: info("Invalid type of Schwarzschild model");
    }
    ui.lineEdit_smNumCoefs->setText(QString::number(numCoefs));
}

void CSmileGUI::ccparams_changed()
{
    if(InternalRecalc) return;
    bool ok;
    double d = ui.lineEdit_MinFreqDiff->text().toDouble(&ok);
    if(ok)  core->configCore.chaoticMinFreqDiff=d;
    d = ui.lineEdit_MinLambda->text().toDouble(&ok);
    if(ok)  core->configCore.chaoticMinLambda=d;
    redrawFreqMap(true);
}

void CSmileGUI::on_comboBox_Lyapunov_currentIndexChanged(int)
{
    if(InternalRecalc) return;
    core->configOrbit.calcLyapunovMethod = ui.comboBox_Lyapunov->currentIndex();
}

void CSmileGUI::on_checkBox_usePS_stateChanged(int Checked)
{
    if(InternalRecalc) return;
    core->configCore.usePS = (Checked>0);
}

void CSmileGUI::on_radioButton_2d_toggled(bool n2d)
{
    core->configOrbit.N_dim=n2d?2:3;
    if(n2d)
    {
        ui.radioButton_2d_xy->setChecked(true);
        ui.radioButton_2d_xz->setEnabled(false);
        ui.radioButton_2d_yz->setEnabled(false);
        ui.radioButton_3dline->setEnabled(false);
        ui.radioButton_3dmesh->setEnabled(false);
        ui.radioButton_axis_z->setEnabled(false);
        ui.lineEdit_z->setEnabled(false);
        ui.lineEdit_vz->setEnabled(false);
        ui.lineEdit_z->setText("0");
        ui.lineEdit_vz->setText("0");
        PScoordChanged();
    }
    else
    {
        ui.lineEdit_z->setEnabled(true);
        ui.lineEdit_vz->setEnabled(true);
        ui.radioButton_2d_xz->setEnabled(true);
        ui.radioButton_2d_yz->setEnabled(true);
        ui.radioButton_axis_z->setEnabled(true);
#ifdef HAVE_QWT3D
        ui.radioButton_3dline->setEnabled(true);
#ifdef USE3DFACETS
        ui.radioButton_3dmesh->setEnabled(true);
#endif
#endif
    }
    if(InternalRecalc) return;
    ICchanged();
}

void CSmileGUI::on_radioButton_ICe_toggled(bool useICe)
{
    if(InternalRecalc) return;
    core->useICe=useICe;
    if(useICe)
    {
        ui.lineEdit_E->setEnabled(true);
        ui.groupBox_ICx->setEnabled(false);
    }
    else
    {
        ui.lineEdit_E->setEnabled(false);
        ui.groupBox_ICx->setEnabled(true);
    }
}

/* ----------------------------  Random initial conditions --------- */
void CSmileGUI::on_ButtonRandom_clicked()
{
    CICGeneratorRandomE icg(core->potential, core->initCondE, core->configOrbit.N_dim, core->configOrbit.Omega);
    CPointMassSet<float> IC;
    icg.createInitialConditions(1, &IC);
    if(IC.size()<1)
    {
        error("Total energy is outside allowed range");
        return;  // error
    }
    InternalRecalc=true;
    core->initCond=IC[0].first;
    ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    InternalRecalc=false;
    on_ButtonStart_clicked();
    if(core->configCore.PSdisplayCoord==1 && core->configCore.PSsectionCoord==2 && 
        (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC)
        on_ButtonClearPS_clicked();
}

/* ----------------------------  Poincare section operations --------- */
void CSmileGUI::PSclicked(const QPointF &pos)
{
    double dirX=(core->configCore.PSdisplayCoord==0), 
           dirY=(core->configCore.PSdisplayCoord==1), 
           dirZ=(core->configCore.PSdisplayCoord==2);
    double secX=(core->configCore.PSsectionCoord==0), 
           secY=(core->configCore.PSsectionCoord==1), 
           secZ=(core->configCore.PSsectionCoord==2);
    bool meridionalPlane = dirY && secZ && 
        (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC;
    double Lz = !meridionalPlane ? 0 :
        core->initCond.Pos[0]*core->initCond.Vel[1]-core->initCond.Pos[1]*core->initCond.Vel[0];
    double newd=pos.x();
    double newvd=pos.y();
    double newvs2=2*(core->initCondE - core->potential->Phi(newd*dirX, newd*dirY, newd*dirZ))
        + pow_2(core->configOrbit.Omega*newd)*(core->configCore.PSdisplayCoord<2) - pow_2(newvd) +
        (meridionalPlane ? -pow_2(Lz/newd) : 0);
    if(newvs2>=0)
    {
        InternalRecalc=true;
        double newvs=sqrt(newvs2);
        if(meridionalPlane)   // vs is the radial velocity (Y); need to determine tangential one (X) as well
            secX=-(Lz/newd)/newvs;
        core->initCond=CPosVelPoint<double>(newd*dirX, newd*dirY, newd*dirZ, 
            newvd*dirX+newvs*secX-core->configOrbit.Omega*newd*dirY, 
            newvd*dirY+newvs*secY+core->configOrbit.Omega*newd*dirX, newvd*dirZ+newvs*secZ);
        ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0], 'g', 7));
        ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1], 'g', 7));
        ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2], 'g', 7));
        ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0], 'g', 7));
        ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1], 'g', 7));
        ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2], 'g', 7));
        InternalRecalc=false;
    }
}

/// \cond INTERNAL_DOCS
struct CMaxVRparam {
    const CPotential* pot;
    double E, Lz2;
};
double findMaxVR(double R, void* param)
{
    double xyz[3]={R,0,0};
    double F[4]={0,0,0,0};
    ((CMaxVRparam*)param)->pot->Force(xyz, 0, F, NULL);
    return F[0]*R*R*R+((CMaxVRparam*)param)->Lz2;
}
double findZeroVR(double R, void* param)
{
    return ((CMaxVRparam*)param)->E - ((CMaxVRparam*)param)->pot->Phi(R, 0, 0) - ((CMaxVRparam*)param)->Lz2/(2*R*R);
}
/// \endcond

void CSmileGUI::on_ButtonClearPS_clicked()
{
    const size_t NUM_POINTS_PLOT_POINCARE=400;
    while(CurvesPoincare.size()>0)
    {
        delete CurvesPoincare[CurvesPoincare.size()-1];
        CurvesPoincare.remove(CurvesPoincare.size()-1);
    }
    double xm[NUM_POINTS_PLOT_POINCARE*4], vxm[NUM_POINTS_PLOT_POINCARE*4];
    double dirX=(core->configCore.PSdisplayCoord==0), 
           dirY=(core->configCore.PSdisplayCoord==1), 
           dirZ=(core->configCore.PSdisplayCoord==2);
    bool meridionalPlane = dirY && core->configCore.PSsectionCoord==2 && 
        (core->potential->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC;
    double Lz2 = !meridionalPlane ? 0 :
        pow_2(core->initCond.Pos[0]*core->initCond.Vel[1]-core->initCond.Pos[1]*core->initCond.Vel[0]);
    double Xmax=findIntersection(core->potential, core->initCondE, dirX, dirY, dirZ, core->configOrbit.Omega);
    if(Xmax<0 && core->configOrbit.Omega!=0) { // use corotation radius instead
        double XX=0;
        findLagrangianPoints(core->potential, core->configOrbit.Omega, &Xmax, &XX);
        Xmax=std::max<double>(Xmax, XX);
    }
    if(Xmax>0 && gsl_finite(Xmax)) 
    {
        double Xmin=0, Xhalf=Xmax/2;
        if(Lz2>0) {
            gsl_function F;
            CMaxVRparam par;
            par.pot=core->potential;
            par.Lz2=Lz2;
            par.E=core->initCondE;
            F.function=&findMaxVR;
            F.params=&par;
            double Rmin=Xmax/100;
            while(Rmin>Xmax*1e-6 && findMaxVR(Rmin, &par)<0) Rmin/=2;
            Xhalf=findRoot(&F, Rmin, Xmax, 0);  // location of maximum in V_R
            if(Xhalf>0) {
                F.function=&findZeroVR;
                while(Rmin>Xhalf*1e-6 && findZeroVR(Rmin, &par)>0) Rmin/=2;
                Xmin=findRoot(&F, Rmin, Xhalf, 0);
                Xmax=findRoot(&F, Xhalf, Xmax, Xmax);
            } else Xhalf=Xmax/2;
        }
        size_t numpt=0;
        for(size_t i=0; i<NUM_POINTS_PLOT_POINCARE; i++)
        {
            double x= i<NUM_POINTS_PLOT_POINCARE/2 ? 
                Xmin+(Xhalf-Xmin)*pow((i+0.5)/(NUM_POINTS_PLOT_POINCARE/2), Xmin>0 ? 2.0 : 1.0) :
                Xhalf + (Xmax-Xhalf)*(1.0-pow_2((NUM_POINTS_PLOT_POINCARE-i-0.5)/(NUM_POINTS_PLOT_POINCARE/2)));
            double vx2=2*(core->initCondE - core->potential->Phi(x*dirX, x*dirY, x*dirZ))
                + pow_2(core->configOrbit.Omega*x)*(core->configCore.PSdisplayCoord<2) +
                (meridionalPlane ? -Lz2/pow_2(x) : 0);
            if(vx2>0) {
                xm[numpt]=x;
                vxm[numpt]=sqrt(vx2);
                numpt++;
            }
        }
        for(size_t i=0; i<numpt; i++) {  // copy nonzero values to all four quadrants (assuming symmetry?..)
            xm[numpt*2-1-i]=xm[i];
            vxm[numpt*2-1-i]=-vxm[i];
            if(!meridionalPlane) {
                xm[numpt*2+i]=-xm[i];
                vxm[numpt*2+i]=-vxm[i];
                xm[numpt*4-1-i]=-xm[i];
                vxm[numpt*4-1-i]=vxm[i];
            }
        }
        QwtPlotCurve* CurveBoundary = new QwtPlotCurve();
        CurveBoundary->setPen(QColor(Qt::red));
        CurveBoundary->attach(PlotPoincare);
        CurveBoundary->SETDATA(xm, vxm, numpt*(meridionalPlane?2:4));
        CurvesPoincare<<(CurveBoundary);
    }
    for(int i=0; i<PlotPoincare->axisCnt; i++) PlotPoincare->setAxisAutoScale(i); 
    PlotPoincare->replot();
    ZoomerPoincare->setZoomBase(true);
    Eprev=core->initCondE;
    Lz2prev=Lz2;
}

void CSmileGUI::addPoincareCurve(const CPoincareInformation* tps)
{
    QwtPlotCurve* CurvePS = new QwtPlotCurve();
    CurvePS->setSymbol(QWTSYM(QwtSymbol::Rect, 2, QColor(rand()%224, rand()%192, rand()%224)));
    CurvePS->setStyle(QwtPlotCurve::Dots);
    CurvePS->attach(PlotPoincare);
    size_t nPS=tps->size();
    vectord psx(nPS), psvx(nPS);
    for(size_t i=0; i<nPS; i++)
    {
        psx[i]=tps->at(i).first;
        psvx[i]=tps->at(i).second;
    }
    CurvePS->SETDATA(&(psx.front()), &(psvx.front()), static_cast<int>(psx.size()));
    CurvesPoincare<<(CurvePS);
}

void CSmileGUI::on_ButtonFindPeriodic_clicked()
{
    unsigned int Nperiods=ui.spinBox_Nperiods->value();
    bool success=findPeriodicOrbit(core->potential, core->configOrbit.Omega, 
        core->configCore.PSsectionCoord, core->configCore.PSdisplayCoord, 
        false/*core->configOrbit.N_dim==N_DIM*/, Nperiods, &(core->initCond));
    if(success) {
        InternalRecalc=true;
        ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0], 'f', 7));
        ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
        ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
        ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0], 'f', 7));
        ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1], 'f', 7));
        ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
        InternalRecalc=false;
    } else 
        my_error(FUNCNAME, "Could not locate periodic orbit");
}

void CSmileGUI::PScoordChanged()
{
    if(InternalRecalc) return;
    unsigned int PSdisplayCoord=ui.comboBox_PSdisplayCoord->currentIndex();
    unsigned int PSsectionCoord=ui.comboBox_PSsectionCoord->currentIndex();
    if(core->configOrbit.N_dim<N_DIM && PSdisplayCoord+PSsectionCoord>=2) { // in case of 2d, check that both coordinate indices are <=1
        core->configCore.PSdisplayCoord=0;
        core->configCore.PSsectionCoord=1;
        InternalRecalc=true;
        ui.comboBox_PSdisplayCoord->setCurrentIndex(core->configCore.PSsectionCoord);
        ui.comboBox_PSsectionCoord->setCurrentIndex(core->configCore.PSdisplayCoord);
        InternalRecalc=false;
        on_ButtonClearPS_clicked();
        return;
    }
    if(PSdisplayCoord==core->configCore.PSdisplayCoord && PSsectionCoord==core->configCore.PSsectionCoord)
        return;
    InternalRecalc=true;
    if(PSdisplayCoord==PSsectionCoord) {  // shouldn't be equal
        if(PSdisplayCoord!=core->configCore.PSdisplayCoord) { // this one has changed -- adjust the other one
            ui.comboBox_PSsectionCoord->setCurrentIndex(core->configCore.PSdisplayCoord);
            PSsectionCoord=core->configCore.PSsectionCoord=core->configCore.PSdisplayCoord;
        } else {  // vice versa
            ui.comboBox_PSdisplayCoord->setCurrentIndex(core->configCore.PSsectionCoord);
            PSdisplayCoord=core->configCore.PSdisplayCoord=core->configCore.PSsectionCoord;
        }
    }
    core->configCore.PSsectionCoord=PSsectionCoord;
    core->configCore.PSdisplayCoord=PSdisplayCoord;
    InternalRecalc=false;
    on_ButtonClearPS_clicked();
}

/* ----------------------------  Frequency map operations --------- */

// disable several buttons for the duration of orbit library integration, optimization, or export to Nbody
void CSmileGUI::enableOrbitLibControls(bool enable)
{
    ui.groupBox_N_dim->setEnabled(enable && !core->isOrbitIntRunning());
    ui.groupBox_potential->setEnabled(enable && !core->isOrbitIntRunning());
    ui.groupBox_smOptim->setEnabled(enable);
    ui.groupBox_smGrid->setEnabled(enable);
    ui.ButtonImportFM->setEnabled(enable);
    ui.ButtonImportSM->setEnabled(enable);
}

void CSmileGUI::on_ButtonFreqMap_clicked()
{
    if(core->isExportNbodyRunning() || core->isSchwOptRunning()) return;
    if(core->isOrbitLibRunning())
    {
        emit stopOrbitLibrary();
        return;
    }
    CurveFreqMapR->SETDATA(NULL, NULL, 0);
    CurveFreqMapC->SETDATA(NULL, NULL, 0);
    // start building freq map
    core->buildFreqMapIC(core->initCondE, ui.checkBox_fmExistingSS->isChecked());
    core->startOrbitLibrary(SLOT(coreOrbitLibraryFinished()));
    ui.progressBarFM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarFM->setValue(0);
    ui.progressBarFM->show();
    timeFM.start();
    ui.ButtonFreqMap->setText("Stop");
    ui.tabWidget_sm->setEnabled(false);   // can't do Schw.modelling while running freq.map
    enableOrbitLibControls(false);
    ui.ButtonSchwNbody->setEnabled(false);
}

void CSmileGUI::OrbitLibraryFinished()
{   // called when either freq.map or schw.orbit library is finished
    enableOrbitLibControls(true);
    ui.ButtonFreqMap->setEnabled(true);
    ui.ButtonFreqMap->setText("Start");
    ui.ButtonSchwOrbitLib->setText("Start");
    ui.tabWidget_sm->setEnabled(true);
    if(core->schwData.size()==0)   //  was not running in the context of creation of Schw.model orbit library
        ui.groupBox_smOptim->setEnabled(false);
    ui.progressBarFM->hide();
    ui.progressBarSM->hide();
    redrawFreqMap(true);
    info(QString::number(core->orbitlib->numComplete())+" orbits, " +
        QString::number(1000.0/timeFM.elapsed()*core->orbitlib->numComplete(),'g',3)+" orbits/second\n" );
    if(core->orbitlib->size()>0 && core->orbitlib->getOrbitDesc(0)->getInfoByType(CBasicInformation::IT_POINCARE)!=NULL) 
    {   // was using Poincare surface in orbit library integration
        const CPoincareInformation* inform;
        for(size_t o=0; o<core->orbitlib->size(); o++) {
            inform=static_cast<const CPoincareInformation*>(
                core->orbitlib->getOrbitDesc(o)->getInfoByType(CBasicInformation::IT_POINCARE));
            if(inform!=NULL)
                addPoincareCurve(inform);
        }
        PlotPoincare->replot();
    }
}

/** Orbit filtering function based on whether an orbit belongs to a given energy shell.
    Energy shells are taken from an existing shell-kinematic Schwarzschild data object.
    if model==NULL or numShell==0, any orbit is acceptable, otherwise 
    only if orbit energy is within numShell-1 energy shell.
    Used in filtering displayed orbits on the frequency map or Schw.model tabs.
    TODO: eliminate the relation to Schw.data object! may use filtering by energy without any model.
*/
class CShellOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    CShellOrbitFilteringFnc(const CSchwDataAngMomDist* _model, const CPotential* _potential, unsigned int _numShell) :
      model(_model), potential(_potential), numShell(_numShell) {};
    virtual double eval(const COrbitDesc* orbit) const; ///< return 1 if the orbit is inside a given energy shell, 0 otherwise
private:
    const CSchwDataAngMomDist* model;  ///< pointer to a shell-kinematic Schwarzschild data object from which the shell index for an orbit is determined
    const CPotential* potential;       ///< pointer to the potential used in computing total energy of an orbit
    unsigned int numShell;             ///< required shell index
};

double CShellOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    if(model==NULL || numShell==0) return 1;
    double E=0;
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL) E=info->getEinit();
    // need to find which energy shell does this orbit belong to
    if(E==0)  // need to figure out orbit energy
    {
        const CPosVelPoint<float>& IC = orbit->getInitCond();
        E = totalEnergy(potential, IC);
    }
    if((numShell>model->numConstraints() && E>model->getShellEnergy(model->numConstraints()-1)) ||
       (E<=model->getShellEnergy(numShell-1) && (numShell==1 || E>model->getShellEnergy(numShell-2)))) 
       return 1;
    else return 0;
}

// return an orbit filtering function which selects orbits belonging to a given energy shell, 
// if a Schw.data object with energy shells was initialized and numShell>0
// may also return shell energy if the last parameter is not NULL
CShellOrbitFilteringFnc shellFilter(const vectorSchwData &schwData, const CPotential* potential, int numShell, double* shellEnergy=NULL)
{
    const CSchwDataAngMomDist* model_sh=NULL;
    for(vectorSchwData::const_iterator iter=schwData.begin(); iter!=schwData.end(); ++iter)
        if((*iter)->dataType() == CBasicSchwData::SD_KINEM_ANGMOMDIST)
        {
            model_sh=static_cast<const CSchwDataAngMomDist*>(*iter);
            if(numShell>0 && shellEnergy!=NULL) *shellEnergy=model_sh->getShellEnergy(numShell-1);
        }
    return CShellOrbitFilteringFnc(model_sh, potential, numShell);
}

void CSmileGUI::redrawFreqMap(bool show)
{
    if(!show || core->orbitlib==NULL) return;
    size_t nFMpoints = core->orbitlib->size();
    if(nFMpoints==0) return;
    int numShell = ui.selectEnergyLevel_fm->value();
    double shellEnergy=12345;
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, core->potential, numShell, &shellEnergy);
    //const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
    if(numShell>0 && shellEnergy!=12345)
        ui.lineEdit_E->setText(QString::number(shellEnergy));
    bool useNonzeroOnly = ui.checkBox_fmConfigSpace->isChecked();
    if(ui.radioButton_fmmapdiff->isChecked() || ui.radioButton_fmmaplyapunov->isChecked())
    {   // draw frequency map itself
        vectord w1w3r(nFMpoints),  w2w3r(nFMpoints),  w1w3c(nFMpoints),  w2w3c(nFMpoints);
        int nr=0, nc=0;
        for(unsigned int i=0; i<nFMpoints; i++)
        {
            bool useOrbit=true;
            useOrbit = filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5 && (!useNonzeroOnly || core->orbitlib->getOrbitDesc(i)->getWeight()>0);
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && useOrbit)
            {
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                bool criterion;
                if(ui.radioButton_fmmapdiff->isChecked()) 
                    criterion = info->getlfccdiff() <= core->configCore.chaoticMinFreqDiff;
                else 
                    criterion = info->getlambda() <= core->configCore.chaoticMinLambda;
                double denom = (core->configOrbit.N_dim==3)?info->getlf(2):1;
                if(denom==0) continue;
                if(criterion)
                {
                    w1w3r[nr] = info->getlf(0)/denom;
                    w2w3r[nr] = info->getlf(1)/denom;
                    nr++;
                }
                else
                {
                    w1w3c[nc] = info->getlf(0)/denom;
                    w2w3c[nc] = info->getlf(1)/denom;
                    nc++;
                }
            }
        }
        CurveFreqMapR->SETDATA(&(w1w3r.front()), &(w2w3r.front()), nr);
        CurveFreqMapC->SETDATA(&(w1w3c.front()), &(w2w3c.front()), nc);
        if(core->configOrbit.N_dim==3)
        {
            PlotFreqMap->setAxisScale(QwtPlot::yLeft, 0.5, 1.1);
            PlotFreqMap->setAxisScale(QwtPlot::xBottom, 0.3, 1.1);
        }
        PlotFreqMap->replot();
        PlotFreqMapHisto->hide();
        PlotFreqMapSS->hide();
        PlotFreqMap->show();

        if(!core->isOrbitLibRunning())
        {
            info(getOrbitPopulation(numShell, useNonzeroOnly));
        }
    }
    else if(ui.radioButton_fmhistodiff->isChecked() || ui.radioButton_fmhistolyapunov->isChecked())
    {   // draw histogram of frequency diffusion rates or Lyapunov exponents
        std::vector<paird> values;
        double totalWeight=0;
        for(unsigned int i=0; i<nFMpoints; i++)
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
            {
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                double weight=(useNonzeroOnly ? core->orbitlib->getOrbitDesc(i)->getWeight() : 0);
                totalWeight+=weight;
                values.push_back(paird(ui.radioButton_fmhistodiff->isChecked() ? info->getlfccdiff() : info->getlambda(), weight));
            }
        std::sort(values.begin(), values.end(), comparepair<double,double>);
        int npoints=static_cast<int>(values.size());
        if(npoints>0)
        {
            vectord val(npoints), np(npoints);
            double minval=1, tw=0;
            for(int i=0; i<npoints; i++)
            {
                val[i]=values[i].first;
                if(totalWeight)
                {
                    np[i]=tw/totalWeight;
                    tw+=values[i].second;
                }
                else
                    np[i]=i*1.0/npoints;
                if((val[i]>0) && (val[i]<minval)) minval=val[i];
            }
            CurveFreqMapHisto->SETDATA(&(val.front()), &(np.front()), npoints);
            PlotFreqMapHisto->setAxisScale(QwtPlot::xBottom, minval/1.5, val[npoints-1]*1.1);
            PlotFreqMapHisto->replot();
        }
        PlotFreqMap->hide();
        PlotFreqMapSS->hide();
        PlotFreqMapHisto->show();
    }
    else if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked() ||
        ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked() ||
        ui.radioButton_fmsydiff->isChecked() || ui.radioButton_fmsylyapunov->isChecked())
    {   // draw start spaces
        vectord Xr(nFMpoints), Yr(nFMpoints), Xc(nFMpoints), Yc(nFMpoints);
        int nr=0, nc=0;
        double rad=0;
        for(unsigned int i=0; i<nFMpoints; i++)
        {
            if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
            {
                bool criterion;
                const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
                if(info==NULL) continue;
                if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsydiff->isChecked()) 
                    criterion = info->getlfccdiff() <= core->configCore.chaoticMinFreqDiff;
                else
                    criterion = info->getlambda() <= core->configCore.chaoticMinLambda;
                const CPosVelPoint<float>& IC=core->orbitlib->getOrbitDesc(i)->getInitCond();
                double X=IC.Pos[0];
                double Y=IC.Pos[1];
                double Z=IC.Pos[2];
                double Xp, Yp;
                if(ui.radioButton_fmsydiff->isChecked() || ui.radioButton_fmsylyapunov->isChecked())
                {
                    Xp=Y;
                    Yp=atan2(IC.Vel[0], IC.Vel[2]);
                }
                else
                    XYZtoXY(X,Y,Z,&Xp,&Yp);
                if(pow_2(Xp)+pow_2(Yp)>pow_2(rad)) rad=sqrt(pow_2(Xp)+pow_2(Yp));
                if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked())
                {
                    if(pow_2(IC.Vel[0])+pow_2(IC.Vel[1])+pow_2(IC.Vel[2])>0) continue;   // this is not a stationary orbit
                }
                else if(ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked())
                {
                    if(X*Y*Z!=0 || (X==0 && Z==0)) continue;   // this is not a principal-plane orbit
                }
                else
                {
                    if(X!=0 || Z!=0) continue;  // this point is not on Y axis
                }
                if(criterion)
                {
                    Xr[nr] = Xp;
                    Yr[nr] = Yp;
                    nr++;
                }
                else
                {
                    Xc[nc] = Xp;
                    Yc[nc] = Yp;
                    nc++;
                }
            }
        }
        CurveFreqMapSSR->SETDATA(&(Xr.front()), &(Yr.front()), nr);
        CurveFreqMapSSC->SETDATA(&(Xc.front()), &(Yc.front()), nc);
        PlotFreqMapSS->setAxisScale(QwtPlot::yLeft, -rad, rad);
        PlotFreqMapSS->setAxisScale(QwtPlot::xBottom, -rad, rad);
        PlotFreqMapSS->replot();
        PlotFreqMap->hide();
        PlotFreqMapHisto->hide();
        PlotFreqMapSS->show();
    }
}

void CSmileGUI::on_selectEnergyLevel_fm_valueChanged(int /*numShell*/)
{
    redrawFreqMap(true);
    //if(numShell>0 && core->model!=NULL && (core->model->ModelType() & CBasicSchwData::SD_DENS_SHELL))  // redundant
    //    ui.lineEdit_E->setText(QString::number(static_cast<CBasicShellSchwModel*>(core->model)->getShellEnergy(numShell-1)));
}

void CSmileGUI::on_checkBox_fmConfigSpace_stateChanged(int)
{
    redrawFreqMap(true);
}

QString CSmileGUI::getOrbitPopulation(int numShell, bool useWeights) const
{
    if(core->orbitlib==NULL) return "Error!";
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, core->potential, numShell);
    const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
    return QString::fromStdString(core->orbitlib->getOrbitPopulation(&filter, &chaos, useWeights));
}

void CSmileGUI::drawFreqMapResonances()
{
    const int NR=18;   // number of resonances depicted on the map
    double res[NR][5] = {
        {0,1,-1 ,0.80,1},
        {1,-1,0 ,0.55,0.55},
        {2,-1,0 ,0.45,0.90},
        {3,-2,0 ,0.34,0.51},
        {4,-3,0 ,0.39,0.52},
        {2,0,-1 ,0.50,1.08},
        {3,0,-2 ,0.66,0.52},
        {1,-3,2 ,0.34,0.78},
        {1,-2,1 ,0.80,0.90},
        {2,-3,1 ,0.33,0.55},
        {4,-2,-1,0.78,1.08},
        {3,-1,-1,0.69,1.08},
        {3,1,-3 ,0.63,1.08},
        {2,1,-2 ,0.74,0.52},
        {2,2,-3 ,0.80,0.70},
        {1,2,-2 ,0.74,0.63},
        {1,1,-1 ,0.47,0.53} };
    double xmin=0.3, xmax=1., ymin=0.4, ymax=1.1;
    double boundrect[4][3] = { {1, 0, -xmin}, {1, 0, -xmax}, {0, 1, -ymin}, {0, 1, -ymax} };
    //QString S;
    for(int i=0; i<NR; i++)
    {
        double x[2], y[2];
        if(res[i][1]==0)
        {
            x[0]=x[1]=-res[i][2]/res[i][0];
            y[0]=ymin;
            y[1]=ymax;
        }
        else if(res[i][0]==0)
        {
            x[0]=xmin;
            x[1]=xmax;
            y[0]=y[1]=-res[i][2]/res[i][1];
        }
        else
        {   // intersect with bounding rectangle
            std::vector<double> ax(0), ay(0);
            for(int c=0; c<4; c++)
            {
                double D = res[i][0]*boundrect[c][1] - res[i][1]*boundrect[c][0];
                ax.push_back( (-res[i][2]*boundrect[c][1] + res[i][1]*boundrect[c][2]) / D);
                ay.push_back(-(-res[i][2]*boundrect[c][0] + res[i][0]*boundrect[c][2]) / D);
            }
            std::sort(ax.begin(), ax.end());
            std::sort(ay.begin(), ay.end());
            int sgn = (res[i][0]*res[i][1]>0)?1:0;
            x[0]=ax[1]; x[1]=ax[2];
            y[0]=ay[1+sgn]; y[1]=ay[2-sgn];
        }
        QwtPlotCurve* c = new QwtPlotCurve();
        c->setPen(QPen(Qt::darkGray, 1, Qt::DotLine));
        c->SETDATA(x, y, 2);
        c->attach(PlotFreqMap);
        QwtPlotMarker* m = new QwtPlotMarker();
        m->setValue(res[i][3], res[i][4]);
        m->setLabel(QString::number(res[i][0])+","+QString::number(res[i][1])+","+QString::number(res[i][2]));
        m->attach(PlotFreqMap);
    }
}

void CSmileGUI::FMclicked(const QPointF &pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    int mode=-1;
    if(ui.radioButton_fmmapdiff->isChecked() || ui.radioButton_fmmaplyapunov->isChecked()) mode=0;
    else if(ui.radioButton_fmssdiff->isChecked() || ui.radioButton_fmsslyapunov->isChecked()) mode=1;
    else if(ui.radioButton_fmspdiff->isChecked() || ui.radioButton_fmsplyapunov->isChecked()) mode=2;
    if(mode<0) return;
    double px=pos.x();
    double py=pos.y();
    double mindist=1; 
    int np=0;
    int numShell = ui.selectEnergyLevel_fm->value();
    //const CBasicShellSchwModel* model_sh = core->model!=NULL && (core->model->ModelType() & CBasicSchwData::SD_DENS_SHELL) ? static_cast<const CBasicShellSchwModel*>(core->model) : NULL;
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, core->potential, numShell);
    for(unsigned int i=0; i<core->orbitlib->size(); i++)
    {
        const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(core->orbitlib->getOrbitDesc(i)->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
        if(info==NULL) continue;
        double dist=1;
        if(mode==0)
        {
            double lf2 = (core->configOrbit.N_dim==3)?info->getlf(2):1;
            double w1w3o = info->getlf(0)/lf2;
            double w2w3o = info->getlf(1)/lf2;
            dist = pow_2(w1w3o-px) + pow_2(w2w3o-py);
        }
        else
        {
            const CPosVelPoint<float>& IC=core->orbitlib->getOrbitDesc(i)->getInitCond();
            double X=IC.Pos[0];
            double Y=IC.Pos[1];
            double Z=IC.Pos[2];
            double Xp, Yp;
            XYZtoXY(X,Y,Z,&Xp,&Yp);
            if((mode==1 && X*Y*Z==0) || (mode==2 && X*Y*Z!=0)) continue; 
            dist = pow_2(Xp-px) + pow_2(Yp-py);
        }
        if(dist<mindist && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            np=i;
            mindist=dist;
        }
    }
    if(mindist>=1) return;  // too far
    // load data from np'th orbit
    core->initCond=core->orbitlib->getOrbitDesc(np)->getInitCond();
    // set exact values
    bool tmpUseE=core->useICe;
    core->useICe=false;  // need to put in values for IC, not for E
    core->initIC();
    core->useICe=tmpUseE;
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->timeUnit));
    InternalRecalc=false;
    info(QString::fromStdString(core->orbitlib->getOrbitDesc(np)->toString()));
}

void CSmileGUI::XYZtoXY(const double X, const double Y, const double Z, double* x, double* y)
{
    *x = (Y-X)/sqrt(2.0);
    *y = (2*Z-X-Y)/sqrt(6.0);
}

/* ----------------------------  Schwarzschild modelling --------- */
void CSmileGUI::on_ButtonSchwOrbitLib_clicked()
{
    if(core->isExportNbodyRunning())  // don't interfere with exporting to Nbody
        return;
    if(core->isSchwOptRunning())  // stop? noooo!
    {
        error("Cannot stop optimization routine");
        return;
    }
    if(core->isOrbitLibRunning())
    {
        emit stopOrbitLibrary();
        return;
    }
    if(core->configOrbit.N_dim!=3)
    {
        error("Schwarzschild modelling is only possible in 3 dimensions");
        return;
    }

    // start building library
    core->buildSchwIC(ui.comboBox_smICGenerator->currentIndex()>=3);  // 3 first elements are different kinds of IC generators, 4th is "use existing IC"
    if(core->orbitlib==NULL || core->orbitlib->size()==0)  // was unsuccessful, get out of here
        return;
    core->startOrbitLibrary(SLOT(coreOrbitLibraryFinished()));
    enableOrbitLibControls(false);
    ui.ButtonFreqMap->setEnabled(false);  // can't run freq.map while integrating orbits for Schw/model
    ui.ButtonSchwOrbitLib->setText("Stop");
    timeFM.start();
    ui.progressBarSM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarSM->setValue(0);
    ui.progressBarSM->show();
}

void CSmileGUI::NbodyReintegrationFinished()    // slot called when core->signalNbodyReintegrationFinished() emitted
{
    ui.progressBarSM->hide();
    info("Reintegration finished, exporting Nbody file");
}

void CSmileGUI::NbodyExportFinished(const QString& result)    // slot called when core->signalschwExportNbodyFinished() emitted
{
    ui.ButtonSchwNbody->setText("Create Nbody IC");
    enableOrbitLibControls(true);
    ui.progressBarSM->hide();
    info(result);
}

void CSmileGUI::on_ButtonSchwLP_clicked()
{
    if(core->isExportNbodyRunning() || core->isOrbitLibRunning() || 
        core->isSchwOptRunning() || core->schwData.size()==0) return;
    enableOrbitLibControls(false);
    core->schwStartOptimization(CBasicSchwModel::SM_LINEAR);
}

void CSmileGUI::on_ButtonSchwQP_clicked()
{
    if(core->isExportNbodyRunning() || core->isOrbitLibRunning() || 
        core->isSchwOptRunning() || core->schwData.size()==0) return;
    enableOrbitLibControls(false);
    core->schwStartOptimization(CBasicSchwModel::SM_QUADRATIC);
}

void CSmileGUI::SchwOptimizationFinished(const QString& message)
{
    enableOrbitLibControls(true);
    ui.ButtonSchwNbody->setEnabled(true);
    SchwModelResult = message;
    redrawSchwModel(true);
}

void CSmileGUI::on_selectEnergyLevel_sm_valueChanged(int)
{
    redrawSchwModel(true);
}

void CSmileGUI::redrawSchwModel(bool updateInfo)
{
    if(/*core->schwData.size()==0 ||*/ !core->orbitlib)
        return;
    int numShell=ui.selectEnergyLevel_sm->value();
    size_t nstart=0, nend=0;      // range of cells to display
    vectorSchwData tmpSchwData;   // temporary vector for keeping only known Schw.density data objects
    // init one or more pointers to specific classes of Schw.data objects
    const CSchwDataKinemShell*          model_sh=NULL;
    const CSchwDataAngMomDist*          model_am=NULL;
    const CSchwDataDensGridClassic*     model_cl=NULL;
    const CSchwDataDensGridCylindrical* model_cyl=NULL;
    const CSchwDataDensSH*              model_sph=NULL;
    for(vectorSchwData::iterator iter=core->schwData.begin(); iter!=core->schwData.end(); iter++)
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_KINEM_SHELL:
                model_sh=static_cast<const CSchwDataKinemShell*>(*iter); 
                break;
            case CBasicSchwData::SD_KINEM_ANGMOMDIST:
                model_am=static_cast<const CSchwDataAngMomDist*>(*iter); 
                break;
            case CBasicSchwData::SD_DENS_GRIDCLASSIC:
                model_cl=static_cast<const CSchwDataDensGridClassic*>(*iter);
                nstart=numShell==0 ? 0 : (numShell-1)*model_cl->getCellsPerShell();
                nend=numShell==0 ? (*iter)->numConstraints() : numShell*model_cl->getCellsPerShell();
                tmpSchwData.push_back(model_cl);
                break;
            case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL:
                model_cyl=static_cast<const CSchwDataDensGridCylindrical*>(*iter);
                nend=(*iter)->numConstraints();
                tmpSchwData.push_back(model_cyl);
                break;
            case CBasicSchwData::SD_DENS_SHMESH:
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                nend=(*iter)->numConstraints();
                tmpSchwData.push_back(model_sph);
                break;
            default: ;
        }
    if(model_cl) {
        ui.selectEnergyLevel_sm->setEnabled(true);
    } else {
        ui.selectEnergyLevel_sm->setEnabled(true);
        numShell=0;
    }
    //if(numShell>0 && model_sh!=NULL)
    //    ui.lineEdit_E->setText(QString::number(model_sh->getShellEnergy(numShell-1)));
    if(ui.radioButton_smviewgrid->isChecked() && (model_cl!=NULL||model_cyl!=NULL||model_sph!=NULL))
    {
        vectord CellsFx, CellsFy, CellsIx, CellsIy;
        // create a temporary instance of Schw Model object to compute statistics
        CSchwModelQuadOpt model(core->orbitlib, tmpSchwData, NULL, NULL, 0, 0, 0, 0);
        double Xmin=0, Xmax=0, Ymin=0, Ymax=0;
        int nn=-1; double nnprev=-1;
        for(size_t c=nstart; c<nend; c++)
        {
            double Xp=0, Yp=0;
            if(model_cl)
            {
                double X=0,Y=0,Z=0;
                model_cl->cellCenter(c, &X, &Y, &Z);
                XYZtoXY(X,Y,Z,&Xp,&Yp);
                if(fabs(Xp)>Xmax) { Xmax=fabs(Xp); Xmin=-Xmax; }
                if(Yp>Ymax) { Ymax=Yp; }
                if(Yp<Ymin) { Ymin=Yp; }
            } 
            else if(model_cyl)
            {
                size_t indR, indz, indphi;
                model_cyl->getCoefDecomposition(c, &indR, &indz, &indphi);
                Xp=indR + indphi/(1.0+std::max<size_t>(core->configCore.sm_numAngularCoefs,1));
                Yp=indz;
                Xmax=std::max<double>(Xmax,Xp);
                Ymax=std::max<double>(Ymax,Yp);
            }
            else if(model_sph)
            {
                double n;
                int l, m;
                model_sph->getCoefDecomposition(c, &n, &l, &m);
                l/=2; m/=2;
                if(n!=nnprev) { nnprev=n; nn++; }
                Xp=l*(l+1)/2 + m*(1.-1./(l+4));  // put the points in row but make somewhat larger spacing between l and l+1 series
                Yp=nn;
                Xmax=std::max<double>(Xmax,Xp);
                Ymax=std::max<double>(Ymax,Yp);
            }
            size_t no, noz;
            double dev;
            model.calcCoefDeviation(c, &dev, &no, &noz);
            if(dev!=0)
            {
                CellsIx.push_back(Xp);
                CellsIy.push_back(Yp);
            }
            else
            {
                CellsFx.push_back(Xp);
                CellsFy.push_back(Yp);
            }
        }
        if(CellsFx.empty())
            CurveSchwCellF->SETDATA(NULL, NULL, 0);
        else
            CurveSchwCellF->SETDATA(&(CellsFx.front()), &(CellsFy.front()), static_cast<int>(CellsFx.size()));
        if(CellsIx.empty())
            CurveSchwCellI->SETDATA(NULL, NULL, 0);
        else
            CurveSchwCellI->SETDATA(&(CellsIx.front()), &(CellsIy.front()), static_cast<int>(CellsIx.size()));
        PlotSchw->setAxisScale(QwtPlot::xBottom, Xmin, Xmax);
        PlotSchw->setAxisScale(QwtPlot::yLeft, Ymin, Ymax);
        PlotSchw->replot();
        PlotSchwHisto->hide();
        PlotSchw->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
    else if(ui.radioButton_smviewbeta->isChecked())
    {   // plot anisotropy coefficient beta
        if(model_sh)
        {
            size_t numKinemShells=model_sh->numConstraints();
            vectord val(numKinemShells), np(numKinemShells);
            for(size_t s=0; s<numKinemShells; s++)
            {
                double vr2=0, vt2=0;
                for(size_t o=0; o<core->orbitlib->size(); o++)
                {
                    const COrbitDesc* orb=core->orbitlib->getOrbitDesc(o);
                    if(orb->getState()==COrbitDesc::OS_DONE && orb->getWeight()>0)
                    {
                        // load kinematic information data from orbit desc
                        const CSchwInformation* infoSchwKinem=NULL;
                        for(size_t inf=0; inf<orb->getInfoNum(); inf++)
                        {
                            const CBasicInformation* info=orb->getInfo(inf);
                            if(info->infoType()==CBasicInformation::IT_SCHW && 
                                static_cast<const CSchwInformation*>(info)->schwDataType() == CBasicSchwData::SD_KINEM_SHELL) 
                                infoSchwKinem=static_cast<const CSchwInformation*>(info);
                        }
                        if(infoSchwKinem!=NULL && infoSchwKinem->getSchwData().size()==numKinemShells*3)
                        {
                            vr2 += infoSchwKinem->getSchwData(s + numKinemShells) * infoSchwKinem->getSchwData(s) * orb->getWeight();
                            vt2 += infoSchwKinem->getSchwData(s+2*numKinemShells) * infoSchwKinem->getSchwData(s) * orb->getWeight();
                        }
                    }
                }
                if(vr2>0)
                    val[s] = 1-vt2/vr2/2;
                else
                    val[s] = 1;
                np[s]=s*1.0;
            }
            if(numKinemShells>0)
                CurveSchwHistoR->SETDATA(&(np.front()), &(val.front()), static_cast<int>(numKinemShells));
            else
                CurveSchwHistoR->SETDATA(NULL, NULL, 0);
            CurveSchwHistoR->setStyle(QwtPlotCurve::Steps);
            CurveSchwHistoC->SETDATA(NULL, NULL, 0);
        } else {
            CurveSchwHistoR->SETDATA(NULL, NULL, 0);
            CurveSchwHistoC->SETDATA(NULL, NULL, 0);
        }
        PlotSchwHisto->setAxisScale(QwtPlot::yLeft, -1, 1);
        PlotSchwHisto->setAxisAutoScale(QwtPlot::xBottom);
        PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLinearScaleEngine);
        PlotSchwHisto->replot();
        PlotSchw->hide();
        PlotSchwHisto->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
    else if(ui.radioButton_smviewweights->isChecked() || ui.radioButton_smviewhisto->isChecked())
    {   // plot histogram of orbit weights
        vectord valuesR, valuesC, npR, npC;
        double minval=1, maxval=0;
        const CShellOrbitFilteringFnc filter(model_am, core->potential, numShell);
        const CChaosOrbitFilteringFnc chaos(1, core->configCore.chaoticMinLambda, core->configCore.chaoticMinFreqDiff);
        for(size_t o=0; o<core->orbitlib->size(); o++)
        {
            const COrbitDesc* orb=core->orbitlib->getOrbitDesc(o);
            if(orb->getState()==COrbitDesc::OS_DONE && filter.eval(orb)>=0.5)
            {
                if(chaos.eval(orb)>=0.5 && !ui.radioButton_smviewhisto->isChecked())
                {
                    valuesC.push_back(orb->getWeight());
                    npC.push_back(o*1.0);
                }
                else
                {
                    valuesR.push_back(orb->getWeight());
                    npR.push_back(o*1.0);
                }
                if(orb->getWeight()>0) minval = std::min<double>(minval, orb->getWeight());
                maxval = std::max<double>(maxval, orb->getWeight());
            }
        }
        if(ui.radioButton_smviewhisto->isChecked()) 
        {
            std::sort(valuesR.begin(), valuesR.end());
            for(size_t i=1; i<valuesR.size(); i++) valuesR[i]+=valuesR[i-1];
            for(size_t i=0; i<valuesR.size(); i++) valuesR[i]/=valuesR.back();
            minval=0; maxval=1;
            CurveSchwHistoC->SETDATA(NULL, NULL, 0);
            PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLinearScaleEngine);
        }
        else
        {
            if(npC.size()==0)
                CurveSchwHistoC->SETDATA(NULL, NULL, 0);
            else
                CurveSchwHistoC->SETDATA(&(npC.front()), &(valuesC.front()), static_cast<int>(valuesC.size()));
            PlotSchwHisto->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
        }
        if(npR.size()==0)
            CurveSchwHistoR->SETDATA(NULL, NULL, 0);
        else
            CurveSchwHistoR->SETDATA(&(npR.front()), &(valuesR.front()), static_cast<int>(valuesR.size()));
        CurveSchwHistoR->setStyle(ui.radioButton_smviewhisto->isChecked() ? QwtPlotCurve::Steps : QwtPlotCurve::Dots);
        PlotSchwHisto->setAxisScale(QwtPlot::yLeft, minval, maxval);
        //PlotSchwHisto->setAxisAutoScale(QwtPlot::xBottom);
        PlotSchwHisto->setAxisScale(QwtPlot::xBottom, 0, core->orbitlib->size());
        PlotSchwHisto->replot();
        PlotSchw->hide();
        PlotSchwHisto->show();
        if(updateInfo)  info(SchwModelResult + "\n" + 
            getOrbitPopulation(numShell, true));
    }
}

void CSmileGUI::SMclicked(const QPointF& pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    double px=pos.x();
    double py=pos.y();
    double mindist=10000; 
    size_t numCoefs=0;
    // init one or more pointers to specific classes of Schw.data objects
    const CSchwDataDensGridClassic*     model_cl=NULL;
    const CSchwDataDensGridCylindrical* model_cyl=NULL;
    const CSchwDataDensSH*              model_sph=NULL;
    vectorSchwData tmpSchwData;         // temporary vector for keeping only known Schw.density data objects
    for(vectorSchwData::iterator iter=core->schwData.begin(); iter!=core->schwData.end(); iter++)
        switch((*iter)->dataType()) {
            case CBasicSchwData::SD_DENS_GRIDCLASSIC:
                model_cl=static_cast<const CSchwDataDensGridClassic*>(*iter);
                numCoefs=model_cl->numConstraints();
                tmpSchwData.push_back(model_cl);
                break;
            case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL:
                model_cyl=static_cast<const CSchwDataDensGridCylindrical*>(*iter);
                numCoefs=model_cyl->numConstraints();
                tmpSchwData.push_back(model_cyl);
                break;
            case CBasicSchwData::SD_DENS_SHMESH:
            case CBasicSchwData::SD_DENS_SHBSE:
                model_sph=static_cast<const CSchwDataDensSH*>(*iter); 
                numCoefs=model_sph->numConstraints();
                tmpSchwData.push_back(model_sph);
                break;
            default: ;
        }
    if(tmpSchwData.size()!=1) return;   // must have one and only one known density data type object
    int nc=-1, nn=-1; double nnprev=-1;
    std::string decomp, decomp1;
    for(size_t c=0; c<numCoefs; c++)
    {
        double Xp=0, Yp=0;
        if(model_cl)
        {
            double X=0,Y=0,Z=0;
            model_cl->cellCenter(c, &X, &Y, &Z);
            XYZtoXY(X,Y,Z,&Xp,&Yp);
            decomp1="X="+convertToString(X)+",Y="+convertToString(Y)+",Z="+convertToString(Z);
        }
        else if(model_cyl)
        {
            size_t indR, indz, indphi;
            model_cyl->getCoefDecomposition(c, &indR, &indz, &indphi);
            Xp=indR + indphi/(1.0+std::max<size_t>(core->configCore.sm_numAngularCoefs,1));
            Yp=indz;
            double X=0,Y=0,Z=0;
            model_cyl->cellCenter(c, &X, &Y, &Z);
            decomp1="X="+convertToString(X)+",Y="+convertToString(Y)+",Z="+convertToString(Z);
        }
        else if(model_sph)
        {
            double n;
            int l, m;
            model_sph->getCoefDecomposition(c, &n, &l, &m);
            decomp1="n="+convertToString(n)+",l="+convertToString(l)+",m="+convertToString(m);
            l/=2; m/=2;
            if(n!=nnprev) { nnprev=n; nn++; }
            Xp=l*(l+1)/2 + m*(1.-1./(l+4));  // put the points in row but make somewhat larger spacing between l and l+1 series
            Yp=nn;
        }
        double dist = pow_2(Xp-px) + pow_2(Yp-py);
        if(dist<mindist)
        {
            nc=static_cast<int>(c);
            mindist=dist;
            decomp=decomp1;
        }
    }
    if(nc<0) return;  // nc= index of cell/coefficient that was clicked
    // create a temporary instance of Schw Model object to compute penalty (deviation of a coef from required value)
    CSchwModelQuadOpt model(core->orbitlib, tmpSchwData, NULL, NULL, 0, 0, 0, 0);
    double coefValue  = tmpSchwData[0]->getConstraint(nc);
    double normFactor = tmpSchwData[0]->getConstraintNormFactor(nc);
    double coefDeviation=0;
    size_t numOrb, numOrbNZ;
    model.calcCoefDeviation(nc, &coefDeviation, &numOrb, &numOrbNZ);
    if(coefDeviation!=0) coefDeviation=coefDeviation/coefValue*100;  // relative deviation in percents
    info("Constraint index="+QString::number(nc) + "\n" + QString::fromStdString(decomp) + 
        "\nRequired value="+QString::number(coefValue) + 
        "\nDeviation="+QString::number(coefDeviation) + "%\nNormalization="+QString::number(normFactor) +
        "\n"+QString::number(numOrb)+" orbits ("+QString::number(numOrbNZ)+
        " with non-zero weight) contribute to this constraint");
}

void CSmileGUI::SMHistoClicked(const QPointF &pos)
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || !ui.radioButton_smviewweights->isChecked())
        return;
    int numShell = ui.selectEnergyLevel_fm->value();
    const CShellOrbitFilteringFnc filter=shellFilter(core->schwData, core->potential, numShell);
    double px=pos.x();
    double py=pos.y();
    // scale clicked value also to [0..1] interval
    double minyval=1, maxyval=0;
    size_t maxxval=core->orbitlib->size(), minxval=0;
    for(size_t i=0; i<core->orbitlib->size(); i++)
        if(core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            //if(i<minxval) minxval=i;
            //if(i>maxxval) maxxval=i;
            float weight=core->orbitlib->getOrbitDesc(i)->getWeight();
            if(weight>0)
            {
                minyval=std::min<double>(minyval, weight);
                maxyval=std::max<double>(minyval, weight);
            }
        }
    if(maxxval==0 || maxyval==0) return;
    minyval=log10(minyval);
    maxyval=log10(maxyval);
    px = (px-minxval)*1.0/(maxxval-minxval);
    py = (log10(py)-minyval)/(maxyval-minyval);
    // find nearest point
    double mindist=1; 
    size_t np=0;
    for(size_t i=0; i<core->orbitlib->size(); i++)
    {
        double dist=1;
        if(core->orbitlib->getOrbitDesc(i)->getWeight()>0 && core->orbitlib->getOrbitDesc(i)->getState()==COrbitDesc::OS_DONE && filter.eval(core->orbitlib->getOrbitDesc(i))>=0.5)
        {
            double ox=(i-minxval)*1.0/(maxxval-minxval);
            double oy=(log10(core->orbitlib->getOrbitDesc(i)->getWeight())-minyval)/(maxyval-minyval);
            dist = pow_2(ox-px) + pow_2(oy-py);
        }
        if(dist<mindist)
        {
            np=i;
            mindist=dist;
        }
    }
    if(mindist>=1) return;  // too far
    // load data from np'th orbit
    core->initCond=core->orbitlib->getOrbitDesc(np)->getInitCond();
    // set exact values
    bool tmpUseE=core->useICe;
    core->useICe=false;  // need to put in values for IC, not for E
    core->initIC();
    core->useICe=tmpUseE;
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->timeUnit));
    InternalRecalc=false;
    info(QString::fromStdString(core->orbitlib->getOrbitDesc(np)->toString()));
}

/*   -------------  export-import ------------- */
void CSmileGUI::on_ButtonSchwNbody_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    if(core->isExportNbodyRunning())
    {   // stop export (determine which stage we are at now)
        if(core->isOrbitLibRunning()) 
        {   // reintegration stage - abort
            emit stopOrbitLibrary();
            return;
        } else {  // perhaps still computing virial ratio?
            emit stopExportNbody();  // if not, this will be discarded
            return;
        }
    }
    if(core->isOrbitLibRunning() || core->isSchwOptRunning())  // can't export while performing regular orbit library integration
        return;
    bool ok;
    int NPoints=QInputDialog::GETINTEGER(this, "Export to N-body", "Enter number of points", 100000, 0, 1e8, 1, &ok);
    if(!ok) return;
    QString fileFormats, selectedFormat;
    for(size_t i=0; i<formatsIOSnapshot.size(); i++) {
        fileFormats += (fileFormats.length()>0 ? ";;" : "") + QString(formatsIOSnapshot[i].c_str()) + " (*)";
        if(i==0) selectedFormat=fileFormats;
    }
    QString fileName = QFileDialog::getSaveFileName(this, "Export to N-body", 
        core->WorkDir, fileFormats, &selectedFormat, QFileDialog::HideNameFilterDetails | QFileDialog::DontUseNativeDialog);
    selectedFormat.replace(" (*)", "");
    if(fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    int refineFactor=QInputDialog::GETINTEGER(this, "Export to N-body", "Mass refine factor (2^N)", 0, 0, 6, 1, &ok);
    if(!ok) refineFactor=0;

    ui.ButtonSchwNbody->setText("Cancel");
    enableOrbitLibControls(false);
    ui.progressBarSM->setValue(0);
    ui.progressBarSM->setRange(0, static_cast<int>(core->orbitlib->size()));
    ui.progressBarSM->show();   // to display progress in orbit library reintegration, if it is necessary
    core->schwExportNbody(NPoints, fileName, selectedFormat, refineFactor);
}

void CSmileGUI::on_ButtonExportFM_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Frequency map", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);

    if(!core->exportOrbitLib(fileName, false))
        error("Cannot write file "+fileName);
    core->saveSettings(fileName.toStdString()+".ini");
}

void CSmileGUI::on_ButtonImportFM_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import Frequency map", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(QFile::exists(fileName+".ini"))
        loadApplySettings(fileName+".ini");
    if (!core->importOrbitLib(fileName, false)) {
        error("Cannot read file "+fileName);
        return;
    }
    redrawFreqMap(true);
}

void CSmileGUI::on_ButtonExportSM_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Schwarzschild model", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);

    if(!core->exportOrbitLib(fileName, true))
        error("Cannot write file "+fileName);
    core->saveSettings(fileName.toStdString()+".ini");
}

void CSmileGUI::on_ButtonImportSM_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import Schwarzschild model", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(QFile::exists(fileName+".ini"))
        loadApplySettings(fileName+".ini");
    else
        info("Warning, no "+fileName+".ini file found");  // normally each model should have its parameters stored in ini file
    if(core->configOrbit.N_dim!=N_DIM)
    {
        error("N_dim should be equal to 3 for Schwarzschild modelling");
        return;
    }
    if(!core->importOrbitLib(fileName, true)) 
    {
        error("Cannot read file "+fileName);
        return;
    }
    ui.groupBox_smGrid->setEnabled(true);
    ui.groupBox_smOptim->setEnabled(true);
    ui.ButtonSchwNbody->setEnabled(true);
    SchwModelResult=ui.label_info->toPlainText();
    redrawFreqMap(true);
    on_selectEnergyLevel_sm_valueChanged(ui.selectEnergyLevel_sm->value());  // redraw grid
}

void CSmileGUI::on_ButtonExportOrbit_clicked()
{
    if(!core->orbit || !core->orbit->getIntTime())
        return;
#ifdef USE3DFACETS
    if(ui.radioButton_3dmesh->isChecked())
    {   // export mesh, not the trajectory
        QString fileName = QFileDialog::getSaveFileName(this, "Export 3d orbit shape to Wavefront OBJ file", core->WorkDir, "*.obj");
        if (fileName.isEmpty())
            return;
        core->WorkDir = getDirName(fileName);
        // export mesh to Wavefront OBJ text file, to render in an external application (e.g. Blender or ParaView)
        std::ofstream strm(fileName.toStdString().c_str(), std::ios_base::out);
        for(size_t t=0; t<TriangThr->trjpos.size(); t++)
            strm << "v " << TriangThr->trjpos[t].x << " " << TriangThr->trjpos[t].y << " " << TriangThr->trjpos[t].z << std::endl;
        for(size_t c=0; c<TriangThr->facets.size(); c++)
            strm << "f " << TriangThr->facets[c][0] <<" " << TriangThr->facets[c][1] <<" " << TriangThr->facets[c][2]<< std::endl; 
        if(!strm) error("Cannot write file "+fileName);
        return;
    }
#endif

    QString fileName = QFileDialog::getSaveFileName(this, "Export orbit", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->exportOrbit(fileName))
        error("Cannot write file "+fileName);
}

void CSmileGUI::on_ButtonImportOrbit_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Import orbit", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->importOrbit(fileName)) {
        error("Cannot read file "+fileName);
        return;
    }
    InternalRecalc=true;
    ui.lineEdit_x->setText(QString::number(core->initCond.Pos[0]));
    ui.lineEdit_y->setText(QString::number(core->initCond.Pos[1]));
    ui.lineEdit_z->setText(QString::number(core->initCond.Pos[2]));
    ui.lineEdit_vx->setText(QString::number(core->initCond.Vel[0]));
    ui.lineEdit_vy->setText(QString::number(core->initCond.Vel[1]));
    ui.lineEdit_vz->setText(QString::number(core->initCond.Vel[2]));
    ui.lineEdit_E->setText(QString::number(core->initCondE));
    ui.lineEdit_Torb->setText(QString::number(core->timeUnit));
    ui.lineEdit_intTime->setText(QString::number(core->configCore.intTimeInPeriods));
    ui.lineEdit_timeStep->setText(QString::number(core->configOrbit.samplingInterval));
    ui.spinBox_intervalCount->setValue(1);
    InternalRecalc=false;
    refreshOrbitAndSpectrum();
}

void CSmileGUI::on_ButtonExportModelStats_clicked()
{
    if(core->orbitlib==NULL || core->orbitlib->size()==0 || core->schwData.size()==0)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Schwarzschild model statistics", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->exportSchwModelStats(fileName))
        error("Cannot write file "+fileName);
}

void CSmileGUI::on_ButtonExportPotential_clicked()
{
    int indComp=ui.comboBox_ptComponents->currentIndex();
    QString fileName = QFileDialog::getSaveFileName(this, QString("Export potential/density profile")+
        QString::fromStdString(indComp>0?" for component "+convertToString(indComp-1):""), core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(!core->exportPotential(fileName, indComp))
        error("Cannot write file "+fileName);
}

void CSmileGUI::on_ButtonExportPS_clicked()
{
#if QWT_VERSION < 0x060000
    if(CurvesPoincare.size()<=1) return;
    QString fileName = QFileDialog::getSaveFileName(this, "Export Poincare surface of section", core->WorkDir);
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    std::ofstream strm(fileName.toStdString().c_str(), std::ios::out);
    if(!strm) { error("Cannot write file "+fileName); return; }
    std::vector< paird > curve;
    for(int c=0; c<CurvesPoincare.size(); c++)
    {
        curve.resize(CurvesPoincare[c]->data().size());
        for(size_t i=0; i<CurvesPoincare[c]->data().size(); i++)
        {
            curve[i].first =CurvesPoincare[c]->data().x(i);
            curve[i].second=CurvesPoincare[c]->data().y(i);
        }
#if 0  // unused
        // sort points in curve in nearest-neighbour order
        for(size_t i=0; i<curve.size()-1; i++)
        {
            double distmin=1e10; 
            size_t indmin=0;
            // find next unused point closest to i-th 
            for(size_t j=i+1; j<curve.size(); j++)
            {
                double dist=pow_2(curve[i].first-curve[j].first)+pow_2(curve[i].second-curve[j].second);
                if(dist<distmin) { distmin=dist; indmin=j; }
            }
            if(indmin>0)
            {
                paird tmp(curve[indmin]);
                curve[indmin]=curve[i+1];
                curve[i+1]=tmp;
            }
        }
#endif
        for(size_t i=0; i<curve.size()-1; i++)
            strm << curve[i].first << "\t" << curve[i].second << "\n";
        strm << "\n";
    }
#endif
}

void CSmileGUI::on_ButtonPrint_clicked()
{
    QwtPlot* fig = NULL; 
    if(PlotOrbit->isVisible())
        fig = PlotOrbit;
    if(PlotPoincare->isVisible())
        fig = PlotPoincare;
    if(PlotFreq->isVisible())
        fig = PlotFreq;
    if(PlotLyapunov->isVisible())
        fig = PlotLyapunov;
    if(PlotFreqMap->isVisible())
        fig = PlotFreqMap;
    if(PlotFreqMapHisto->isVisible())
        fig = PlotFreqMapHisto;
    if(PlotFreqMapSS->isVisible())
        fig = PlotFreqMapSS;
    if(PlotSchw->isVisible())
        fig = PlotSchw;
    if(PlotSchwHisto->isVisible())
        fig = PlotSchwHisto;
#ifdef HAVE_QWT3D
    bool Plot3dVisible=Plot3d->isVisible();
    // 3d plot is not inherited from QwtPlot, so treat it differently
#else
    bool Plot3dVisible=false;
#endif
    if(fig==NULL && !Plot3dVisible)
        return;
    QString fileName = QFileDialog::getSaveFileName(this, "Print figure to file", core->WorkDir, "PostScript (*.ps);;PDF (*.pdf)");
    if (fileName.isEmpty())
        return;
    core->WorkDir = getDirName(fileName);
    if(fig)
    {
        QPrinter P;
        P.setOutputFileName(fileName);
#if QWT_VERSION >= 0x060000
        QwtPlotRenderer ren;
        ren.renderTo(fig, P);
#else
        fig->print(P);
#endif
    }
#ifdef HAVE_QWT3D
    else if(Plot3dVisible)
    {
        QString Format=fileName.endsWith(".pdf", Qt::CaseInsensitive) ? "PDF" : "PS";
        Qwt3D::VectorWriter* printer = (Qwt3D::VectorWriter*)Qwt3D::IO::outputHandler(Format);
        printer->setTextMode(Qwt3D::VectorWriter::NATIVE);
        printer->setSortMode(Qwt3D::VectorWriter::BSPSORT);
        printer->setLandscape(Qwt3D::VectorWriter::OFF);
        Qwt3D::IO::save(Plot3d, fileName, Format);
    }
#endif
}

/* ------------- Triangulation thread helper class -------------- */
#ifdef USE3DFACETS
CTriangThread::CTriangThread(const QString& _appPath) : appPath(_appPath)
{
    finish = restart = false;
}

CTriangThread::~CTriangThread()
{
    mutex.lock();
    finish=restart=true;
    condition.wakeOne();
    mutex.unlock();
}
 
void CTriangThread::run()
{
    while(!finish)
    {
        createTriangulation();
        mutex.lock();
        if(!restart)
            condition.wait(&mutex);
        restart=false;
        mutex.unlock();
    }
}

void CTriangThread::startTriangulation(const std::vector<CPosPoint<double> > &_trajp, double _maxdist, int _slice)
{
    QMutexLocker locker(&mutex);
    // store data 
    nPointsMy = static_cast<int>(_trajp.size());
    trajpMy=_trajp;
    sliceMy = _slice;
    maxdistMy = _maxdist;
    // start or restart thread
    if(!isRunning())
        start(QThread::LowPriority);
    else
    { 
        restart=true;
        condition.wakeOne();
    }
}

bool CTriangThread::toolong(int v1, int v2)
{
    return (pow_2(trjpos[v1].x-trjpos[v2].x) + pow_2(trjpos[v1].y-trjpos[v2].y) + pow_2(trjpos[v1].z-trjpos[v2].z) > pow_2(maxdist));
}
void CTriangThread::addface(int v1, int v2, int v3)
{
    CTriangle T(v1, v2, v3);
    if(Triangles.remove(T)) 
        T.dup=true;
    Triangles << T;
}

void CTriangThread::createTriangulation()
{
    {
        Triangles.clear();
        // before proceeding further, copy data to local variables
        mutex.lock();
        facets.clear();
        trjpos.clear();
        maxdist=maxdistMy;
        int coord = sliceMy-1;
        int nPoints=0;
        for(int i=0; i<nPointsMy; i++)
        {
            if((sliceMy==0) || (trajpMy[i].Pos[coord]>=0))
            {
                trjpos.push_back(Qwt3D::Triple(trajpMy[i].Pos[0], trajpMy[i].Pos[1], trajpMy[i].Pos[2]));
                nPoints++;
            }
        }
        mutex.unlock();

        QProcess process;
        process.setReadChannel(QProcess::StandardOutput);
        process.start(appPath, QStringList("i"));
        if(!process.waitForStarted()) return;
        QTextStream strmI(&process);
        strmI << "3" << endl << nPoints << endl;
        for(int i=0; i<nPoints; i++)
        {
            strmI << trjpos[i].x << " " << trjpos[i].y << " " << trjpos[i].z << endl;
            if(restart) return;
        }
        process.closeWriteChannel();
        int nttot=0, nt=0;
        if(process.waitForFinished())
        {
            QTextStream strmO(&process);
            QString ln = strmO.readLine();
            nttot = ln.toInt();
            while(!strmO.atEnd())
            {
                ln = strmO.readLine();
                QStringList abcd=ln.split(" ");
                if(abcd.size() >= 4)
                {
                    int v1=abcd[0].toInt();
                    int v2=abcd[1].toInt();
                    int v3=abcd[2].toInt();
                    int v4=abcd[3].toInt();
                    if((v1>=nPoints) || (v2>=nPoints) || (v3>=nPoints) || (v4>=nPoints)) continue;
                    if(!(toolong(v1,v2) || toolong(v1,v3) || toolong(v1,v4) ||
                         toolong(v2,v3) || toolong(v2,v4) || toolong(v3,v4)))
                    {
                        addface(v1, v2, v3);
                        addface(v1, v2, v4);
                        addface(v1, v3, v4);
                        addface(v2, v3, v4);
                        nt++;
                    }
                }
                if(restart) return;
            }
        }
        // update faces
        int ntot=0, nout=0;
        QList<CTriangle> TList = Triangles.toList();
        Qwt3D::Cell tricell(3);
        foreach (CTriangle T, TList)
        {
            ntot++;
            if(!T.dup)
            {
                tricell[0] = T.c1;
                tricell[1] = T.c2;
                tricell[2] = T.c3;
                facets.push_back(tricell);
                nout++;
            }
            if(restart) return;
        }
        emit finished(true);
    }
}

#endif

QwtText QwtPlotPicker2::trackerText( const QPointF &pos ) const
{
    QString text;

    switch ( rubberBand() )
    {
        case HLineRubberBand:
            text.sprintf( "%.4g", pos.y() );
            break;
        case VLineRubberBand:
            text.sprintf( "%.4g", pos.x() );
            break;
        default:
            text.sprintf( "%.4g, %.4g", pos.x(), pos.y() );
    }
    return QwtText( text );
}

QwtText QwtPlotZoomer2::trackerText( const QPointF &pos ) const
{
    QString text;

    switch ( rubberBand() )
    {
        case HLineRubberBand:
            text.sprintf( "%.4g", pos.y() );
            break;
        case VLineRubberBand:
            text.sprintf( "%.4g", pos.x() );
            break;
        default:
            text.sprintf( "%.4g, %.4g", pos.x(), pos.y() );
    }
    return QwtText( text );
}

}  // namespace
