#include "icgenerator.h"
#include <algorithm>
#include <set>
#include <cassert>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_monte_vegas.h>
#include "potential.h"
#include "massmodel.h"
#include "massmodelaxi.h"
#include "utils.h"
#include "managepotential.h"

namespace smile{

/* --------------  CICGenerator*** classes - preparation of initial conditions for orbit library ------------- */
void CICGeneratorStationary::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    double Xmax = longAxisRadius(potential, E, Omega);
    if(Xmax<0) return;  // error...
    CPosVelPoint<double> IC;
    if(N_dim==N_DIM) {
#if SCHWARZSCHILD_START_SPACE
        size_t ngrd=static_cast<size_t>(sqrt(numPoints/3.0));
#else
        size_t ngrd=static_cast<size_t>(sqrt(1.0*numPoints*M_PI/2));
#endif
        for(size_t ith=0; ith<ngrd; ith++)
        {
#if SCHWARZSCHILD_START_SPACE
            double theta = tan((ith+0.5)/ngrd * M_PI/4);
            for(size_t iph=0; iph<ngrd*3; iph++)
            {
                double phi = tan((iph/3+0.5)/ngrd * M_PI/4);
                double X=0, Y=0, Z=0;
                switch(iph%3)
                {
                case 0:
                    X=Xmax;
                    Y=Xmax*theta;
                    Z=Xmax*phi;
                    break;
                case 1:
                    Y=Xmax;
                    Z=Xmax*theta;
                    X=Xmax*phi;
                    break;
                case 2:
                    Z=Xmax;
                    X=Xmax*theta;
                    Y=Xmax*phi;
                    break;
                }
#else
            double theta = (1 - (ith+0.5)/ngrd) * M_PI/2;
            size_t nph = static_cast<size_t>(ngrd * sin(theta) + 1);
            for(size_t iph=0; iph<nph; iph++)
            {
                double phi = (iph+0.5)/nph * M_PI/2;
                double X=0, Y=0, Z=0;

                X=Xmax*sin(theta)*cos(phi);
                Y=Xmax*sin(theta)*sin(phi);
                Z=Xmax*cos(theta);
#endif
                double r=findIntersection(potential, E, X, Y, Z, Omega);
                IC.Pos[0]=X*r; IC.Pos[1]=Y*r; IC.Pos[2]=Z*r;
                IC.Vel[0]=IC.Vel[1]=IC.Vel[2]=0;
                // here no prior weight is assigned, although in principle one could use Thomas et al.(2004) 
                // tesselation method for determining the volume of phase space adjacent to the given orbit.
                points->add(IC, 0);
            }
        }
    } else {  // N_dim==2
        for(size_t iph=0; iph<numPoints; iph++)
        {
            double phi = (iph+0.5)/numPoints * M_PI/2;
            double X=Xmax*cos(phi);
            double Y=Xmax*sin(phi);
            double r=findIntersection(potential, E, X, Y, 0, Omega);
            IC.Pos[0]=X*r; IC.Pos[1]=Y*r; IC.Pos[2]=0;
            IC.Vel[0]=IC.Vel[1]=IC.Vel[2]=0;
            points->add(IC, 0);
        }
    }
};

void CICGeneratorPrincipalPlane::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    double Xmax = longAxisRadius(potential, E, Omega);
    if(Xmax<0) return;  // error...
    CPosVelPoint<double> IC;
    if(N_dim==N_DIM) {
#if ANNULUS_PP_START_SPACE  // unused
        double rmin=Xmax, rminc;
        findPeriodicOrbit(potential, E, 0, 1, rminc);
        if(rminc<rmin) rmin=rminc;
        if(rmin>0) findPeriodicOrbit(potential, E, 1, 2, rminc);
        if(rminc<rmin) rmin=rminc;
        if(rmin>0) findPeriodicOrbit(potential, E, 0, 2, rminc);
        if(rminc<rmin) rmin=rminc;
        rminc=rmin/Xmax;
#else
        double rminc=0;
#endif
        // use elliptic annulus with inner radius rmin and outer xmax
        size_t ngrd = static_cast<size_t>(sqrt(4*numPoints/N_DIM/M_PI/(1+rminc))+0.5);
        for(size_t ir=0; ir<ngrd; ir++)
        {
            double r = (ir+0.5)/ngrd * (1-rminc) + rminc;
            size_t nph = static_cast<size_t>(r * ngrd * M_PI/2 + 0.22);
            for(size_t iph=0; iph<nph; iph++)
            {
                double phi = (iph+0.5)/nph * M_PI/2;
                double rmax=findIntersection(potential, E, Xmax*cos(phi), Xmax*sin(phi), 0, Omega);
                IC.Pos[0]=r*rmax*Xmax*cos(phi); IC.Pos[1]=r*rmax*Xmax*sin(phi); IC.Pos[2]=0;
                IC.Vel[0]=IC.Vel[1]=0; IC.Vel[2]=sqrt(2*(E-potential->Phi(IC.Pos[0], IC.Pos[1], IC.Pos[2])));
                if(IC.Vel[2]>0)
                    points->add(IC, 0);
                rmax=findIntersection(potential, E, 0, Xmax*cos(phi), Xmax*sin(phi), Omega);
                IC.Pos[0]=0; IC.Pos[1]=r*rmax*Xmax*cos(phi); IC.Pos[2]=r*rmax*Xmax*sin(phi);
                IC.Vel[1]=IC.Vel[2]=0; IC.Vel[0]=sqrt(2*(E-potential->Phi(IC.Pos[0], IC.Pos[1], IC.Pos[2])));
                if(IC.Vel[0]>0)
                    points->add(IC, 0);
                rmax=findIntersection(potential, E, Xmax*sin(phi), 0, Xmax*cos(phi), Omega);
                IC.Pos[0]=r*rmax*Xmax*sin(phi); IC.Pos[1]=0; IC.Pos[2]=r*rmax*Xmax*cos(phi);
                IC.Vel[0]=IC.Vel[2]=0; IC.Vel[1]=sqrt(2*(E-potential->Phi(IC.Pos[0], IC.Pos[1], IC.Pos[2])));
                if(IC.Vel[1]>0)
                    points->add(IC, 0);
            }
        }
    } else {  // N_dim==2
        for(size_t ir=0; ir<numPoints/2; ir++)
        {
            double r = (ir+0.5)/(numPoints/2);
            IC.Pos[0]=r*Xmax; IC.Pos[1]=0; IC.Pos[2]=0;
            IC.Vel[0]=IC.Vel[2]=0; IC.Vel[1]=sqrt(2*(E-potential->Phi(IC.Pos[0], IC.Pos[1], IC.Pos[2])));
            if(IC.Vel[1]>0)
                points->add(IC, 0);
            IC.Pos[0]=0; IC.Pos[1]=r*Xmax; IC.Pos[2]=0;
            IC.Vel[1]=IC.Vel[2]=0; IC.Vel[0]=sqrt(2*(E-potential->Phi(IC.Pos[0], IC.Pos[1], IC.Pos[2])));
            if(IC.Vel[0]>0)
                points->add(IC, 0);
        }
    }
};

void CICGeneratorYalpha::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    CPosVelPoint<double> IC;
    size_t ny=N_dim==N_DIM ? static_cast<size_t>(sqrt(numPoints*1.0)) : numPoints;
    size_t na=numPoints/ny;
    double Xmax = longAxisRadius(potential, E, Omega);
    if(Xmax<0) return;  // error...
    double Ymax=0;
    for(int niter=0; niter<50 && Xmax-Ymax>Xmax*1e-8; niter++)  // should be smarter, use findIntersection
    {
        double Ey=potential->Phi(0,(Xmax+Ymax)/2,0);
        if(Ey>E) Xmax=(Xmax+Ymax)/2; else Ymax=(Xmax+Ymax)/2;
    }   // thus we have found approximately the intermediate axis radius at given energy;
    for(size_t iy=0; iy<ny; iy++)
    {
        double Y = (iy+0.5)/ny * Ymax;
        double V=E-potential->Phi(0,Y,0);
        if(V<0) continue;   // should not happen though..
        V=sqrt(2*V);
        IC.Pos[0]=0; IC.Pos[1]=Y; IC.Pos[2]=0; IC.Vel[1]=0;
        for(size_t ia=0; ia<na; ia++)
        {
            double alpha = (ia+0.5)/na * M_PI;
            IC.Vel[0]=V*cos(alpha); IC.Vel[2]=V*sin(alpha);
            points->add(IC, 0);
        }
    }
};

void CICGeneratorRandomE::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    double Xmax = longAxisRadius(potential, E, Omega);
    if(Xmax<0) return;  // error...
    CPosVelPoint<double> IC;
    const size_t npsampling = std::max<size_t>(numPoints*10, 1000);
    std::vector<double> Xs(npsampling), Ys(npsampling), Zs(npsampling), Ws(npsampling), Vs(npsampling);
    double X, Y, Z, totalWeight=0;
    for(size_t s=0; s<npsampling; s++)
    {   // sample accessible configuration space with appropriate weight
        double v2;
        do{
            X=getRandomNumber()*Xmax;
            Y=getRandomNumber()*Xmax;
            Z=N_dim==N_DIM ? getRandomNumber()*Xmax : 0;
            v2=2*(E-potential->Phi(X,Y,Z))+pow_2(Omega)*(pow_2(X)+pow_2(Y));
        } while(v2<0);
        Xs[s]=X; Ys[s]=Y; Zs[s]=Z;
        Vs[s]=sqrt(v2);
        Ws[s]=N_dim==N_DIM ? Vs[s] : 1.0;
        totalWeight+=Ws[s];
    }
    // now select points from sampled
    for(size_t np=0; np<numPoints; np++)
    {
        double indw=getRandomNumber() * totalWeight;
        size_t i=0;
        while(i<npsampling-1 && indw>=Ws[i]) 
        {
            indw-=Ws[i];
            i++;
        }
        IC.Pos[0]=Xs[i]; IC.Pos[1]=Ys[i]; IC.Pos[2]=Zs[i]; 
        double vx, vy, vz=0;
        getRandomUnitVector(&vx, &vy, N_dim==N_DIM ? &vz : NULL);
        IC.Vel[0]=Vs[i]*vx - Omega*Ys[i];
        IC.Vel[1]=Vs[i]*vy + Omega*Xs[i];
        IC.Vel[2]=Vs[i]*vz;
        points->add(IC, 0);
        // remove this point from sample list, to avoid duplication
        totalWeight-=Ws[i];
        Ws[i]=0;  // it won't occur again
    }
};

// generate points in proportion to density, using recursive space splitting 
double densSampler(double x[N_DIM], size_t /*dim*/, void* param)
{
    return ((const CDensity*)param)->Rho(x[0], x[1], x[2]);
}

void CICGeneratorDensitySampler::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    double mass0=density->Mass(0);
    double totalMass=density->totalMass()-mass0;
    if(totalMass<0) {
        my_error(FUNCNAME, "Error, cannot sample density distribution for an infinite-mass model");
        return;
    }
    double maxPointMass=totalMass/numPoints;
    double Rmax=1;
    while(density->Mass(Rmax)-mass0<totalMass*MAX_LAGRANGIAN_RADIUS) Rmax*=2;   // obtain the initial box size
    CPointMassSet<double> tmpPoints;
    gsl_monte_function F;
    gsl_monte_vegas_state* state=gsl_monte_vegas_alloc(N_DIM);
    gsl_monte_vegas_init(state);
    state->mode=GSL_VEGAS_MODE_IMPORTANCE_ONLY;
    F.f=&densSampler;
    F.dim=N_DIM;
    F.params=const_cast<CDensity*>(density);
    double xlow[N_DIM]={-Rmax,-Rmax,-Rmax };
    double xupp[N_DIM]={ Rmax, Rmax, Rmax };
    double result, error;
    // determine number of sample points and bins
    const size_t MINBINS=25;  // default maximum is 50
    const double MINPOINTSPERBIN=5.0;
    size_t Nsamples=std::max<size_t>(numPoints*10, static_cast<size_t>(MINBINS*MINBINS*MINBINS*MINPOINTSPERBIN));
    state->bins_max=std::min<size_t>(state->bins_max, static_cast<size_t>(pow(1.0+Nsamples/MINPOINTSPERBIN, 1.0/3)));
    // warm-up run
    size_t nIter=0;
    double chisq;
    do{
        gsl_monte_vegas_integrate(&F, xlow, xupp, N_DIM, Nsamples, randgen, state, &result, &error);  // randgen is the global instance of random number generator
        chisq=gsl_monte_vegas_chisq(state);
#ifdef DEBUGPRINT
        my_message(FUNCNAME,
            "VEGAS mass estimate="+convertToString(result)+"; chi^2="+convertToString(chisq));
#endif
        nIter++;
    } while(nIter<10 && (fabs(result/totalMass-1)>0.1 && chisq>2));
    unsigned int numCells=state->bins*state->bins*state->bins;
    double meanNumSamplePointsPerCell = Nsamples*1.0/numCells;   // initially all cells have the same number of sampling points
    vectord cellVolumes(numCells);  // physical volume of each cell
    std::vector<unsigned int> cellNumPoints(numCells);  // number of sampling points in each cell
    std::vector<unsigned int> cellIndices(numCells);    // array of cells that need more sampling points (initially all cells are included)
    // compute cell volumes
    for(unsigned int indCell=0; indCell<numCells; indCell++) {
        unsigned int indCellCoord[N_DIM] =   // indices of the cell in three coordinates
        { indCell % state->bins, (indCell / state->bins) % state->bins, indCell / (state->bins * state->bins) };
        double cellVolume=1;
        for(unsigned int d=0; d<N_DIM; d++) {
            double cellSize=(state->xi[(indCellCoord[d]+1)*N_DIM+d]-state->xi[indCellCoord[d]*N_DIM+d])*state->delx[d];
            cellVolume*=cellSize;
        }
        cellVolumes[indCell]=cellVolume;
        cellIndices[indCell]=indCell;
    }
    double corrFactor;  // ratio of monte-carlo mass estimate to the true mass
    nIter=0;
    do{
        // sprinkle some sample points
        size_t numSamplesIteration=static_cast<size_t>(meanNumSamplePointsPerCell*cellIndices.size());
        for(size_t s=0; s<numSamplesIteration; s++) {
            // pick up a cell
            unsigned int indCellArray=std::min<unsigned int>(cellIndices.size()-1,
                static_cast<unsigned int>(gsl_rng_uniform(randgen)*cellIndices.size()) );
            unsigned int indCell=cellIndices[indCellArray];
            unsigned int indCellCoord[N_DIM] =   // indices of the cell in three coordinates
            { indCell % state->bins, (indCell / state->bins) % state->bins, indCell / (state->bins * state->bins) };
            double coords[N_DIM];
            for(unsigned int d=0; d<N_DIM; d++) {
                double cellSize=(state->xi[(indCellCoord[d]+1)*N_DIM+d]-state->xi[indCellCoord[d]*N_DIM+d])*state->delx[d];
                double fracCoord=gsl_rng_uniform(randgen);  // fractional d-th coordinate in the given bin
                coords[d] = xlow[d] + state->xi[indCellCoord[d]*N_DIM+d]*state->delx[d] + fracCoord*cellSize;
            }
            cellNumPoints[indCell]++;
            double dens=density->Rho(coords[0], coords[1], coords[2]);
            tmpPoints.add(CPosVelPoint<double>(coords[0], coords[1], coords[2], static_cast<double>(indCell), 0, 0), dens);  // x-component of velocity holds the cell index
        }
        // check that all sample points have associated mass less than the target mass of a single point
        double sumMass=0;
        for(size_t s=0; s<tmpPoints.size(); s++) {
            unsigned int indCell=static_cast<unsigned int>(tmpPoints[s].first.Vel[0]);
            double samplePointMass = tmpPoints[s].second * cellVolumes[indCell] / cellNumPoints[indCell];
            sumMass+=samplePointMass;
        }
        corrFactor=totalMass/sumMass;
        std::set< unsigned int > badCells;
        for(size_t s=0; s<tmpPoints.size(); s++) {
            unsigned int indCell=static_cast<unsigned int>(tmpPoints[s].first.Vel[0]);
            double samplePointMass = tmpPoints[s].second * cellVolumes[indCell] / cellNumPoints[indCell] * corrFactor;
            if(samplePointMass>maxPointMass)   // the cell that this point belongs to needs more study..
                badCells.insert(indCell);
        }
        cellIndices.clear();
        if(!badCells.empty()) {  // need another iteration
            for(unsigned int indCell=0; indCell<numCells; indCell++) 
                if(badCells.find(indCell)!=badCells.end()) {  // need to sample more points from this cell
                    cellIndices.push_back(indCell);
                }
            if(nIter>0)  // if this is the first iteration, don't need to double #samplePointsPerCell, 
                meanNumSamplePointsPerCell*=2; // because on the next iteration we add the same number of points to each cell to be refined
#ifdef DEBUGPRINT
            my_message(FUNCNAME,
                "Mass estimate="+convertToString(sumMass)+"; "+
                convertToString(cellIndices.size())+" cells need refinement");
#endif
        }
#ifdef DEBUGPRINT
        else 
            my_message(FUNCNAME,
                "Mass estimate="+convertToString(sumMass)+"; all done");
#endif
        nIter++;
    } while(nIter<15 && cellIndices.size()>0);  // repeat until there are no more cells to be refined
    if(corrFactor<0.9 || corrFactor>1.1)
        my_message(FUNCNAME,
            "Warning, mass estimate from Monte-Carlo sampling="+convertToString(totalMass/corrFactor)+
            " significantly differs from the actual mass="+convertToString(totalMass));
    // sample output points from the list of all points
    size_t ind_out=1;
    double weight=0;
    float pointMass=static_cast<float>(maxPointMass);
    for(size_t ind_in=0; ind_in<tmpPoints.size(); ind_in++) {
        unsigned int indCell=static_cast<unsigned int>(tmpPoints[ind_in].first.Vel[0]);
        double samplePointMass = tmpPoints[ind_in].second * cellVolumes[indCell] / cellNumPoints[indCell] * corrFactor;
        weight+=samplePointMass;
        if(weight>=totalMass*ind_out/(numPoints+1e-6)) {
            tmpPoints[ind_in].first.Vel[0]=0;  // clear the temporary stored cell index in this field
            points->add(tmpPoints[ind_in].first, pointMass);
            ind_out++;
        }
    }
    gsl_monte_vegas_free(state);
};

// sample points from spherical model 
void CICGeneratorDensitySpherical::createInitialConditions(size_t numPoints, smile::CPointMassSet<float> *points)
{
    double totalmass=model->getTotalMass(), centralmass=model->mass(0);
    float pointmass=static_cast<float>((totalmass-centralmass)/numPoints);
    points->data.reserve(numPoints);
    for(size_t ind=0; ind<numPoints; ind++)
    {
        double radius=0, encmass;
        if (quiet==1) 
            encmass=(ind+getRandomNumber())/numPoints;
        else if (quiet==2) 
            encmass = (ind+0.5)/numPoints;
        else
            encmass=getRandomNumber();
        encmass=(totalmass-centralmass)*encmass + centralmass;
        radius=model->rad(encmass);
        double X, Y, Z;
        getRandomUnitVector(&X, &Y, &Z);
        CPosVelPoint<float> pt(
            static_cast<float>(radius*X),
            static_cast<float>(radius*Y), 
            static_cast<float>(radius*Z), 0, 0, 0);
        points->add(pt, pointmass);
    }
}

/// \cond INTERNAL_DOCS
struct CDensityFinderParam {
    const CDensity* P;
    double dens;
    double X, Y, Z; 
};
double densityFinder(double r, void* params)
{
    return ((CDensityFinderParam*)params)->P->Rho(
        r*((CDensityFinderParam*)params)->X, r*((CDensityFinderParam*)params)->Y, r*((CDensityFinderParam*)params)->Z) 
        - ((CDensityFinderParam*)params)->dens;
}
/// simple linear interpolation function (could use a GSL analog?)
double piecewise_linear(double x, const vectord& values_x, const vectord& values_y)
{
    //assert(values_x.size()==values_y.size());
    size_t indbin=0;
    while(indbin<values_x.size() && x>values_x[indbin]) indbin++;
    if(indbin==0) 
        return values_y.front();
    else if(indbin>=values_x.size()) 
        return values_y.back();
    else   // linear interpolation
        return (values_x[indbin]*values_y[indbin-1]-values_x[indbin-1]*values_y[indbin]
            + x*(values_y[indbin]-values_y[indbin-1]))/(values_x[indbin]-values_x[indbin-1]);
}
/// \endcond

// sample points using ellipsoidal approximation to isodensity surfaces
void CICGeneratorDensityEllipsoidal::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    double totalMassWithoutBH = density->totalMass()-density->Mass(0);
    if(totalMassWithoutBH<0) {
        my_error(FUNCNAME, "Error, total mass is not finite, cannot create initial conditions for this model");
        return;
    }
    const size_t NUM_RADIAL_POINTS_AXIS_RATIO=10;       // evaluate shape at this many points in radius
    // create spherical mass model with its distribution function, which will be used to draw initial conditions
    const CMassModel* massModel = createMassModel(density);
    if(massModel==NULL) {
        my_error(FUNCNAME, "Unknown error in creating equivalent spherical mass model");
        return;
    }
    totalMassWithoutBH = std::min<double>(totalMassWithoutBH, massModel->getTotalMass()-massModel->mass(0));   // to ensure correct initialization of outermost initial conditions
    float pointMass=static_cast<float>(totalMassWithoutBH/numPoints);

    // evaluate shape as function of radius (defined as axis ratios of equidensity ellipsoids)
    // NOTE: this works well only if density is decreasing with radius. A fail-safe approach is to assign axes ratio of unity if anything goes wrong.
    vectord radii_shape, axes_a, axes_b, axes_c;
    radii_shape.assign(NUM_RADIAL_POINTS_AXIS_RATIO, 0);
    axes_a.assign(NUM_RADIAL_POINTS_AXIS_RATIO, 0);
    axes_b.assign(NUM_RADIAL_POINTS_AXIS_RATIO, 0);
    axes_c.assign(NUM_RADIAL_POINTS_AXIS_RATIO, 0);
    gsl_function F;
    CDensityFinderParam PP;
    PP.P=density;
    F.function=&densityFinder;
    F.params=&PP;
    for(size_t i=0; i<NUM_RADIAL_POINTS_AXIS_RATIO; i++)
    {
        radii_shape[i]=massModel->rad((i+0.5)/NUM_RADIAL_POINTS_AXIS_RATIO * totalMassWithoutBH + massModel->mass(0));
        PP.dens=massModel->dens(radii_shape[i]);
        if(radii_shape[i]<0 || !gsl_finite(radii_shape[i]) || PP.dens<0 || !gsl_finite(PP.dens)) 
        {
            my_error(FUNCNAME, "Warning, at r="+convertToString(radii_shape[i])+" can't evaluate shape profile");
            axes_a[i]=1.0; axes_b[i]=1.0; axes_c[i]=1.0;
            continue;  // weird things do happen, just skip it.
        }
        // find the radii along x,y,z axes where density in real model is equal to that of spherical model
        double Rroot;
        const double ARlow=0.1, ARupp=10;    // we won't be able to deal correctly with highly flattened models
        // X axis
        PP.X=radii_shape[i]; PP.Y=0; PP.Z=0;
        Rroot = findRoot(&F, ARlow, ARupp, 1.0, 1e-3);
        if(i>0 && Rroot*radii_shape[i] < axes_a[i-1]*radii_shape[i-1])
            Rroot = axes_a[i-1]*radii_shape[i-1]/radii_shape[i] * 1.01;  // ensure monotonicity
        axes_a[i]=Rroot;
        // Y axis
        PP.X=0; PP.Y=radii_shape[i]; PP.Z=0;
        Rroot = findRoot(&F, ARlow, ARupp, 1.0, 1e-3);
        if(i>0 && Rroot*radii_shape[i] < axes_b[i-1]*radii_shape[i-1])
            Rroot = axes_b[i-1]*radii_shape[i-1]/radii_shape[i] * 1.01;  // ensure monotonicity
        axes_b[i]=Rroot;
        // Z axis
        PP.X=0; PP.Y=0; PP.Z=radii_shape[i];
        Rroot = findRoot(&F, ARlow, ARupp, 1.0, 1e-3);
        if(i>0 && Rroot*radii_shape[i] < axes_c[i-1]*radii_shape[i-1])
            Rroot = axes_c[i-1]*radii_shape[i-1]/radii_shape[i] * 1.01;  // ensure monotonicity
        axes_c[i]=Rroot;
#ifdef DEBUGPRINT
        my_message(FUNCNAME,
            "r="+convertToString(radii_shape[i])+",X="+convertToString(axes_a[i])+
            ",Y="+convertToString(axes_b[i])+",Z="+convertToString(axes_c[i]));
#endif
    }

    for(size_t o=0; o<numPoints; o++)
    {   // sample accessible configuration space with appropriate weight
        double fracmass = (o+0.5)/numPoints;  // fraction of enclosed mass
        double encmass = fracmass * totalMassWithoutBH;
        double radius=massModel->rad(encmass + massModel->mass(0));
        if(radius<0 || !gsl_finite(radius)) 
            continue;  // weird things do happen, just skip it.
        // find axis ratios at this radius
        double axis_a=piecewise_linear(radius, radii_shape, axes_a);
        double axis_b=piecewise_linear(radius, radii_shape, axes_b);
        double axis_c=piecewise_linear(radius, radii_shape, axes_c);
        //int ntrial=0;
        double X, Y, Z;//, poten;
        do{
            getRandomUnitVector(&X, &Y, &Z);
            X*=axis_a*radius; 
            Y*=axis_b*radius; 
            Z*=axis_c*radius; 
            //poten=potential->Phi(X, Y, Z);
            //ntrial++;
        } while(0/*poten>=maxEnergy && ntrial<1e3*/);
        points->add(
            CPosVelPoint<float>((float)X, (float)Y, (float)Z, 0, 0, 0), pointMass);
    }
    delete massModel;
};

// estimate the upper bound on energy as one half of the potential energy of the outermost particle
double computeEnergyCutoff(const CPotential* potential, const CPointMassSet<float> & points)
{
    double maxRad2=0;
    for(CPointMassSet<float>::Type::const_iterator p=points.data.begin(); p!=points.data.end(); p++)
        maxRad2=std::max<double>(maxRad2, pow_2(p->first.Pos[0])+pow_2(p->first.Pos[1])+pow_2(p->first.Pos[2]));
    double val=potential->Phi(sqrt(maxRad2),0,0);
    return val<0 ? val*0.5 : val+1.0;  // take a fail-safe value if val>=0 (shouldn't occur for a finite-mass model)
}

// initial conditions for velocity based on spherical Eddington distribution function
void CICGeneratorVelocityEddington::createInitialConditions(size_t /*numPoints*/, CPointMassSet<float> *points)
{
    const CMassModel* massModel = createMassModel(potential);
    if(massModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    const CMassModel* densityModel = 
        density==NULL || density==potential ? massModel :
        createMassModel(density, NUM_RADIAL_POINTS_SPHERICAL_MODEL, 0, 0, massModel);  // need to create a separate model for density
    if(densityModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    double Emax=computeEnergyCutoff(potential, *points);
    for(CPointMassSet<float>::Type::iterator point=points->data.begin(); point!=points->data.end(); point++)
    {
        // correction factor for escape velocity
        double radius=sqrt(pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]));
        double poten=potential->Phi(point->first.Pos[0], point->first.Pos[1], point->first.Pos[2]);
        double velcorr = sqrt(poten / massModel->pot(radius));
        if(gsl_isnan(velcorr)) velcorr=1.0;
        double veloc, E;
        size_t ntrial=0;
        do{
            veloc=densityModel->samplev(radius) * velcorr;
            ntrial++;
            E=poten+veloc*veloc/2;
        } while(E >= Emax && ntrial<1e3);  // to avoid infinite loop in the case of weird potential
        if(E>=Emax) veloc=0;
        float prior=1;
        double X,Y,Z;
        if(weightSkewFactorL==0) {  // direct velocity vector isotropically
            getRandomUnitVector(&X, &Y, &Z);
            point->first.Vel[0]=static_cast<float>(veloc*X); 
            point->first.Vel[1]=static_cast<float>(veloc*Y); 
            point->first.Vel[2]=static_cast<float>(veloc*Z);
        } else {   // create more orbits with lower angular momentum
            double w=getRandomNumber()*2-1;
            double sign=w>0?1:-1; 
            w=fabs(w);
            double oneminuscosth=w*(2+(weightSkewFactorL-1)*w)/(1+weightSkewFactorL);  // 1-cos(theta) is more frequently sampled at low values
            prior=static_cast<float>(2*(1+(weightSkewFactorL-1)*w)/(1+weightSkewFactorL));   // this bias is accounted for in the choice of prior weight
            double Vpar=veloc*(1-oneminuscosth)*sign;   // component of velocity parallel to radius-vector
            double Vper=sqrt(pow_2(veloc)-pow_2(Vpar)); // ..and perpendicular
            getRandomUnitVector(&X, &Y, &Z);
            double WdotR=X*point->first.Pos[0]+Y*point->first.Pos[1]+Z*point->first.Pos[2];  // projection of random vector W onto the radius vector R
            double Rmag2=pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]);
            double Rmag=sqrt(Rmag2);
            X -= WdotR*point->first.Pos[0]/Rmag2;
            Y -= WdotR*point->first.Pos[1]/Rmag2;
            Z -= WdotR*point->first.Pos[2]/Rmag2; // subtract parallel component
            double Wmag=sqrt(pow_2(X)+pow_2(Y)+pow_2(Z));
            if(Wmag==0) Wmag=1;  // avoid singularity
            point->first.Vel[0] = static_cast<float>(point->first.Pos[0]/Rmag*Vpar + X/Wmag*Vper);
            point->first.Vel[1] = static_cast<float>(point->first.Pos[1]/Rmag*Vpar + Y/Wmag*Vper);
            point->first.Vel[2] = static_cast<float>(point->first.Pos[2]/Rmag*Vpar + Z/Wmag*Vper);
        }
        point->second*=prior;
    }
    if(densityModel!=massModel) delete densityModel;
    delete massModel;
};

// initial conditions for velocity based on spherical Jeans equation
/// \cond INTERNAL_DOCS
struct CJeansIntegrandParam {
    const CMassModel* potentialModel;
    const CMassModel* densityModel;
    double beta;
};
double jeansIntegrandSph(double r, void* params)
{
    return pow(r, 2*((CJeansIntegrandParam*)params)->beta-2) * 
        ((CJeansIntegrandParam*)params)->potentialModel->mass(r) *
        ((CJeansIntegrandParam*)params)->densityModel->dens(r);
}
/// \endcond

void CICGeneratorVelocityJeansSph::createInitialConditions(size_t /*numPoints*/, CPointMassSet<float> *points)
{
    const CMassModel* massModel = createMassModel(potential);
    if(massModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    const CMassModel* densityModel = 
        density==NULL || density==potential ? massModel :
        createMassModel(density, NUM_RADIAL_POINTS_SPHERICAL_MODEL, 0, 0, massModel);  // need to create a separate model for density
    if(densityModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    vectord radii_veldisp, sigma_r;
    gsl_function F;
    CJeansIntegrandParam param;
    F.params=&param;
    F.function=&jeansIntegrandSph;
    param.potentialModel=massModel;
    param.densityModel=densityModel;
    param.beta=beta;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    const size_t NUM_RADIAL_POINTS_VEL_DISP=NUM_RADIAL_POINTS_SPHERICAL_MODEL;
    double Mbh=massModel->mass(0);
    double rout=massModel->rad(massModel->getTotalMass()*(1-1.0/NUM_RADIAL_POINTS_VEL_DISP) + Mbh);
    double rin =massModel->rad(massModel->getTotalMass()*0.5 + Mbh)/NUM_RADIAL_POINTS_VEL_DISP*2;
    if(Mbh>0) rin=std::min<double>(rin, Mbh/10);
    rin=std::min<double>(rin, rout/NUM_RADIAL_POINTS_VEL_DISP);
    createNonuniformGrid(NUM_RADIAL_POINTS_VEL_DISP, rin, rout, true, &radii_veldisp);
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Grid in radius from "+convertToString(rin)+" to "+convertToString(rout));
#endif
    radii_veldisp[0] = Mbh==0 ? radii_veldisp[1]*0.01 : 
        std::min<double>(massModel->rad(Mbh*1.001),radii_veldisp[1]*0.1);
    double prevint=0;
    sigma_r.assign(radii_veldisp.size(), 0);
    for(int ind=static_cast<int>(radii_veldisp.size())-1; ind>=0; ind--)
    {
        if(ind==static_cast<int>(radii_veldisp.size())-1)
            gsl_integration_qagiu(&F, radii_veldisp.back(), 0, 1e-4, 1000, ws, &result, &error);
        else
            gsl_integration_qags(&F, radii_veldisp[ind], radii_veldisp[ind+1], 0, 1e-4, 1000, ws, &result, &error);
        prevint+=result;
        sigma_r[ind] = pow(radii_veldisp[ind], 1-2*beta) / densityModel->dens(radii_veldisp[ind]) * prevint;
    }
    radii_veldisp[0]=0;  // pretend that we have computed central velocity dispersion
    if(Mbh==0) sigma_r[0]=0;  // r*sigma^2 should be zero at center if no BH is present
    gsl_integration_workspace_free(ws);

    double Emax=computeEnergyCutoff(potential, *points);
    // now sample velocities
    for(CPointMassSet<float>::Type::iterator point=points->data.begin(); point!=points->data.end(); point++)
    {
        double radius=sqrt(pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]));
        double poten=potential->Phi(point->first.Pos[0], point->first.Pos[1], point->first.Pos[2]);
        double velcorr = sqrt(poten / massModel->pot(radius));
        if(gsl_isnan(velcorr)) velcorr=1.0;
        double veldisp_r = sqrt(piecewise_linear(radius, radii_veldisp, sigma_r)/radius) * velcorr;
        double veldisp_t = veldisp_r*sqrt(1-beta);
        double Vpar, Vper, E;
        size_t ntrial=0;
        do{
            double v1, v2;
            getNormalRandomNumbers(&v1, &v2);
            Vpar=v1*veldisp_r;
            getNormalRandomNumbers(&v1, &v2);
            Vper=sqrt(v1*v1+v2*v2)*veldisp_t;
            ntrial++;
            E=poten+(Vpar*Vpar+Vper*Vper)/2;
        } while(E >= Emax && ntrial<1e3);  // to avoid infinite loop in the case of weird potential
        if(E>=Emax) {
            Vpar=0; Vper=0;
        }
        double X,Y,Z;
        //prior=static_cast<float>(2*(1+(weightSkewFactorL-1)*w)/(1+weightSkewFactorL));
        getRandomUnitVector(&X, &Y, &Z);
        double WdotR=X*point->first.Pos[0]+Y*point->first.Pos[1]+Z*point->first.Pos[2];  // projection of random vector W onto the radius vector R
        double Rmag2=pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]);
        double Rmag=sqrt(Rmag2);
        X -= WdotR*point->first.Pos[0]/Rmag2;
        Y -= WdotR*point->first.Pos[1]/Rmag2;
        Z -= WdotR*point->first.Pos[2]/Rmag2; // subtract parallel component
        double Wmag=sqrt(pow_2(X)+pow_2(Y)+pow_2(Z));
        if(Wmag==0) Wmag=1;  // avoid singularity
        point->first.Vel[0] = static_cast<float>(point->first.Pos[0]/Rmag*Vpar + X/Wmag*Vper);
        point->first.Vel[1] = static_cast<float>(point->first.Pos[1]/Rmag*Vpar + Y/Wmag*Vper);
        point->first.Vel[2] = static_cast<float>(point->first.Pos[2]/Rmag*Vpar + Z/Wmag*Vper);
    }
    if(densityModel!=massModel) delete densityModel;
    delete massModel;
}

#ifdef HAVE_INTERP2D
// initial conditions for velocity based on axisymmetric Jeans equation
void CICGeneratorVelocityJeansAxi::createInitialConditions(size_t /*numPoints*/, CPointMassSet<float> *points)
{
    CMassModelAxisymmetricJeans* modelAxi=new CMassModelAxisymmetricJeans(potential, 
        NUM_RADIAL_POINTS_SPHERICAL_MODEL, NUM_RADIAL_POINTS_SPHERICAL_MODEL, beta_m, density);
    double Emax=computeEnergyCutoff(potential, *points);
    for(CPointMassSet<float>::Type::iterator point=points->data.begin(); point!=points->data.end(); point++)
    {
        double R=sqrt(pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1]));
        double z=point->first.Pos[2];
        double poten=potential->Phi(point->first.Pos[0], point->first.Pos[1], point->first.Pos[2]);
        double veldisp_z   = modelAxi->sigma_z(R, z);     // velocity dispersion in the meridional plane: z-component..
        double veldisp_R   = veldisp_z/(1-beta_m);        // ..and R-component
        double veldisp_phi = modelAxi->vphisq_av(R, z);   // total mean square velocity in the azimuthal direction
        double velstrm2    = std::max<double>(veldisp_phi-veldisp_R, 0);  // intermediate quantity
        double kappa_this  = std::min<double>(kappa, sqrt(veldisp_phi/velstrm2));  // rotation parameter should not exceed the upper limit determined by sigma_phi^2>=0
        double velstrm_phi = kappa_this*sqrt(velstrm2);   // streaming velocity (average v_phi)
        veldisp_phi-=pow_2(velstrm_phi);   // subtract avg.velocity and obtain velocity dispersion in the azimuthal direction
        veldisp_R=sqrt(veldisp_R);         // now this holds the "dispersion in the galactic dynamicist's sense", i.e. mean-square deviation from mean
        veldisp_z=sqrt(veldisp_z);
        veldisp_phi=sqrt(std::max<double>(veldisp_phi, 0));
        double VR, Vphi, Vz, E;
        size_t ntrial=0;
        do{
            double v1, v2;
            getNormalRandomNumbers(&v1, &v2);
            Vphi=v1*veldisp_phi + velstrm_phi;
            getNormalRandomNumbers(&v1, &v2);
            VR=v1*veldisp_R;
            Vz=v2*veldisp_z;
            ntrial++;
            E=poten+(Vphi*Vphi+VR*VR+Vz*Vz)/2;
        } while(E >= Emax && ntrial<1e3);  // to avoid infinite loop in the case of weird potential
        if(E>=Emax) {
#ifdef DEBUGPRINT
            my_message(FUNCNAME, "Can't initialize velocity for ("+convertToString(point->first.Pos[0])+
                ","+convertToString(point->first.Pos[1])+","+convertToString(point->first.Pos[2])+"): "
                "Phi="+convertToString(poten)+",Vphi="+convertToString(velstrm_phi)+",Sphi="+convertToString(veldisp_phi)+
                ",SR="+convertToString(veldisp_R)+",Sz="+convertToString(veldisp_z));
#endif
            Vphi=std::min<double>(sqrt(2*(Emax-poten)), velstrm_phi);
            VR=0; Vz=0;
        }
        double Vx = (point->first.Pos[0]*VR - point->first.Pos[1]*Vphi)/R;
        double Vy = (point->first.Pos[1]*VR + point->first.Pos[0]*Vphi)/R;
        point->first.Vel[0] = static_cast<float>(Vx);
        point->first.Vel[1] = static_cast<float>(Vy);
        point->first.Vel[2] = static_cast<float>(Vz);
    }
    delete modelAxi;
}
#endif

// combines positions and velocities
typedef std::pair<float, CPointMassSet<float>::ElemType> CPointMassEnergy;
void CICGeneratorSchw::createInitialConditions(size_t numPoints, CPointMassSet<float> *points)
{
    if(weightSkewFactorE<0) weightSkewFactorE=0;
    size_t numSampleOrbits=static_cast<size_t>(numPoints * (1+weightSkewFactorE/2));   // if skewFactor>0, initially create more orbits than final output number
    CPointMassSet<float> tmpPoints;
    my_message(FUNCNAME, "Generating positions");
    genPosition->createInitialConditions(numSampleOrbits, &tmpPoints);  // assign positions
    my_message(FUNCNAME, "Generating velocities");
    genVelocity->createInitialConditions(0, &tmpPoints);  // assign velocities
    numSampleOrbits=tmpPoints.size();  // could happen that it is not exactly the required number
    std::vector<CPointMassEnergy> samplePoints(numSampleOrbits);
    for(size_t i=0; i<numSampleOrbits; i++) {
        samplePoints[i].second=tmpPoints[i];
        samplePoints[i].first=static_cast<float>(totalEnergy(potential, tmpPoints[i].first));
    }
    // sort orbits by energy (in a stationary frame)
    std::sort(samplePoints.begin(), samplePoints.end(), comparepair<float, CPointMassSet<float>::ElemType>);
    // now retain only necessary number of orbits (leave out some of them if skewFactor>0)
    size_t kinprev=0;
    size_t numOutputOrbits = std::min<size_t>(numPoints, numSampleOrbits);
    for(size_t kout=0; kout<numOutputOrbits; kout++)
    {   // kin - original index of orbit from the input list to be retained at kout'th position in the output list
        size_t kin=static_cast<size_t>(numSampleOrbits * 
            (1+weightSkewFactorE/2*kout/numOutputOrbits)*kout/numOutputOrbits/(1+weightSkewFactorE/2));
        if(kin==kinprev && kout!=0) kin++;   // shouldn't happen though..
        if(kin>=numSampleOrbits) kin=numSampleOrbits-1;  // also shouldn't exceed the number of input orbits
        points->add(samplePoints[kin].second.first,    // PosVelPoint
            static_cast<float>(samplePoints[kin].second.second *   // retain the weight prior from uneven ang.mom.sampling and
            (1+weightSkewFactorE*kout/numOutputOrbits)));          // multiply it by the prior from uneven energy sampling
    }
    my_message(FUNCNAME, "Initial conditions created");
}

//----- utility functions -----//
const char* getICgenNameByType(CBasicICGenerator::ICTYPE ICType)
{
    switch(ICType) {
        case CBasicICGenerator::IC_VEL_EDDINGTON: return "Eddington";
        case CBasicICGenerator::IC_VEL_JEANSSPH:  return "JeansSph";
        case CBasicICGenerator::IC_VEL_JEANSAXI:  return "JeansAxi";
        default: return "Other";
    }
}

CBasicICGenerator::ICTYPE getICgenTypeByName(const std::string& ICName)
{
    if(strings_equal(ICName, "Eddington")) return CBasicICGenerator::IC_VEL_EDDINGTON;
    if(strings_equal(ICName, "JeansSph" )) return CBasicICGenerator::IC_VEL_JEANSSPH;
    if(strings_equal(ICName, "JeansAxi" )) return CBasicICGenerator::IC_VEL_JEANSAXI;
    return CBasicICGenerator::IC_VEL_EDDINGTON;  // default
}

}; // namespace
