/** \file    icgenerator.h
    \brief   Classes for generating the initial conditions for the orbit library
    \author  EV
    \date    2009-2014

    This file contains the family of classes that are used to create initial conditions 
    for orbit library (frequency map or Schwarzschild model).
    All descendants of the base class CBasicICGenerator provide a virtual function 
    to create a necessary number of points, and output them as a CPointMassSet.

    There are three kinds of IC generators: one family is intended to construct IC 
    for frequency map, i.e. for a given value of energy.
    The other two are for the entire range of energies, used in the context of Schwarzschild modelling.
    They are further divided into position generators and velocity generators;
    fundamentally, the first ones sample points from the density profile of the model, with zero velocities, 
    and the second ones assign velocities for already existing positions.
    Furthermore, CICGeneratorSchw class combines these two aspects (taking an instance of each class).
*/

#pragma once
#include "common.h"

namespace smile{

/** Base class for creating initial conditions for orbit library.
    The output consists of starting points and optionally the priors on orbit weights 
    (if used for Schwarzschild modelling) stored in a point mass set.
*/
class CBasicICGenerator
{
public:
/// list all possible types of descendant classes
enum ICTYPE {
    IC_E_STATIONARY,     ///< CICGeneratorStationary
    IC_E_PRINCIPALPLANE, ///< CICGeneratorPrincipalPlane
    IC_E_YALPHA,         ///< CICGeneratorYalpha
    IC_E_RANDOM,         ///< CICGeneratorRandomE
    IC_DENS_SAMPLER,     ///< CICGeneratorDensitySampler
    IC_DENS_ELLIPSOIDAL, ///< CICGeneratorDensityEllipsoidal
    IC_VEL_EDDINGTON,    ///< CICGeneratorVelocityEddington
    IC_VEL_JEANSSPH,     ///< CICGeneratorVelocityJeansSph
    IC_VEL_JEANSAXI      ///< CICGeneratorVelocityJeansAxi
};
    virtual ~CBasicICGenerator() {};
    /** Create a desired number of points and add them to the point mass set. 
        \param[in] numPoints is the requested number of points; the actual output may slightly differ
        \param[in,out] points is the pointer to the array that will contain the points.
        For the IC velocity generators, this array should already contain positions of particles,
        and only the velocities will be initialized. */
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points)=0;
};

/// \name --- Initial conditions for the frequency map (using single value of energy) ---
///@{

/** Initial conditions from stationary start space */
class CICGeneratorStationary: public CBasicICGenerator
{
public:
    CICGeneratorStationary(const CPotential* _potential, double _E, unsigned int _N_dim, double _Omega=0) : 
      CBasicICGenerator(), potential(_potential), E(_E), N_dim(_N_dim), Omega(_Omega) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;
    const double E;
    const unsigned int N_dim;
    const double Omega;
};

/** Initial conditions from principal-plane start space */
class CICGeneratorPrincipalPlane: public CBasicICGenerator
{
public:
    CICGeneratorPrincipalPlane(const CPotential* _potential, double _E, unsigned int _N_dim, double _Omega=0) : 
      CBasicICGenerator(), potential(_potential), E(_E), N_dim(_N_dim), Omega(_Omega) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;
    const double E;
    const unsigned int N_dim;
    const double Omega;
};

/** Initial conditions from Y-alpha start space */
class CICGeneratorYalpha: public CBasicICGenerator
{
public:
    CICGeneratorYalpha(const CPotential* _potential, double _E, unsigned int _N_dim, double _Omega=0) : 
      CBasicICGenerator(), potential(_potential), E(_E), N_dim(_N_dim), Omega(_Omega) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;
    const double E;
    const unsigned int N_dim;
    const double Omega;
};

/** Initial conditions from random start space at given energy */
class CICGeneratorRandomE: public CBasicICGenerator
{
public:
    CICGeneratorRandomE(const CPotential* _potential, double _E, unsigned int _N_dim, double _Omega=0) : 
      CBasicICGenerator(), potential(_potential), E(_E), N_dim(_N_dim), Omega(_Omega) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;
    const double E;
    const unsigned int N_dim;
    const double Omega;
};

///@}
/// \name --- Samplers of the entire density model (no assignment of velocities) ---
///@{

/** sample points from a spherical mass model */
class CICGeneratorDensitySpherical: public CBasicICGenerator
{
public:
    CICGeneratorDensitySpherical(const CMassModel* _model, int _quiet) :
      CBasicICGenerator(), model(_model), quiet(_quiet) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CMassModel* model;
    /// controls the regularity of sampling of radius: 
    /// 0 - random sampling; 2 - quiet: place points at successive radii enclosing equal amount of mass;
    /// 1 - intermediate: place points randomly inside successive intervals enclosing equal amount of mass.
    int quiet;
};

/** sample points coordinates proportionally to local density, no velocities assigned */
class CICGeneratorDensitySampler: public CBasicICGenerator
{
public:
    CICGeneratorDensitySampler(const CDensity* _density) :
      CBasicICGenerator(), density(_density) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CDensity* density;
};

/** sample point coordinates using ellipsoidal equidensity surface approximation; 
    this method is generally inferior to CICGeneratorDensitySampler */
class CICGeneratorDensityEllipsoidal: public CBasicICGenerator
{
public:
    CICGeneratorDensityEllipsoidal(const CDensity* _density):
      CBasicICGenerator(), density(_density) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CDensity* density;
};

///@}
/// \name --- Velocity samplers (take existing points and assign velocities) ---
///@{

/** sample velocities from the Eddington distribution function for equivalent spherical system */
class CICGeneratorVelocityEddington: public CBasicICGenerator
{
public:
    /// construct the instance of spherical mass model for a given density model in an external potential;
    /// if the density and potential are the same, only the latter is provided
    CICGeneratorVelocityEddington(const CPotential* _potential, const CDensity* _density=NULL, double _weightSkewFactorL=0) :
      CBasicICGenerator(), potential(_potential), density(_density), weightSkewFactorL(_weightSkewFactorL) {}; 
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;  ///< total potential of the system
    const CDensity* density;      ///< density of the component considered, or NULL if entire potential is used
    double weightSkewFactorL;     ///< biasing factor for increasing the sampling frequency of low angular momentum orbits
};

/** sample velocities from the spherical Jeans equation */
class CICGeneratorVelocityJeansSph: public CBasicICGenerator
{
public:
    CICGeneratorVelocityJeansSph(const CPotential* _potential, const CDensity* _density=NULL, double _beta=0) :
      CBasicICGenerator(), potential(_potential), density(_density), beta(_beta) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;  ///< total potential of the system
    const CDensity* density;      ///< density of the component considered, or NULL if entire potential is used
    double beta;                  ///< velocity anisotropy coefficient in Jeans equation
};

#ifdef HAVE_INTERP2D
/** sample velocities from Jeans equation for an axisymmetric system */
class CICGeneratorVelocityJeansAxi: public CBasicICGenerator
{
public:
    CICGeneratorVelocityJeansAxi(const CPotential* _potential, const CDensity* _density=NULL, double _beta_m=0, double _kappa=0) :
      CBasicICGenerator(), potential(_potential), density(_density), beta_m(_beta_m), kappa(_kappa) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;  ///< total potential of the system
    const CDensity* density;      ///< density of the component considered, or NULL if entire potential is used
    double beta_m;                ///< velocity anisotropy in the meridional plane (1-sigma_z^2/sigma_R^2)
    double kappa;                 ///< degree of rotation
};
#endif

///@}
/// \name --- Aggregate initial conditions generator for Schwarzschild model ---
///@{

/** class that combines two samplers -- for positions and velocities */
class CICGeneratorSchw: public CBasicICGenerator
{
public:
    CICGeneratorSchw(CBasicICGenerator* _genPosition, CBasicICGenerator* _genVelocity, const CPotential* _potential, double _weightSkewFactorE=0) :
      CBasicICGenerator(),
      genPosition(_genPosition), genVelocity(_genVelocity), potential(_potential), weightSkewFactorE(_weightSkewFactorE) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    CBasicICGenerator* genPosition; ///< class for creating initial positions for the given density component
    CBasicICGenerator* genVelocity; ///< class for sampling initial velocities in the total potential
    const CPotential* potential;    ///< total potential of the system (for sorting particles in energy)
    double weightSkewFactorE;       ///< bias factor for increasing the sampling coverage of most bound orbits
};

///@}
/// \name --- non-class utility routines ---
///@{

//  Correspondence between IC generator names and types 
/// return the name for a given IC generator data type
const char* getICgenNameByType(CBasicICGenerator::ICTYPE ICType);
/// return IC generator data type associated with the text name
CBasicICGenerator::ICTYPE getICgenTypeByName(const std::string& ICName);

///@}

}; // namespace
