/** \file    managepotential.h
    \brief   Creation of CPotential and CMassModel instances, and input/output of potentials
    \author  EV
    \date    2010-2015

    This file provides several utility function to manage instances of CDensity and CPotential: 
    creating a density or potential model from parameters provided in CConfigPotential, 
    creating a potential from a set of point masses or from an N-body snapshot file,
    loading potential coefficients from a text file, 
    writing expansion coefficients to a text file,
    and creating a spherical mass model that approximates the given density model.
    Note that potential here is elementary (non-composite, no central black hole).
*/

#pragma once
#include "common.h"

namespace smile {
struct CConfigPotential; // forward declaration

/** create a density model according to the parameters. 
    This only deals witj finite-mass models, including some of the CPotential descendants.
    \param[in] configPotential contains the parameters (density type, mass, shape, etc.)
    \return    the instance of CDensity, or NULL in case of incorrect parameters
*/
const CDensity* createDensity(const CConfigPotential* config);

/** create an instance of CPotential according to the parameters passed. 
    \param[in,out] config specifies the potential parameters, which could be modified, 
                   e.g. if the potential coefficients are loaded from a file.
                   Massive black hole (config->Mbh) is not included in the potential 
                   (the returned potential is always non-composite)
    \return        the instance of potential, or NULL in case of failure
*/
const CPotential* createPotential(CConfigPotential* config);

/** create a potential of a generic expansion kind from a set of point masses.
    \param[in] configPotential contains the parameters (potential type, number of terms in expansion, etc.)
    \param[in] points is the array of particles that are used in computing the coefficients; 
               if potential type is PT_NB, then an instance of tree-code potential is created.
    \return    a new instance of potential on success, or NULL on failure (e.g. if potential type is inappropriate).
*/ 
template<typename NumT>
const CPotential* createPotentialFromPoints(const CConfigPotential *configPotential, const CPointMassSet<NumT> *points);

/** load a potential from a text or snapshot file.

    The input file may contain one of the following kinds of data:
    - a Nbody snapshot in a text or binary format, handled by classes derived from CBasicIOSnapshot;
    - a potential coefficients file for CPotentialBSE, CPotentialBSECompact, CPotentialSpline, or CPotentialCylSpline;
    - a density model described by CDensityEllipsoidal or CDensityMGE.
    The data format is determined from the first line of the file, and 
    if it is allowed by the parameters passed in configPotential, then 
    the file is read and the instance of a corresponding potential is created. 
    If the input data was not the potential coefficients and the new potential 
    is of BSE or Spline type, then a new file with potential coefficients is 
    created and written via writePotential(), so that later one may load this 
    coef file instead of the original one, which speeds up initialization.
    \param[in,out] configPotential contains the potential parameters and may be updated 
                   upon reading the file (e.g. the number of expansion coefficients 
                   may change). If the file doesn't contain appropriate kind of potential
                   (i.e. if configPotential->PotentialType is PT_NB but the file contains 
                   BSE coefficients or a description of MGE model), an error is returned.
                   configPotential->NbodyFile contains the file name from which the data is loaded.
    \return        a new instance of CPotential on success, or NULL on failure 
                   (display an error message by calling my_error() in that case).
*/
const CPotential* readPotential(CConfigPotential *configPotential);

/** write potential expansion coefficients to a text file.

    The potential must be one of the following expansion classes: 
    CPotentialBSE, CPotentialBSECompact, CPotentialSpline, CPotentialCylSpline, CPotentialScaleFreeSH. 
    The coefficients stored in a file may be later loaded by readPotential() function.
    \param[in] fileName is the output file
    \param[in] potential is the pointer to potential
    \return    true on success, false on failure (file not writeable or potential is of inappropriate type).
*/
bool writePotential(const std::string &fileName, const CPotential* potential);

/** create an equivalent spherical mass model for the given density profile. 
    \param[in] density is the non-spherical density model to be approximated
    \param[in] poten (optional) another spherical mass model that provides the potential 
               (if it is not given self-consistently by this density profile).
    \param[in] numNodes specifies the number of radial grid points in the M(r) profile
    \param[in] Rmin is the radius of the innermost non-zero grid node (0 means auto-select)
    \param[in] Rmax is the radius outermost grid node (0 means auto-select)
    \return    new instance of CMassModel on success, or NULL on fail (e.g. if the density model was infinite) 
*/
CMassModel* createMassModel(const CDensity* density, int numNodes=NUM_RADIAL_POINTS_SPHERICAL_MODEL, 
    double Rmin=0, double Rmax=0, const CMassModel* poten=NULL);

}; // namespace
