/** \file    massmodelaxi.h
    \brief   Class for dealing with axisymmetric mass models
    \author  EV
    \date    2014

    This file defines two classes that are used to deal with axisymmetric density or potential models.
    They are initialized from a given CDensity or CPotential instance 
    (which needs not be axisymmetric itself, in which case all quantities will be averaged 
    over azimuthal angle using an auxiliary class CPotentialAxisymmetrized).
    CMassModelAxisymmetric then provides fast interpolation-based routines to compute 
    surface density and mass in a given slab |z|<Z0 or a cylinder R<R0.
    CMassModelAxisymmetricJeans further provides velocity dispersions obtained by solving 
    the axisymmetric Jeans equations under the assumption of separability in cylindrical coordinates 
    (a.k.a. Jeans Anisotropic Model).
*/

#pragma once
#include "common.h"
#include "potential.h"
#include <gsl/gsl_spline.h>
#ifdef HAVE_INTERP2D
#include <interp2d_spline.h>
#endif

namespace smile{

/** Class for representing an axisymmetric density model.
    It takes an instance of CDensity class which provides a 3d mass model;
    if this model is not axisymmetric itself then it is averaged over azimuthal angle.
    Then it constructs auxiliary splines that interpolate the one-dimensional 
    mass profiles in R and z directions, and provides functions for their fast evaluation.
*/
class CMassModelAxisymmetric
{
public:
    /// construct the axisymmetric approximation to a given density model
    CMassModelAxisymmetric(const CDensity* density, size_t ngrid_R, size_t ngrid_z, bool initSplineMass=true);
    /// destroy object, deallocate memory
    ~CMassModelAxisymmetric();
    /// interpolated mass in a cylinder of radius R (extending to infinity in z direction)
    double mass_R(double R) const;
    /// find radius R containing a given mass in the cylindrical volume
    double find_R_mass(double m) const;
    /// interpolated surface density (integral of rho over z=-infinity..+infinity) as a function of radius R
    double Sigma(double R) const;
    /// interpolated mass in a slice from -z to +z, extending to infinity in R direction
    double mass_z(double z) const;
    /// find value of z such that the mass of the slice -z..+z equals the given amount
    double find_z_mass(double m) const;
    /// return the estimate of total mass
    double getTotalMass() const { return totalMass; }
protected:
    vectord grid_R, grid_z;  ///< grid nodes in R and z directions
    double totalMass;        ///< total mass of the original density model
    gsl_spline* spl_mass_R;  ///< interpolating spline for the enclosed mass in a cylindrical radius R
    gsl_spline* spl_mass_z;  ///< interpolating spline for the enclosed mass in a slice in z direction
};

#ifdef HAVE_INTERP2D
/** An extension of axisymmetric model which uses anisotropic Jeans equations 
    to compute velocity dispersions in cylindrical coordinates.
    The solution is specified by a single parameter -- the meridional velocity anisotropy coefficient 
    \f$ \beta_m = 1-\sigma_z^2/\sigma_R^2 \f$, which is constant with radius;
    furthermore, the velocity dispersion tensor is assumed to be aligned with cylindrical coordinates.
    The density of tracer stars whose velocity dispersions we need may be different from 
    the potential in which they move.
*/
class CMassModelAxisymmetricJeans: public CMassModelAxisymmetric
{
public:
    /// construct the axisymmetric approximation for a given potential model
    CMassModelAxisymmetricJeans(const CPotential* potential, size_t ngrid_R, size_t ngrid_z, double _beta_m=0, const CDensity* density=NULL);
    /// destroy object, deallocate memory
    ~CMassModelAxisymmetricJeans();
    /// velocity dispersion (1d) in the z direction (sigma_z^2) from axisymmetric Jeans equation
    double sigma_z(double R, double z) const { return spl_eval(R, z, spl_vmer); };
    /// mean squared velocity in the azimuthal direction from axisymmetric Jeans equations
    double vphisq_av(double R, double z) const { return spl_eval(R, z, spl_vphi); };
private:
    double beta_m;              ///< meridional velocity anisotropy parameter
    interp2d_spline* spl_vmer;  ///< 2d spline used for interpolating the velocity dispersion in meridional plane
    interp2d_spline* spl_vphi;  ///< 2d spline for interpolating the mean-square azimuthal velocity
    /// compute the interpolating spline with the appropriate scaling
    double spl_eval(double R, double z, interp2d_spline* spl) const;
};
#endif

/** A helper class for representing axisymmetrized versions of the given potential and/or density model.
    The class is a descendant of CPotential, but may also be used for density model only, in which case 
    the potential and force functions return zero.
    The density, potential and forces are computed as averages over azimuthal angle of the original model.
*/
class CPotentialAxisymmetrized: public CPotential
{
public:
    /** The constructor call determines the usage mode of the class: 
        if _potential==NULL, then _density is used for computing the density only, 
        otherwise _density parameter is ignored and _potential is used for density as well. */
    CPotentialAxisymmetrized(const CPotential* _potential, const CDensity* _density=NULL) :
      CPotential(), potential(_potential), density(_potential==NULL?_density:_potential) {};
    virtual CPotential* clone() const { return new CPotentialAxisymmetrized(*this); };
    virtual POTENTIALTYPE PotentialType() const { return potential->PotentialType(); };
    virtual const char* PotentialName() const  { return potential->PotentialName(); };
    virtual SYMMETRYTYPE symmetry() const { return ST_AXISYMMETRIC; };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    //virtual bool finite() const { return potential->finite(); }
    virtual double Mass(const double r) const { return density->Mass(r); }
private:
    const CPotential* potential; ///< the original potential model, or NULL if only the density is used
    const CDensity* density;     ///< the original density model
};

}; // namespace
