#ifdef HAVE_CVXOPT
#include <cvxopt.h>  // C interface to LP/QP solver written in Python
#endif
#include "optimization.h"
#include <QFile>
#include <QTextStream>
#include <QProcess>
#ifdef HAVE_GLPK
#include <glpk.h>    // GNU linear programming kit - optimization problem solver
#include <setjmp.h>  // to handle errors
#endif
#include "utils.h"

namespace smile{

#ifdef HAVE_GLPK
/*------- LP solver based on GNU linear programming kit -------*/

/// helper routine to capture debug and progress messages from GLPK and redirect them to my_message()
int glpk_output_hook(void * tmp, const char *data)
{
    std::string* infostr = static_cast<std::string*>(tmp);
    infostr->append(data);
    if(infostr->find('\r') || infostr->find('\n'))
    {
        size_t ind=infostr->find_first_of("\r\n");
        if(ind!=std::string::npos) infostr->resize(ind);
        my_message("COptimizationSolverGLPK", *infostr);
        infostr->clear();
    }
    return 1;
}

static jmp_buf err_buf;
/// change the default behaviour on GLPK error to something that doesn't crash program right away
void glpk_error_harmless(void*)
{
    longjmp(err_buf, 1);
}

template<typename NumT> int COptimizationSolverGLPK<NumT>::callSolver(
    const CMatrix<NumT> *linearMatrix,
    const std::vector<NumT> &rhs,
    const std::vector<NumT> &maxSolValue,
    const std::vector<NumT> &objFuncLinear,
    const CMatrix<NumT> * /*objFuncQuadratic*/,  // unused
    std::vector<NumT> *sol) const
{
    try{
        // to redirect messages from GLPK to the internal SMILE message routine
        std::string infostr;
        glp_term_hook(glpk_output_hook, &infostr); 
        // error handling
        glp_error_hook(glpk_error_harmless, NULL);
        if(setjmp(err_buf)) { 
            // this trick saves the current execution state on the first call, and if an error occurs 
            // inside GLPK then the modified error handler, the control is returned here to the following statement:
            glp_free_env();
            my_error("COptimizationSolverGLPK", "Internal error in GLPK solver, please save your data and restart program");
            return -2;  // most likely it was no memory error
        }
        size_t numVariables = linearMatrix->size1();
        size_t numConstraints = linearMatrix->size2();
        if(rhs.size()!=numConstraints || maxSolValue.size()!=numVariables || objFuncLinear.size()!=numVariables) 
        { my_error("COptimizationSolverGLPK", "Invalid call to solver"); return -999; }
        // count non-zero values in matrix to reserve the right amount of space in vectors containing sparse matrix coefs
        size_t nonZero=0;
        for(size_t v=0; v<numVariables; v++)
        {
            for(size_t c=0; c<numConstraints; c++)
            {
                double val=linearMatrix->at(v, c);
                if(val!=0) nonZero++;
                if(!gsl_finite(val))
                    return -99;  // invalid numeric format (need to handle here or the solver will crash without warning)
            }
        }
        glp_prob *lp;
        lp = glp_create_prob();
        glp_set_obj_dir(lp, GLP_MIN);
        glp_add_rows(lp, static_cast<int>(numConstraints));
        for(size_t c=0; c<numConstraints; c++)
            glp_set_row_bnds(lp, static_cast<int>(c+1), GLP_FX, rhs[c], rhs[c]);
        glp_add_cols(lp, static_cast<int>(numVariables));
        for(size_t v=0; v<numVariables; v++)
            if(maxSolValue[v]>0)
                glp_set_col_bnds(lp, static_cast<int>(v+1), GLP_DB, 0.0, maxSolValue[v]);
            else
                glp_set_col_bnds(lp, static_cast<int>(v+1), GLP_LO, 0.0, 0.0);
        {  // open a block so that temp variables are destroyed after it is closed
            std::vector<int> ic;
            std::vector<int> iv;
            std::vector<double> matrixval;
            ic.reserve(nonZero+1);  ic.push_back(0);
            iv.reserve(nonZero+1);  iv.push_back(0);
            matrixval.reserve(nonZero+1);  matrixval.push_back(0);
            for(size_t v=0; v<numVariables; v++)
            {
                for(size_t c=0; c<numConstraints; c++)
                {
                    double val=linearMatrix->at(v,c);
                    if(val!=0) {
                        ic.push_back(static_cast<int>(c+1));
                        iv.push_back(static_cast<int>(v+1));
                        matrixval.push_back(val);
                    }
                }
                if(objFuncLinear[v]!=0)
                    glp_set_obj_coef(lp, static_cast<int>(v+1), objFuncLinear[v]);
            }
            glp_load_matrix(lp, static_cast<int>(ic.size())-1, &(ic.front()), &(iv.front()), &(matrixval.front()));
        }
        my_message("COptimizationSolverGLPK", "Starting optimization");

        //int status = glp_simplex(lp, NULL);
        int status = glp_interior(lp, NULL);   // both methods give reasonable result, but the interior-point seems to be more efficient
        if(glp_ipt_status(lp)!=GLP_OPT) status=1;  // infeasible

        sol->assign(numVariables, 0);
        for(unsigned int v=0; v<numVariables; v++)
        {
            double val = glp_ipt_col_prim(lp, v+1);
            //double val = glp_get_col_prim(lp, v+1);
            sol->at(v) = val;
        }
        glp_delete_prob(lp); 
        return -status;  // 0 for success, <0 for failure

    }
    catch(const std::bad_alloc&) {
        return -2;
    }
    catch(...) { return -100; }
}

template<typename NumT> std::string COptimizationSolverGLPK<NumT>::errorDescription(int result) const
{
    switch(result)
    {
    case 0: return "Success"; break;
    case -1: return "Problem is infeasible"; break;
    case -2: return "Not enough memory in GLPK solver"; break;
    case -99: return "Invalid numeric format"; break;
    case GLP_EFAIL: return "Empty problem in GLPK solver"; break;
    case GLP_ENOCVG: return "Bad convergence in GLPK solver"; break;
    case GLP_EITLIM: return "Number of iterations exceeded limit in GLPK solver"; break;
    case GLP_EINSTAB: return "Numerical instability in GLPK solver"; break;
    default: return "Unknown error in GLPK solver";
    }
}
// explicit instantiation
template class COptimizationSolverGLPK<smNumType>;
#endif

/*------- B P M P D  - based LP/QP solver -------*/
void COutputHookQProcess::readLineOutput()
{
    QString out=proc->readAllStandardOutput();
    std::string str(out.trimmed().toStdString());
    if(!str.empty()) my_message("COptimizationSolverBPMPD", str);
};

// format literal for sprintf corresponding to size_t
#define FMT "zu"
template<typename NumT> int COptimizationSolverBPMPD<NumT>::callSolver(
    const CMatrix<NumT> *linearMatrix,
    const std::vector<NumT> &rhs,
    const std::vector<NumT> &maxSolValue,
    const std::vector<NumT> &objFuncLinear,
    const CMatrix<NumT> *objFuncQuadratic,
    std::vector<NumT> *sol) const
{
    try{
        size_t numVariables = linearMatrix->size1();
        size_t numConstraints = linearMatrix->size2();
        if(rhs.size()!=numConstraints || maxSolValue.size()!=numVariables || objFuncLinear.size()!=numVariables || 
            (objFuncQuadratic!=NULL && (objFuncQuadratic->size1()!=numVariables || objFuncQuadratic->size2()!=numVariables))) 
        { my_error("COptimizationSolverBPMPD", "Invalid call to solver"); return -999; }
        size_t numVarActive = 0;
        // first write out MPS file
        my_message("COptimizationSolverBPMPD", "Writing file");
        size_t numCoefs=linearMatrix->numElems();  // if it is large enough, display progress in writing data:)
        const size_t displayProgressChunk=2000000;
        size_t numProgressChunks=(numCoefs>=5*displayProgressChunk ? numCoefs/displayProgressChunk : 1);
        size_t progressVarChunk=numVariables/numProgressChunks;

        QString dataFileName = QString::fromStdString(tempDir)+"data.mps";
        QFile file(dataFileName);
        file.open(QFile::WriteOnly | QFile::Text);
        QTextStream out(&file);
        QString Str;
        out << "NAME  test" << endl;
        out << "ROWS" << endl << " N  OBJFUN" << endl;
        for(size_t c=0; c<numConstraints; c++)
            out << Str.sprintf(" E  R%07"FMT, c+1) << endl;
        for(size_t v=0; v<numVariables; v++)
            if(maxSolValue[v]>0)
                out << Str.sprintf(" L  W%07"FMT, v+1) << endl;
        out << "COLUMNS";
        int parity=0;
        for(size_t v=0; v<numVariables; v++)
        {
            if(v>0 && v%progressVarChunk==0)
                my_message("COptimizationSolverBPMPD", "Writing file.. "+
                convertToString(static_cast<int>(v*100/numVariables))+"%");
            bool active=false;
            if(objFuncLinear[v]!=0)
            {
                out << endl << Str.sprintf("    C%07"FMT"  OBJFUN    %12G ", v+1, static_cast<double>(objFuncLinear[v]));
                parity=1;
                active=true;
            }
            if(maxSolValue[v]>0)
            {
                if(parity==0) 
                    out << endl << Str.sprintf("    C%07"FMT, v+1);
                out << Str.sprintf("  W%07"FMT"             1 ", v+1);
                parity=1-parity;
                active=true;
            }
            for(size_t c=0; c<numConstraints; c++)
            {
                double val=static_cast<double>(linearMatrix->at(v,c));
                if(!gsl_finite(val)) {  // NaN or other incorrect value, shouldn't normally happen
                    file.close();
                    file.remove();
                    return -99;
                }
                if(val!=0)
                {
                    if(parity==0) 
                        out << endl << Str.sprintf("    C%07"FMT, v+1);
                    out << Str.sprintf("  R%07"FMT"  %12G ", c+1, val);
                    parity = 1-parity;
                    active=true;
                }
            }
            parity=0;
            if(active) numVarActive++;
        }
        out << endl;
        out << "RHS";
        parity=0;
        for(size_t c=0; c<numConstraints; c++)
        {
            if(parity==0)
                out << endl << "    RHS1    ";
            out << Str.sprintf("  R%07"FMT"  %12G ", c+1, static_cast<double>(rhs[c]));
            parity=1-parity;
        }
        for(size_t v=0; v<numVariables; v++) {
            if(maxSolValue[v]>0)
            {
                if(parity==0)
                    out << endl << "    RHS1    ";
                out << Str.sprintf("  W%07"FMT"  %12G ", v+1, static_cast<double>(maxSolValue[v]));
                parity=1-parity;
            }
        }
        if(objFuncQuadratic!=NULL)
        {   // add Quadratic optimization block
            out << endl << "QUADOBJ";
            if(objFuncQuadratic->isDense())
            {
                for(size_t v1=0; v1<numVariables; v1++)
                    for(size_t v2=0; v2<numVariables; v2++)
                    {
                        double val=static_cast<double>(objFuncQuadratic->at(v1, v2));
                        if(val!=0)
                            out << endl << Str.sprintf("    C%07"FMT"  C%07"FMT"  %12G", v1+1, v2+1, val);
                    }
            } else {
                size_t numElem=objFuncQuadratic->numElems();
                for(size_t i=0; i<numElem; i++)
                {
                    size_t v1, v2;
                    double val=static_cast<double>(objFuncQuadratic->getElem(i, &v1, &v2));
                    if(val!=0)
                        out << endl << Str.sprintf("    C%07"FMT"  C%07"FMT"  %12G", v1+1, v2+1, val);
                }
            }
        }
        out << endl << "ENDATA" << endl;
        file.close();

        // then run BPMPD
        my_message("COptimizationSolverBPMPD", "Starting optimization");
        QProcess process;
        COutputHookQProcess outputHook(&process);
        QObject::connect(&process, SIGNAL(readyReadStandardOutput()), &outputHook, SLOT(readLineOutput()));
        process.start(QString::fromStdString(execName), QStringList(dataFileName) );
        if(!process.waitForStarted()) return -3;  // report error
        QProcess::ProcessState st=process.state();
        QProcess::ProcessError er=process.error();
        if(st!=QProcess::Running) return -3-er;  // report error
        process.waitForFinished(-1);
#ifndef DEBUGPRINT
        file.remove();
#endif
        // finally, read results
        my_message("COptimizationSolverBPMPD", "Reading results");
        QFile fileo(QString::fromStdString(tempDir)+"data.out");
        int status=-5;
        if(fileo.open(QIODevice::ReadOnly))
        {
            sol->assign(numVariables, 0);
            QTextStream in(&fileo);
            QString S;
            unsigned int numColumnsResult=0;
            do{
                S = in.readLine();
                if(S.contains("infeasib")) return -1;
                if(S.contains("General memory limit")) return -2;
                if(S.contains("Limit of the evaluation version is reached")) return -6;  // BPMPD: sadly we won't know the answer...
                bool ok=false;
                QStringList fields=S.trimmed().split(QRegExp("\\s+"));
                unsigned int c=fields[0].mid(1,7).toInt(&ok);
                if(fields[0].mid(0,1)=="C" && ok && c>0 && c<=numVariables)
                {
                    double val=fields[1].toDouble(&ok);
                    if(ok)
                    {
                        sol->at(c-1) = static_cast<NumT>(val);
                        numColumnsResult++;
                    }
                }
            }
            while(!in.atEnd());
            status=(numColumnsResult==numVarActive ? 0 : -7);
            fileo.close();
#ifndef DEBUGPRINT
            fileo.remove();
#endif
        }
        return status;
    }
    catch(...) { return -100; }
}

template<typename NumT> std::string COptimizationSolverBPMPD<NumT>::errorDescription(int result) const
{
    switch(result)
    {
    case -1: return "Problem is infeasible"; break;
    case -2: return "Not enough memory in BPMPD solver"; break;
    case -3: return "BPMPD: Process failed to start"; break;
    case -4: return "BPMPD: Process crashed unexpectedly"; break;
    case -5: return "BPMPD: Result data file not found"; break;
    case -6: return "BPMPD: Problem size exceeds limit of evaluation version"; break;
    case -7: return "BPMPD: Unrecognized output format"; break;
    case -99:return "Invalid numeric format"; break;
    default: return "BPMPD: Unknown error";
    }
}
// explicit instantiation
template class COptimizationSolverBPMPD<smNumType>;

#ifdef HAVE_CVXOPT

/// flag set if Python reports about an out-of-memory error
static bool memoryError=false;
/// helper routine to record messages from Python interpreter and pass them to my_message()
static PyObject* redirection_stdoutredirect(PyObject *, PyObject *args)
{
    const char *pchar;
    if(!PyArg_ParseTuple(args, "s", &pchar))
        return NULL;
    std::string info(pchar);
    if(info.size()>2)      // prevent empty strings
        my_message("COptimizationSolverCVXOPT", info);  // pass string to console/gui of the main program
    if(info.find("MemoryError")!=std::string::npos)
        memoryError=true;  // record the out-of-memory condition
    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef RedirectionMethods[] = {
    {"stdoutredirect", redirection_stdoutredirect, METH_VARARGS,
        "stdout redirection helper"},
    {NULL, NULL, 0, NULL}
};

/// helper routine to create CVXOPT-compatible dense or sparse matrix
template<typename NumT> PyObject* initPyMatrix(const CMatrix<NumT> *srcmatrix)
{
    assert(srcmatrix);
    if(srcmatrix->isDense())
    {   // dense matrices are initialized item-by-item
        size_t numcol=srcmatrix->size1(), numrow=srcmatrix->size2();
        PyObject *pyMatrix = (PyObject*)Matrix_New(numrow, numcol, DOUBLE);
        if(!pyMatrix) return NULL;
        for(size_t i1=0; i1<numcol; i1++)
        {
            for(size_t i2=0; i2<numrow; i2++)
            {
                NumT value=srcmatrix->at(i1, i2);
                if(!gsl_finite(value)) {  // incorrect value
                    Py_XDECREF(pyMatrix); 
                    return NULL;
                }
                MAT_BUFD(pyMatrix)[i1*numrow+i2] = value;
            }
        }
        return pyMatrix;
    } else {  // To initialize sparse matrices, additional 3 vectors are allocated which hold col, row, and value triplets.
        size_t numElem = srcmatrix->numElems();
        PyObject *col= (PyObject*)Matrix_New(numElem, 1, INT);
        PyObject *row= (PyObject*)Matrix_New(numElem, 1, INT);
        PyObject *val= (PyObject*)Matrix_New(numElem, 1, DOUBLE);
        if(!col || !row || !val) {
            Py_XDECREF(col);
            Py_XDECREF(row);
            Py_XDECREF(val);
            return NULL;
        }
        // init quadratic penalty matrix values(triplets)
        for(size_t ind=0; ind<numElem; ind++)
        {
            size_t i1, i2;
            NumT value=srcmatrix->getElem(ind, &i1, &i2);
            if(!gsl_finite(value)) {  // incorrect value
                Py_XDECREF(col);
                Py_XDECREF(row);
                Py_XDECREF(val);
                return NULL;
            }
            MAT_BUFI(col)[ind]=i1;
            MAT_BUFI(row)[ind]=i2;
            MAT_BUFD(val)[ind]=value;
        }
        PyObject* pyMatrix = (PyObject*)SpMatrix_NewFromIJV(
            (matrix*)col, (matrix*)row, (matrix*)val, srcmatrix->size2(), srcmatrix->size1(), DOUBLE);
        Py_DECREF(col);
        Py_DECREF(row);
        Py_DECREF(val);
        return pyMatrix;
    }
}

template<typename NumT> int COptimizationSolverCVXOPT<NumT>::callSolver(
    const CMatrix<NumT> *linearMatrix,
    const std::vector<NumT> &rhs,
    const std::vector<NumT> &maxSolValue,
    const std::vector<NumT> &objFuncLinear,
    const CMatrix<NumT> *objFuncQuadratic,
    std::vector<NumT> *sol) const
{
    assert(linearMatrix);
    try{
        bool QP = objFuncQuadratic != NULL;  // whether to solve linear or quadratic problem
        size_t numVariables = linearMatrix->size1();
        size_t numConstraints = linearMatrix->size2();
        if(rhs.size()!=numConstraints || maxSolValue.size()!=numVariables || objFuncLinear.size()!=numVariables ||
            (QP && (objFuncQuadratic->size1()!=numVariables || objFuncQuadratic->size2()!=numVariables)))
        { 
            my_error("COptimizationSolverCVXOPT", "Invalid call to solver");
            return -999;
        }
        memoryError=false;
        Py_Initialize();
        Py_InitModule("redirection", RedirectionMethods);
        PyRun_SimpleString(
            "import redirection\n"
            "import sys\n"
            "class StdoutCatcher:\n"
            "    def write(self, stuff):\n"
            "        redirection.stdoutredirect(stuff)\n"
            "sys.stdout = StdoutCatcher()\n"
            "sys.stderr = StdoutCatcher()\n");
        if(!modulePath.empty()) {
            PyRun_SimpleString(("sys.path.append('"+modulePath+"')\n").c_str());
        }

        if (import_cvxopt() < 0) {   // error importing cvxopt
            //Py_Finalize();
            return -21;
        }

        // import cvxopt.solvers
        PyObject *solvers = PyImport_ImportModule("cvxopt.solvers");
        if (!solvers) {   // error importing cvxopt.solvers
            //Py_Finalize();
            return -22;
        }

        // get reference to solvers.solveqp
        PyObject *qp = PyObject_GetAttrString(solvers, QP ? "qp" : "lp");
        if (!qp) {   // error referencing cvxopt.solvers.qp
            Py_DECREF(solvers);
            //Py_Finalize();
            return -23;
        }

        // count upper inequality constraints
        size_t numConsIneq = numVariables;
        for(size_t i=0; i<numVariables; i++)
            if(maxSolValue[i]>0) numConsIneq++;  // one or two inequality constraints per variable

        // create matrices
        PyObject *objectiveQuad = QP ? initPyMatrix(objFuncQuadratic) : NULL;
        PyObject *coefMatrix    = initPyMatrix(linearMatrix);
        PyObject *objectiveLin  = (PyObject*)(Matrix_New(numVariables, 1, DOUBLE));
        PyObject *consIneq      = (PyObject*)(SpMatrix_New(numConsIneq, numVariables, numConsIneq, DOUBLE));
        PyObject *consIneqRhs   = (PyObject*)(Matrix_New(numConsIneq, 1, DOUBLE));
        PyObject *coefRhs       = (PyObject*)(Matrix_New(numConstraints, 1, DOUBLE));
        PyObject *pArgs = PyTuple_New(QP ? 6 : 5);
        if(!objectiveLin  || (QP && !objectiveQuad) || !consIneq || !consIneqRhs || !coefMatrix || !coefRhs || !pArgs) {
            Py_DECREF(qp);
            Py_DECREF(solvers);
            Py_XDECREF(objectiveLin); 
            Py_XDECREF(objectiveQuad); 
            Py_XDECREF(consIneq); 
            Py_XDECREF(consIneqRhs); 
            Py_XDECREF(coefMatrix); 
            Py_XDECREF(coefRhs); 
            Py_XDECREF(pArgs);
            //Py_Finalize();
            return -24;
        }

        // init objective func, linear part
        for(size_t v=0; v<numVariables; v++)
        {
            MAT_BUFD(objectiveLin)[v] = objFuncLinear[v];
        }
        // init inequality constraints
        for(size_t v=0, c=0; v<numVariables; v++, c++)
        {
            SP_ROW(consIneq) [c]= v;
            SP_VALD(consIneq)[c]= -1.;
            SP_COL(consIneq) [v]=c;
            MAT_BUFD(consIneqRhs)[v]= 0;
            if(maxSolValue[v]>0) { // add upper constraint
                SP_ROW(consIneq) [c+1]= c-v+numVariables;
                SP_VALD(consIneq)[c+1]= 1.;
                MAT_BUFD(consIneqRhs)[c-v+numVariables] = maxSolValue[v];
                c++;
            }
        }
        SP_COL(consIneq)[numVariables]=numConsIneq;
        // init rhs
        for(unsigned int c=0; c<numConstraints; c++)
        {
            MAT_BUFD(coefRhs)[c] = rhs[c];
        }

        /* pack matrices into an argument tuple - references are stolen*/
        if(QP) PyTuple_SetItem(pArgs, 0, objectiveQuad);
        PyTuple_SetItem(pArgs, QP?1:0, objectiveLin);
        PyTuple_SetItem(pArgs, QP?2:1, consIneq);
        PyTuple_SetItem(pArgs, QP?3:2, consIneqRhs);
        PyTuple_SetItem(pArgs, QP?4:3, coefMatrix);
        PyTuple_SetItem(pArgs, QP?5:4, coefRhs);
        my_message("COptimizationSolverCVXOPT", "Starting optimization");
        PyObject *solver = PyObject_CallObject(qp, pArgs);
        if (!solver) {
            PyErr_Print();
            Py_DECREF(pArgs);
            Py_DECREF(qp);
            Py_DECREF(solvers);
            //Py_Finalize();
            return memoryError? -2 : -26; // flag set internally in debugprinting indicating whether Python complained about out-of-memory
        }

        PyObject *status = PyDict_GetItemString(solver, "status");
        bool feasible = std::string(PyString_AsString(status)) == "optimal";
        PyObject *result = PyDict_GetItemString(solver, "x");

        sol->assign(numVariables, 0);
        for(unsigned int o=0; o<numVariables; o++)
            (*sol)[o] = MAT_BUFD(result)[o];

        Py_DECREF(status);
        Py_DECREF(result);
        Py_DECREF(pArgs);
        //Py_DECREF(solver);  // uncommenting this causes crash on repeated calls to solver..??
        Py_DECREF(qp);
        Py_DECREF(solvers);
        //Py_Finalize();  // commented out because repeated initialization of python seems to crash on 3rd-4th occasion...
        return feasible ? 0 : -1;
    }
    catch(const std::bad_alloc&) {
        //Py_Finalize();
        return -2;
    }
    catch(...) {
        //Py_Finalize();
        return -23;  // hmmm
    }
}

template<typename NumT> std::string COptimizationSolverCVXOPT<NumT>::errorDescription(int result) const
{
    switch(result)
    {
    case -1: return "Problem is infeasible"; break;
    case -2:  return "Not enough memory in CVXOPT solver"; break;
    case -21: return "error importing CVXOPT python module"; break;
    case -22: return "error importing CVXOPT.solvers python module"; break;
    case -23: return "error initializing CVXOPT python module"; break;
    case -24: return "error allocating matrices "; break;
    case -26: return "CVXOPT solver returned error"; break;
    default: return "CVXOPT: Unknown error";
    }
}
// explicit instantiation
template class COptimizationSolverCVXOPT<smNumType>;
#endif

}
