/** \file    orbitlib.h
    \brief   Classes for handling the orbit library and creating an Nbody model
    \author  EV
    \date    2009-2014

    This file defines classes for managing the orbit library, i.e. a collection of 
    orbit initial conditions and results of integration and analysis, used for 
    creating a frequency map or a Schwarzschild model. 

    The main class, smile::COrbitLibrary, implements methods for storing initial 
    conditions, and for creating an Nbody model out of the orbit library.
    An interface class smile::COrbitDesc is a container for the initial data and 
    results of integration of a single orbit.
    A multimass scheme for Nbody models is implemented by a couple of classes 
    for mass refinement.

*/

#pragma once
#include "common.h"

namespace smile{

struct CConfigOrbit;  // forward declaration
class COrbitRuntimeTrajSampleCreator;  // forward declaration

/** An utility class that keeps all necessary data to start orbit integration.
    These data are: potential, library of runtime function creators, and common parameters for integrator.
    ATTENTION:  unlike (almost?) all other classes in SMILE, this class does not simply copy pointers 
    to potential, creators and parameters, but takes over the ownership. This means that 
    - in the constructor, potential passed as parameter gets cloned, and in the destructor 
      the cloned potential is deleted (the original potential is unaffected);
    - the structure CConfigOrbit is copied fieldwise in the constructor;
    - most importantly, pointers to runtime function creators are added by a special method, 
      to underline that these pointers are "taken over" by this object and will be destroyed in the destructor.
    This is done to facilitate handling copies of data in CSmileCore (only this object needs to be 
    created before and destroyed after the orbit integration or orbit library integration is performed).
*/
class COrbitLauncher
{
public:
    COrbitLauncher(const CPotential* _potential, const CConfigOrbit& _configOrbit);
    ~COrbitLauncher();
    void addCreator(const CBasicOrbitRuntimeFncCreator* crt) { creators.push_back(crt); }
    /// create orbit with given initial conditions and parameters stored in this class instance
    template<typename NumT> COrbit* createOrbit(const CPosVelPoint<NumT>& initCond, NumT intTime, NumT maxTime) const;
    /// return pointer to the potential
    const CPotential* getPotential() const { return potential; }
    /// return pointer to the config
    const CConfigOrbit* getConfig() const { return configOrbit; }
private:
    const CPotential* potential;
    const CConfigOrbit* configOrbit;
    vectorRuntimeFncCreators creators;
};

/** Container for orbit initial conditions and information objects.
    This class serves as an interface between a single COrbit and the entire COrbitLibrary.
    It keeps the initial data for an orbit, and internally creates an instance 
    of COrbit for the integration and analysis; then it records the results in the form 
    of an array of CBasicInformation and deletes the COrbit to free up memory.

    A collection of this objects constitutes the COrbitLibrary.
*/
class COrbitDesc
{
public:
    /// lists all possible states of orbit integration
enum ORBITSTATE{
    OS_INITIALIZED,      ///< the object is created but only contains initial data
    OS_PREPARING,        ///< the run() method is called
    OS_RUNNING,          ///< the internal COrbit::run() is in progress
    OS_NEEDTOTERMINATE,  ///< the user requested to stop orbit integration
    OS_DONE,             ///< integration completed, all information objects are stored (also set when all data is loaded from a file).
    OS_FAILED            ///< if the integration failed for some reason (did not produce any orbit)
};
    /// Create the object for the given initial conditions, without the results of orbit integration (OS_INITIALIZED)
    template<typename NumT> COrbitDesc(
        const CPosVelPoint<NumT> &_initCond, const NumT _intTime, const NumT _maxTime, 
        const double _weight, const double _weightPrior, const vectorInformation* _info=NULL);

    /// destroy all information containers
    ~COrbitDesc();

    /// create an internal COrbit instance using the launcher, and run orbit integration and analysis
    void run(const COrbitLauncher* launcher);

    /// if running, immediately stop computation (set the state to OS_NEEDTOTERMINATE)
    void halt();

    // get/set functions

    /// return the initial conditions for the orbit
    const CPosVelPoint<float>& getInitCond() const { return initCond; };

    /// assign initial conditions
    void setInitCond(const CPosVelPoint<float>& ic) { initCond=ic; }

    // return pointer to the potential
    //   const CPotential* getPotential() const { return potential; }

    /// return the integration time, whether the actual one if orbit is done, or the intended one if only initialized
    float getIntTime() const { return intTime; };

    /// assign the integration time
    void setIntTime(float time) { intTime=time; };

    /// return the maximum time for adaptive integration
    float getMaxTime() const { return maxTime; };

    /// assign maximum integration time
    void setMaxTime(float time) { maxTime=time; };

    /// return the number of CBasicInformation objects
    size_t getInfoNum() const { return info.size(); };

    /// return a pointer to the given information object (a private data member!)
    const CBasicInformation* getInfo(size_t index) const 
    { return index<info.size()? info.at(index) : NULL; };

    /// convenience function to get a pointer to information of a given type, or NULL if it doesn't exist
    const CBasicInformation* getInfoByType(CBasicInformation::INFOTYPE InfoType) const {
        for(size_t i=0; i<info.size(); i++)
            if(info[i]->infoType()==InfoType) return info[i];
        return NULL; };

    /// returns string representation of all information objects associated with the orbit, 
    /// plus the weight and integration time
    std::string toString() const;

    /// return orbit weight in the Schwarzschild model
    double getWeight() const { return weight; };

    /// assign the orbit weight
    void setWeight(double w) { weight=w; };

    /// return orbit weight prior in the Schwarzschild model
    double getWeightPrior() const { return weightPrior; };

    // assign the orbit weight prior
    void setWeightPrior(double w) { weightPrior=w; };

    /// return the state of orbit integration (init/running/done)
    ORBITSTATE getState() const { return state; };

    /// set the state to OS_PREPARING if it is going to be started by the orbit library class,
    /// or reset it to OS_INITIALIZED if it was completed.
    void setState(ORBITSTATE _state) 
    { if((state==OS_INITIALIZED && _state==OS_PREPARING)||(state==OS_DONE && _state==OS_INITIALIZED)) state=_state; };
private:
    /*const*/ CPosVelPoint<float> initCond; ///< orbit initial conditions
    COrbit* orbit;                      ///< pointer to the internal object created in run()
    float intTime, maxTime;             ///< integration and max integration time
    vectorInformation info;             ///< array of CBasicOrbitInformation objects
    double weight;                      ///< weight of the orbit in the Schwarzschild model
    double weightPrior;                 ///< prior on orbit weight (for proper normalization of entropy; default 0 means Mtotal/Norbits)
    volatile ORBITSTATE state;          ///< current state of affairs
};

/** The collection of orbits.
    Contains the array of COrbitDesc objects, methods to generate initial conditions 
    for the orbit library, to get the statistical information about the orbit ensemble, 
    and to create an Nbody model from the orbit library.
*/
class COrbitLibrary
{
public:
    /// Construct an empty orbit library
    COrbitLibrary() : launcher(NULL), nextOrbitInd(0) {};

    /// Destroy the orbit library and delete all orbit data
    ~COrbitLibrary();

    /// Assign or modify initial conditions for orbit list
    void assignInitialConditions(
        const CPointMassSet<float>* points, ///< array of points to be used as initial conditions; if NULL then retain the existing OrbitList
        const COrbitLauncher* newLauncher,  ///< pointer to the instance of orbit launcher (which contains potential, runtime fnc creator lib, and common parameters)
        double intTimeInPeriods,            ///< integration time in units of dynamical time
        double intTimeMaxAdaptive,          ///< maximal integration time (if 0, equal to the previous)
        double intTimeMinAbs                ///< minimum integration time in global time units (not energy-dependent units of dynamical time)
    );

    /// if integration of the whole library was aborted by user, delete unfinished orbits
    void removeUnused();

    /// reset the orbit integration state to OS_INITIALIZED and reset the counter of complete orbits
    void rewind();

    /** try to create Nbody model from orbit library (with trajectory sample data).
        If number of sampling points was not enough for some orbits, put duplicate points 
        in the output for orbits with insufficient sampling, mark those orbits as unstarted 
        and return list of orbits for reintegration.
        \param[in] numPoints  is number of points to be created in the Nbody snapshot;
        \param[in] massRefineFactor  may be used to generate a multimass model: 
                   the mass of points sampled from a given orbit is proportional to 
                   the number returned by this function for that orbit), 
                   if this parameter is NULL then just assign equal mass to all points.
        \param[in,out] result  is a pointer to the point mass set that is created in this routine;
                   if this parameter points to a NULL pointer then a new CPointMassSet* is allocated, 
                   otherwise the existing snapshot is replaced with the new points;
        \param[in,out] trajSampleCreator  is a pointer to the runtime function creator that manages 
                   collection of sampling points from trajectories.
                   If any of the orbit lacks enough sampling points, it will be added to 
                   the list of orbits for reintegration, and the runtime function will be 
                   informed of the required number of sampling points.
        \return some statistics about number of orbits and points.
    **/
    std::string exportNbody(
        size_t numPoints,
        const CBasicOrbitFilteringFnc* massRefineFactor,
        CPointMassSet<float>* *result,
        COrbitRuntimeTrajSampleCreator* trajSampleCreator
    ); 

    // get functions

    /// return total number of orbits in the library (both completed and unstarted)
    size_t size() const { return OrbitList.size(); };

    /// returns sum of all orbit weights (if they were assigned)
    double totalMass() const;

    /// return pointer to the orbit launcher (why would it be necessary?)
    const COrbitLauncher* getLauncher() const { return launcher; }

    /// return a pointer to a particular orbit descriptor
    const COrbitDesc* getOrbitDesc(size_t index) const { return OrbitList.at(index); };
    COrbitDesc* getOrbitDesc(size_t index) { return OrbitList.at(index); };

    /// find how many orbits are already finished
    size_t numComplete() const;

    /// return statistics about the orbit library.
    std::string getOrbitPopulation(
        const CBasicOrbitFilteringFnc* filter, ///< if not NULL, select only orbits for which this function returns nonzero
        const CBasicOrbitFilteringFnc* chaos,  ///< if not NULL, provide information separately for regular and chaotic orbits determined by this function
        bool useWeight                         ///< whether to use orbit weights in computing the fraction of orbits, or use equal weighting
    ) const;  ///< \return a string containing orbit population (5 most abundant orbit families with their combined number, possibly weighted).

    /** find an orbit that hasn't been started yet.
        search for the first orbit with state=OS_INITIALIZED and set it to OS_PREPARING.
        Should be called with external mutex/lock in a multi-threaded environment.
        \return the index of this orbit or -1 if none was found.
    */
    int findUnstartedOrbit();

    // set functions

    /// set new launcher
    void setLauncher(const COrbitLauncher* _launcher) { launcher=_launcher; }

    /// append an orbit descriptor to the library (used in loading from a file)
    void addOrbitDesc(COrbitDesc* const &od) { OrbitList.push_back(od); };

    /// integrate the orbit with the given index
    void runOrbit(size_t index) 
    { if(launcher!=NULL) OrbitList.at(index)->run(launcher); };

    /// stop all orbit integrations (call halt for all orbit descriptors)
    void halt() {
        for(std::vector<COrbitDesc*>::iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
            (*iter)->halt();
    }
private:
    const COrbitLauncher* launcher;
    std::vector<COrbitDesc*> OrbitList;       ///< array of orbit descriptors
    volatile int nextOrbitInd;
};

/// A helper object for computing the virial ratio for a set of point masses
template<typename NumT> class CPointMassSetHandler
{
public:
    CPointMassSetHandler(const CPointMassSet<NumT>* _points) : points(_points), state(PS_IDLE) {};
    /// compute the kinetic and potential energy and return them in these two pointers
    void computeVirialRatio(double* T, double* W);
    /// stop the computation
    void halt() { if(state==PS_RUNNING) state=PS_NEEDTOTERMINATE; };
private:
    /// internal state of the object (created/running/done)
enum PSTATE{
    PS_IDLE,
    PS_RUNNING,
    PS_NEEDTOTERMINATE,
    PS_DONE
};
    const CPointMassSet<NumT>* points;  ///< pointer to the array of bodies
    volatile PSTATE state;              ///< computation state
};

/** A mass refinement function is used to create an Nbody snapshot with unequal mass particles. 
    It ranks orbits into several bins, based on another evaluation function passed as a parameter; 
    mass factors of adjacent bins differ by a factor of two, and each bin contains the number of 
    orbits that also differ by a factor of two.
    Used in creating multimass Nbody models: one may wish to have a better mass resolution for 
    the inner parts of the model, based on orbit energy, or pericenter distance, or anything else.
*/
class CBinnedMassRefinementFnc: public CBasicOrbitFilteringFnc
{
public:
    /// initialize bin boundaries by sorting the output of evalFnc for all orbits in the library
    CBinnedMassRefinementFnc(const COrbitLibrary* orbitlib, size_t numBins, const CBasicOrbitFilteringFnc* _evalFnc);
    /// return the bin into which the orbit falls, using the pre-initialized array of bins
    virtual double eval(const COrbitDesc* orbit) const;
private:
    /// the evaluation function; orbits are sorted according to the value returned by this function
    const CBasicOrbitFilteringFnc* rankFnc;
    /// divide the entire orbit library into numBins; the bin boundaries 
    /// in terms of evaluation function values are stored in this array
    vectord binBoundary;
};

/// an example of evaluation function for multimass Nbody models, which sorts orbits in total energy
class CEnergyOrbitFilteringFnc: public CBasicOrbitFilteringFnc
{
public:
    explicit CEnergyOrbitFilteringFnc(double _Omega=0) : Omega(_Omega) {};
    virtual double eval(const COrbitDesc* orbit) const;   ///< return total energy of an orbit for the initial conditions in the inertial frame
private:
    const double Omega;  ///< frequency of figure rotation of the potential (need to compensate the value of Jacobi constant read from the orbit desc)
};

/// Another mass refinement function that aims to reduce the average mass of particles inside a given radius
class CRadiusMassRefinementFnc: public CBasicOrbitFilteringFnc
{
public:
    /// initialize bin boundaries by sorting the output of evalFnc for all orbits in the library
    CRadiusMassRefinementFnc(double _refinementRadius, double _massRatio) :
      refinementRadius(_refinementRadius), massRatio(_massRatio) {};
    /// return the mass multiplier for an orbit, based on the fraction of time it spends inside the given radius
    virtual double eval(const COrbitDesc* orbit) const;
private:
    double refinementRadius;  ///< increase the mass resolution within the given radius
    double massRatio;         ///< ratio between maximum and minumum allowed particle mass
};

}
