/** \file    utils.h
    \brief   several mathematical, statistical, string formatting and error reporting routines
    \author  EV
    \date    2009-2014

*/
#pragma once
#include <string>
#include <sstream>
#include <iomanip>
#include <vector>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_bspline.h>
#include <gsl/gsl_rng.h>

namespace smile{

/* ------- message and error reporting routines ------- */
// the following definitions ensure that function name will appear with full class qualifier in my_message calls
#ifdef __GNUC__
#define FUNCNAME __PRETTY_FUNCTION__
#else
#define FUNCNAME __FUNCTION__
#endif

/// a function that shows a given message string.
/// implementation is program-specific, may dump text to stderr or do something else.
/// \param[in] origin is the name of calling function
/// \param[in] message is the text to be displayed
typedef void show_message_type(const std::string &origin, const std::string &message);

/// global variable to the routine that displays errors (if it is NULL then nothing is done)
extern show_message_type* my_error_ptr;

/// global variable to the routine that shows information messages.
/// called by various time-consuming routines to display progress, if NULL then no messaging is done
extern show_message_type* my_message_ptr;

/// the interface routine for error reporting (redirects the call to my_error_ptr if it is defined)
inline static void my_error(const std::string &origin, const std::string &message)
{ if(my_error_ptr!=NULL) my_error_ptr(origin, message); }

/// the interface routine for message reporting (redirects the call to my_message_ptr if it is defined)
inline static void my_message(const std::string &origin, const std::string &message)
{ if(my_message_ptr!=NULL) my_message_ptr(origin, message); }

/// default message printing routine
void my_stderr_show_message(const std::string &origin, const std::string &message);

/** error reporting routine which is invoked from inside GSL. 
    By default GSL crashes program on any error, which is not fun. 
    To change this behaviour, add the following line at the start of the program:
      gsl_set_error_handler(&smile::gsl_verbose_error_handler);
*/
void gsl_verbose_error_handler (const char *reason, const char *file, int line, int gsl_errno);

/*------------- string functions ------------*/
/** The StringVariant class is a simple string-based variant implementation that allows
    the user to easily convert between simple numeric/string types.
*/
class StringVariant
{
private:
    std::string data;
public:
    StringVariant() : data() {}
    StringVariant(const std::string &src) : data(src) {};
    template<typename ValueType> StringVariant(ValueType val) {
        std::ostringstream stream;
        stream << val;
        data.assign(stream.str());
    };
    template<typename ValueType> StringVariant(ValueType val, unsigned int width) {
        std::ostringstream stream;
        stream << std::setprecision(width) << val;
        data.assign(stream.str());
    };
    template<typename ValueType> StringVariant& operator=(const ValueType val) {
        std::ostringstream stream;
        stream << val;
        data.assign(stream.str());
        return *this;
    };
    template<typename NumberType> NumberType toNumber() const {
        NumberType result = 0;
        std::istringstream stream(data);
        if(stream >> result)
            return result;
        else if(data == "yes" || data == "true")
            return 1;
        return 0;
    };
    bool toBool() const { return(data == "yes" || data == "Yes" || data == "true" || data == "True" || data == "t" || data == "1"); }
    double toDouble() const { return toNumber<double>(); }
    float toFloat() const { return toNumber<float>(); }
    int toInt() const { return toNumber<int>(); }
    std::string toString() const { return data; }
};

/// a shorthand function to convert a string to a number
template<typename NumberType, typename ValueType> NumberType convertTo(ValueType val) 
{ return StringVariant(val).toNumber<NumberType>(); }
    
/// a shorthand function to convert a string to a bool
template<typename ValueType> bool convertToBool(ValueType val) 
  { return StringVariant(val).toBool(); }

/// a shorthand function to convert a number to a string
template<typename ValueType> std::string convertToString(ValueType val) 
  { return StringVariant(val).toString(); }

/// a shorthand to convert a bool to a string
inline static const char* convertToString(bool val)
  { return val?"true":"false"; }

/// a shorthand function to convert a number to a string with a given precision
template<typename ValueType> std::string convertToString(ValueType val, unsigned int width) 
  { return StringVariant(val,width).toString(); }

/// Pretty-print: convert floating-point or integer numbers to a string of fixed length.
std::string pp(double num, unsigned int width);

/** routine that splits one line from a text file into several items.
    \param[in]  src -- string to be split
    \param[in]  delim -- array of characters used as delimiters
    \param[out] result -- pointer to an existing array of strings in which the elements will be stored
*/
void splitString(const std::string& src, const std::string& delim, std::vector<std::string> *result);

/// routine that checks if a string ends with another string
bool ends_with_str(const std::string& str, const std::string& end);

/// routine that compares two strings in a case-insensitive way
bool strings_equal(const std::string& str1, const std::string& str2);

/// overloaded routine that compares two strings in a case-insensitive way
bool strings_equal(const std::string& str1, const char* str2);

/*---------- math functions ------------*/

extern gsl_rng* randgen;  ///< an instance of global random number generator

/** initialize random number generator using current time */
void randomize();

/** get a random number uniformly distributed in [0..1] */
double getRandomNumber();

/** generates two random numbers with normal distribution (zero mean, dispersion=1) */
void getNormalRandomNumbers(double* num1, double* num2);

/** generates 2d or 3d random vector with unit length; if z==NULL then create 2d vector, else 3d */
void getRandomUnitVector(double* x, double* y, double* z);

/** generate an uniformly distributed random rotation matrix in 3d (output array should have 9 items) */
void getRandomRotationMatrix(double* matr3x3);

/** generates a grid with exponentially growing spacing. 
    x[k] = (exp(Z k) - 1)/(exp(Z) - 1), 
    and the value of Z is computed so the the 1st element is at xmin and last at xmax.
    \param[in]  nnodes -- total number of grid points
    \param[in]  xmin, xmax -- location of the first and the last node
    \param[in]  zeroelem -- if true, 0th node is at zero (otherwise at xmin)
    \param[out] grid -- pointer to an existing array which will be overwritten by this routine
*/
void createNonuniformGrid(size_t nnodes, double xmin, double xmax, bool zeroelem, std::vector<double> *grid);

/** creates an almost uniform grid so that each bin contains at least minbin points from input array.
    input points are in srcpoints array and MUST BE SORTED in ascending order (assumed but not cheched).
    *grid must point to an existing (but possibly uninitialized) vector which will have length at most gridsize. 
    NB: in the present implementation, the algorithm is not very robust and works well only for gridsize*minbin << srcpoints.size,
    assuming that 'problematic' bins only are found close to endpoints but not in the middle of the grid.
*/
void createAlmostUniformGrid(const std::vector<double> &srcpoints, size_t minbin, size_t *gridsize, std::vector<double> *grid);

/*---------- convenience function for numerically solving an equation ------*/
/** wrapper for GSL root solver running multiple iterations until the convergence is achieved.
    may additionally specify desired relative tolerance and default value to be returned 
    if endpoints don't enclose root. 
*/
double findRoot(gsl_function *F, double xlower, double xupper, double xdefault=0, double reltoler=1e-8);

/** wrapper for GSL minimization routine to find local minimum enclosed between xlower < xinit < xupper.
    xinit is the initial guess and the value to be returned in the case of error. 
*/
double findMin(gsl_function *F, double xlower, double xupper, double xinit, double reltoler=1e-8);

/** wrapper for GSL minimization routine to find local minimum enclosed between xlower < xupper.
    employed when the initial minimum location is not really known but suspected to exist in the interval.
    performs binary search to find the initial minimum location and then calls GSL routines to refine it.
    if the interval doesn't enclose minimum then one of the endpoints is returned. 
*/
double findMinGuess(gsl_function *F, double xlower, double xupper, double reltoler=1e-8);

/*---------- spline functions -----------*/
/** a variant of cubic spline with given values of first derivative at endpoints.
    these values are passed as two additional points in y[i] array 
    ( y[npoints] for leftmost, y[npoints+1] for rightmost).
    if one wants to leave the standard, natural boundary condition (second derivative is zero),
    then pass NaN as a correspoinding 1st derivative parameter. 
*/
extern const gsl_interp_type * gsl_interp_ccspline;

/// modified functions for spline evaluation that extrapolate last segment linearly beyond the definition region
double gsl_spline_evalx(const gsl_spline *spline, double x, gsl_interp_accel *a );
double gsl_spline_eval_derivx(const gsl_spline *spline, double x, gsl_interp_accel *a );
double gsl_spline_eval_deriv2x(const gsl_spline *spline, double x, gsl_interp_accel *a );

/// convenience function that evaluates cubic spline value, derivative and 2nd derivative at once 
/// (faster than doing it separately)
void cspline_eval_all(const gsl_spline* spline, double x, double* val, double* der, double* der2);

/// check if a cubic spline is everywhere monotonic
bool check_spline_monotonic(const gsl_spline* spline);

/// create a copy of a spline by allocating new memory to all internal arrays and copying their data
gsl_spline* copyspline(const gsl_spline* src);

/*--------- penalized spline approximation class --------*/

/** Penalized linear least-square fitting problem.
    Approximate the data series  {x[i], y[i], i=0..numDataPoints-1} 
    with spline defined at  {X[k], Y[k], k=0..numKnots-1} in the least-square sense, 
    possibly with additional penalty term for 'roughness' (curvature).

    Initialized once for a given set of x, X, and may be used to fit multiple sets of y 
    with arbitrary degree of smoothing \f$ \lambda \f$.

    Internally, the approximation is performed by multi-parameter linear least-square fitting:
    minimize 
      \f[ 
        \sum_{i=0}^{numDataPoints-1}  (y_i - \hat y(x_i))^2 + \lambda \int [\hat y''(x)]^2 dx,
      \f]
    where 
      \f[
        \hat y(x) = \sum_{p=0}^{numBasisFnc-1} w_p B_p(x) 
      \f]
    is the approximated regression for input data, 
    \f$ B_p(x) \f$ are its basis functions and \f$ w_p \f$ are weights to be found.

    Basis functions are b-splines with knots at X[k], k=0..numKnots-1; the number of basis functions is numKnots+2.
    Equivalently, the regression can be represented by clamped cubic spline with numKnots control points; 
    b-splines are only used internally.

    LLS fitting is done by solving the following linear system:
      \f$ (\mathsf{A} + \lambda \mathsf{R}) \mathbf{w} = \mathbf{z} \f$, 
    where  A  and  R  are square matrices of size numBasisFnc, 
    w and z are vectors of the same size, and \f$ \lambda \f$ is a scalar (smoothing parameter).

    A = C^T C, where C_ip is a matrix of size numDataPoints*numBasisFnc, 
    containing value of p-th basis function at x[i], p=0..numBasisFnc-1, i=0..numDataPoints-1.
    z = C^T y, where y[i] is vector of original data points
    R is the roughness penalty matrix:  \f$ R_pq = \int B''_p(x) B''_q(x) dx \f$.

    Internally several arrays are initialized, the largest being C_ip matrix;
    in case of out-of-memory error no work is done and status is set to AS_NOMEM.
*/
class CSplineApprox {
public: 
    /// status codes returned by status() function and by fitting routines. 
    /// Everything other than AS_OK or AS_SINGULAR means that return arrays are uninitialized
enum APPROXSTATUS { 
    AS_OK,           ///< everything is fine
    AS_NOMEM,        ///< not enough memory in initializing arrays
    AS_INVRANGE,     ///< invalid range of input x and X values (the former should lie within the latter)
    AS_NODATA,       ///< no y-values have been set
    AS_SINGULAR      ///< linear system is [near-]singular, in this case a backup SVD-based algorithm is used to fit but it doesn't deal with smoothing
};
    /// initialize workspace for xvalues=x, knots=X in the above formulation.
    /// knots must be sorted in ascending order, and all xvalues must lie between knots.front() and knots.back()
    CSplineApprox(const std::vector<double> &_xvalues, const std::vector<double> &_knots);
    ~CSplineApprox();

    /// return status of the object
    APPROXSTATUS status() const { return mystatus; };

    /// initialize fitting process for given data points y. 
    /// returns AS_INVRANGE if the number of elements in yvalues != xvalues.size()
    APPROXSTATUS loadyvalues(const std::vector<double> &_yvalues);

    /** perform actual fitting for y loaded previously, with given smoothing parameter lambda.
        return spline values Y at knots X, and if necessary, rms error and EDF 
        (equivalent degrees of freedom) in corresponding output parameters (if they are not NULL).
        Y has length of numKnots+2, the last two values are spline derivatives at endpoints 
        (to pass to initialization of clamped spline):  the internal fitting process uses b-splines, 
        not natural cubic splines, therefore the endpoint derivatives are non-zero.
        EDF is equivalent number of free parameters in fit, increasing smoothing decreases EDF: 
        2<=EDF<=numKnots.  */
    APPROXSTATUS fitData(const double lambda, std::vector<double> *splineValues, double *rmserror=NULL, double* edf=NULL);

    /** perform fitting with adaptive choice of smoothing parameter lambda, to minimize AIC.
        AIC (Akaike information criterion) is defined as 
          log(rmserror*numDataPoints) + 2*EDF/(numDataPoints-EDF-1) .
        return spline values Y, rms error, equivalent degrees of freedom (EDF) and best-choice value of lambda. */
    APPROXSTATUS fitDataOptimal(std::vector<double> *splineValues, double *rmserror=NULL, double* edf=NULL, double *lambda=NULL);

    /** perform an 'oversmooth' fitting with adaptive choice of smoothing parameter lambda.
        The difference in AIC (Akaike information criterion) between the solution with no smoothing 
        and the returned solution is equal to deltaAIC (i.e. smooth more than optimal amount defined above).
        return spline values Y, rms error, equivalent degrees of freedom and best-choice value of lambda. */
    APPROXSTATUS fitDataOversmooth(const double deltaAIC, std::vector<double> *splineValues, double *rmserror=NULL, double* edf=NULL, double *lambda=NULL);

    /// convenience function to evaluate regression function at a given set of points, 
    /// after fitData*** was called and solution for weights has been obtained
    APPROXSTATUS computeRegressionAtPoints(const std::vector<double> &xpoints, std::vector<double> *ypoints) const;

private:
    APPROXSTATUS mystatus;        ///< internal status of the object
    const size_t numDataPoints;   ///< number of x[i],y[i] pairs (original data)
    const size_t numKnots;        ///< number of X[k] knots in the fitting spline
    const size_t numBasisFnc;     ///< number of b-spline basis functions = numKnots+2
    gsl_vector* knots;            ///< b-spline knots  X[k], k=0..numKnots-1
    gsl_vector* xvalues;          ///< x[i], i=0..numDataPoints-1
    gsl_vector* yvalues;          ///< y[i], overwritten each time loadYvalues is called
    gsl_vector* weightCoefs;      ///< w_p, weight coefficients for basis functions to be found in the process of fitting 
    gsl_vector* zRHS;             ///< z_p = C^T y, right hand side of linear system
    gsl_matrix* bsplineMatrix;    ///< matrix "C_ip" used in fitting process; i=0..numDataPoints-1, p=0..numBasisFnc-1
    gsl_matrix* LMatrix;          ///< lower triangular matrix L is Cholesky decomposition of matrix A = C^T C, of size numBasisFnc*numBasisFnc
    gsl_matrix* MMatrix;          ///< matrix "M" which is the transformed version of roughness matrix "R_pq" of integrals of product of second derivatives of basis functions; p,q=0..numBasisFnc-1
    gsl_vector* singValues;       ///< part of the decomposition of the roughness matrix
    gsl_vector* MTz;              ///< pre-computed M^T z
    gsl_vector* tempv;            ///< some routines require temporary storage
    gsl_bspline_workspace* bsplineWorkspace;  ///< workspace for b-spline evaluation
    gsl_vector* bsplineValues;      ///< to compute values of all b-spline basis functions at a given point x
    gsl_bspline_deriv_workspace* bsplineDerivWorkspace;  ///< workspace for derivative computation
    gsl_matrix* bsplineDerivValues; ///< to compute values and derivatives of basis functions
    double ynorm2;                  ///< |y|^2 - used to compute residual sum of squares (RSS)
    void initRoughnessMatrix();     ///< compute integrals over products of second derivatives of basis functions, and transform R to M+singValues
    void computeYvalues(std::vector<double> *splineValues) const;   ///< compute Y-values at spline knots X[k] after the weights w have been determined, and also two endpoint derivatives
    /** In the unfortunate case that the fit matrix appears to be singular, another algorithm 
        is used which is based on the GSL multifit routine, which performs SVD of bsplineMatrix. 
        It is much slower and cannot accomodate nonzero smoothing.
        This routine is automatically called from all fitData*** if mystatus==AS_SINGULAR. */
    APPROXSTATUS fitDataSingular(std::vector<double> *splineValues, double *rmserror, double* edf);
};

}  // namespace
