#include <vector>
#include <gsl/gsl_errno.h>
#include "../../../../smile/massmodel.h"
#include "../../../../smile/utils.h"

std::vector<double> radii;
std::vector<double> masses;
std::vector<double> positions;
std::vector<double> velocities;
double particlemass=0, blackholemass=0;

int set_mass_profile(double radius[], double mass[], int numpoints)
{
    radii.resize(numpoints);
    masses.resize(numpoints);
    for(int i=0; i<numpoints; i++) {
        radii[i]=radius[i];
        masses[i]=mass[i];
    }
    return 0;
}

int generate_particles(int count)
{
    gsl_set_error_handler(&(smile::gsl_verbose_error_handler));
    smile::CMassModel model(radii, masses);
    switch(model.getStatus())
    {
    case smile::CMassModel::MS_ARRAY_NOT_EQ:  return -1;  // arrays of unequal length -- can't happen
    case smile::CMassModel::MS_ARRAY_SHORT:   return -2;  // Too few points
    case smile::CMassModel::MS_RADII_NOT_ASC: return -3;  // Radii not in ascending order
    case smile::CMassModel::MS_MASS_NOT_ASC:  return -4;  // Mass not increasing with radius
    case smile::CMassModel::MS_DF_NEGATIVE:   break;      // f(E) not positive - not an error
    default: ;
    }
    smile::randomize();
    double totalmasswithoutbh=model.getTotalMass()-model.mass(0);
    blackholemass=model.mass(0);  // central black hole
    particlemass=totalmasswithoutbh/count;
    positions.assign (count*3+(blackholemass>0?3:0), 0);
    velocities.assign(count*3+(blackholemass>0?3:0), 0);
    double sumx=0, sumy=0, sumz=0, sumvx=0, sumvy=0, sumvz=0;
    for(int i=0; i<count; i++)
    {
        double radius=model.rad(smile::getRandomNumber()*totalmasswithoutbh + blackholemass);
        double veloc=model.samplev(radius);
        int ind=i*3+(blackholemass>0?3:0);
        smile::getRandomUnitVector(&(positions[ind]), &(positions[ind+1]), &(positions[ind+2]));
        positions[ind]   *= radius * (positions[ind]  >0 ^ sumx>=0 ? 1: -1);
        positions[ind+1] *= radius * (positions[ind+1]>0 ^ sumy>=0 ? 1: -1);
        positions[ind+2] *= radius * (positions[ind+2]>0 ^ sumz>=0 ? 1: -1);
        sumx+=positions[ind];
        sumy+=positions[ind+1];
        sumz+=positions[ind+2];
        smile::getRandomUnitVector(&(velocities[ind]), &(velocities[ind+1]), &(velocities[ind+2]));
        velocities[ind]   *= veloc * (velocities[ind  ]>0 ^ sumvx>=0 ? 1: -1);
        velocities[ind+1] *= veloc * (velocities[ind+1]>0 ^ sumvy>=0 ? 1: -1);
        velocities[ind+2] *= veloc * (velocities[ind+2]>0 ^ sumvz>=0 ? 1: -1);
        sumvx+=velocities[ind];
        sumvy+=velocities[ind+1];
        sumvz+=velocities[ind+2];
    }
    return 0;
}

int get_position(int i, double* x, double* y, double* z)
{
    if(i*3>=(int)positions.size()) return -1;
    *x=positions[i*3];
    *y=positions[i*3+1];
    *z=positions[i*3+2];
    return 0;
}

int get_velocity(int i, double* vx, double* vy, double* vz)
{
    if(i*3>=(int)velocities.size()) return -1;
    *vx=velocities[i*3];
    *vy=velocities[i*3+1];
    *vz=velocities[i*3+2];
    return 0;
}

int get_mass(int i, double* m)
{
    *m=(i==0&&blackholemass) ? blackholemass : particlemass;
    return 0;
}

int get_number_of_particles_updated(int* num)
{
    *num=positions.size()/3;
    return 0;
}
