from amuse.community import *
from amuse.community.interface.common import CommonCodeInterface, CommonCode

class MksphericalInterface(CodeInterface, CommonCodeInterface, LiteratureReferencesMixIn):
    """
    Mkspherical is a module for generating initial conditions for spherical stellar systems
    with isotropic velocity distribution function, using the Eddington inversion formula.
    The model is initialized from an array of values of enclosed mass within given radii;
    typically ~20 points covering the enclosed mass from 0.001 to 0.999 is enough.
    A simplified interface is given by 
    >>> amuse.ext.mkspherical_model.new_mkspherical_model(npoints, [radii], [masses], optional_unit_converter)
    Example of usage is given in the file test_mkspherical.py
    This is an AMUSE interface to the code contained in the SMILE software toolbox:

    .. [#] Vasiliev E., 2013, MNRAS, 434, 3174
    """

    include_headers = ['worker_code.h']

    def __init__(self, **keyword_arguments):
        CodeInterface.__init__(self, name_of_the_worker="mkspherical_worker", **keyword_arguments),
        LiteratureReferencesMixIn.__init__(self)

    def initialize_code():
        pass

    def cleanup_code():
        pass

    def commit_parameters():
        pass

    def recommit_parameters():
        pass

    new_particle = None

    def delete_particle(self, index_of_the_particle):
        pass

    @legacy_function
    def set_mass_profile():
        function = LegacyFunctionSpecification()
        function.addParameter('radius', dtype='float64', direction=function.IN, description = "distance from origin")
        function.addParameter('mass',   dtype='float64', direction=function.IN, description = "enclosed mass")
        function.addParameter('npoints',dtype='int32',   direction=function.LENGTH)
        function.must_handle_array = True
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_number_of_particles_updated():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def generate_particles():
        function = LegacyFunctionSpecification()
        function.addParameter('count', dtype='int32', direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_mass():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('mass', dtype='float64', direction=function.OUT, description = "Particle mass")
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def get_position():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float64', direction=function.OUT, description = "x-coordinate of particle")
        function.addParameter('y', dtype='float64', direction=function.OUT, description = "y-coordinate of particle")
        function.addParameter('z', dtype='float64', direction=function.OUT, description = "z-coordinate of particle")
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def get_velocity():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('vx', dtype='float64', direction=function.OUT, description = "x-velocity of particle")
        function.addParameter('vy', dtype='float64', direction=function.OUT, description = "y-velocity of particle")
        function.addParameter('vz', dtype='float64', direction=function.OUT, description = "z-velocity of particle")
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

class Mkspherical(CommonCode):

    def __init__(self, unit_converter = None, **options):
        self.unit_converter = unit_converter
        InCodeComponentImplementation.__init__(self, MksphericalInterface(**options), **options)

    def initialize_code(self):
        result = self.overridden().initialize_code()
        self.parameters.set_defaults()

    def define_converter(self, object):
        if not self.unit_converter is None:
            object.set_converter(self.unit_converter.as_converter_from_si_to_nbody())

    def define_methods(self, object):
        CommonCode.define_methods(self, object)
        object.add_method("set_mass_profile", (nbody_system.length,nbody_system.mass), (object.ERROR_CODE) )
        object.add_method("generate_particles", (object.NO_UNIT), (object.ERROR_CODE,) )
        object.add_method("get_number_of_particles_updated", (), (object.NO_UNIT, object.ERROR_CODE,) )
        object.add_method("get_mass", (object.INDEX,), (nbody_system.mass, object.ERROR_CODE) )
        object.add_method("get_position", (object.INDEX,), 
            (nbody_system.length, nbody_system.length, nbody_system.length, object.ERROR_CODE) )
        object.add_method("get_velocity", (object.INDEX,), 
            (nbody_system.speed, nbody_system.speed, nbody_system.speed, object.ERROR_CODE) )

    def define_particle_sets(self, object):
        object.define_set('particles', 'index_of_the_particle')
        object.set_new   ('particles', 'new_particle')
        object.set_delete('particles', 'delete_particle')
        object.add_getter('particles', 'get_mass')
        object.add_getter('particles', 'get_position')
        object.add_getter('particles', 'get_velocity')

    def generate_particles(self, count):
        result = self.overridden().generate_particles(count)
        if result == -1:
            raise AmuseException('Input arrays of unequal length')
        elif result == -2:
            raise AmuseException('Too few points in input arrays')
        elif result == -3:
            raise AmuseException('Input radii should be in ascending order')
        elif result == -4:
            raise AmuseException('Interpolated mass is not monotonic with radius')
        elif result == -5:
            print 'Warning, distribution function is not positive';
        number_of_updated_particles = self.get_number_of_particles_updated()
        if number_of_updated_particles:
            self.particles._private.attribute_storage._add_indices(range(number_of_updated_particles))
