/** \file    measureshape.cpp
    \brief   Tool: Measure the shape (axis ratio) of an N-body snapshot as a function of radius
    \author  EV, based on the code written by Rubens Machado
    \date    2008-2015

Two methods for shape determination are available:
- Equidensity ellipsoid (provided that density is a decreasing function of radius);
  this method is used in Athanassoula & Misiriotis(2002) and requires that density information 
  be present in the snapshot. 
- Inertia tensor of particles with iteratively determined axes (see Zemp et al.(2011) for an 
  extended discussion on variations of this method). The axes are determined from the inertia
  tensor of particles within ellipsoidal volume, with the ellipsoid itself using the axes 
  determined on the previous iteration, until it converges.

The program processes input snapshots in NEMO format (a file may contain snapshots at multiple 
times), which should be centered but not necessarily rotated to match principal planes.

The axis ratios are computed for several bins in radius, either using all particles inside 
the radius of this bin (i.e. including those in all previous bins), or in the current bin only.
Output may be presented either in the short or the long forms:
- for the short format, each line contains the time and 2*nbins values for b/a and c/a axes, 
  one line per snapshot at a given time;
- for the long one, each snapshot is displayed in nbins lines, with the radius, mass(r), 
  b/a and c/a, and optionally orientations of the axes of the inertia tensor.

*/

#include <math.h>
#include <stdlib.h>
#include <stdinc.h>
#include <getparam.h>
#include <io_nemo.h>
#include <snapshot/snapshot.h>
#include <vector>
#include <algorithm>
#include <iostream>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_errno.h>
#include "utils.h"

extern "C" {
string defv[]=
{
  const_cast<char*>("in=???\n         input snapshot"),
  const_cast<char*>("out=\n           optional output file for rectified snapshot"),
  const_cast<char*>("nbins=20\n       number of bins in which to calculate the axis ratios"),
  const_cast<char*>("dens=f\n         choice of method: equidensity ellipsoid (t) or inertia tensor (f)"),
  const_cast<char*>("cutoff=0.999\n   last bin encompasses this fraction of total mass"),
  const_cast<char*>("center=t\n       shift snapshot to the center (defined as median of the particle coordinates)"),
  const_cast<char*>("minbin=0\n       fraction of mass in the first bin (relative to Mtot*cutoff; by default = 1/nbins, may be smaller in which case bins will be spaced quasi-exponentially in enclosed mass)"), 
  const_cast<char*>("cumul=t\n        each bin contains all stars within radius (t) or only those not in previous bins (f); the latter option should be used with caution since it may not always converge"),
  const_cast<char*>("compact=f\n      print all data at given time in one line (t) or in multiple (one per radius)"),
  const_cast<char*>("angles=f\n       also print the orientations of the major and minor axes (angles between xy-component of major axis and X axis, and angle between minor axis and Z axis)"),
  const_cast<char*>("VERSION=2.5\n    Jun 2008 Rubens, 2010-2015 EV"),
  NULL
};
string usage=const_cast<char*>("Measure axis ratios as a function of radius. Two methods: equidensity ellipsoids axis ratio (if density is decreasing with radius) or inertia tensor. Snapshot must be centered on the density maximum or results will be rubbish.");
}
/// \cond INTERNAL_DOCS
struct particle{
  float x, y, z, dens, mass, rad, erad;
  particle() {x=y=z=dens=mass=rad=0;}
  particle(float _x, float _y, float _z, float _mass, float _dens)
  { x=(_x), y=(_y), z=(_z), mass=(_mass), dens=(_dens); rad=(sqrt(x*x+y*y+z*z)); erad=rad; }
  inline bool operator<(particle p) const
  { return (dens>=0) ? (dens>p.dens) : (erad<p.erad); }
};
/// \endcond

int main(int argc,char *argv[])
{
  float * pos=NULL, * mass=NULL, * dens=NULL;
  float * time=NULL;
  int   * nbody=NULL, *bits=NULL;

  initparam(argv,defv);
  string in   = getparam(const_cast<char*>("in"));
  string out  = getparam(const_cast<char*>("out"));
  int nbins   = getiparam(const_cast<char*>("nbins"));
  bool angles = getbparam(const_cast<char*>("angles"));
  bool usedens= getbparam(const_cast<char*>("dens"));
  bool compact= getbparam(const_cast<char*>("compact"));
  bool cumul  = getbparam(const_cast<char*>("cumul"));
  bool center = getbparam(const_cast<char*>("center"));
  double minbin=getdparam(const_cast<char*>("minbin"));
  double cutoff=getdparam(const_cast<char*>("cutoff"));
  if(cutoff>1) cutoff=1;
  std::vector<double> grid;
  if(minbin==0) minbin=1.0/nbins;
  smile::createNonuniformGrid(nbins, cutoff*minbin, cutoff, false, &grid);

  if(!compact) std::cerr << "time\tradius\tinmass\tb/a\tc/a" << (angles?"\tangle_x\tangle_z":"")<<std::endl;
  /* This reads the data from the input snapshot */
  while (int res=io_nemo(in,const_cast<char*>("float,n,t,x,m,d,b,read"),&nbody,&time,&pos,&mass,&dens,&bits))
  {
    if(res==-1)
    {
      bool abort=true;
      if(!(*bits & PosBit)) std::cerr<<"No position data! ";
      if(!(*bits &MassBit)) std::cerr<<"No mass data! ";
      if(!(*bits &DensBit)) {if(usedens) std::cerr<<"No density data! "; else abort=false;}
      if(abort){
        std::cerr<<"Aborting.\n";
        return 0;
      }
    }

    int Nbody = *nbody;
    if(Nbody==0) { 
      std::cerr << "No particles! Aborting.\n"; 
      return 0; 
    }
    double currenttime=(time!=NULL ? *time : 0);

    if(center) { // find median values of each coordinates and shift the snapshot there
      std::vector<double> coords(Nbody);
      for(int c=0; c<3; c++) {
        for(int b=0; b<Nbody; b++) coords[b]=pos[b*3+c];
        std::nth_element(coords.begin(), coords.begin() + Nbody/2, coords.end());
        double val=coords[Nbody/2];
        for(int b=0; b<Nbody; b++) pos[b*3+c]-=val;
      }
    }

    std::vector<particle> points(Nbody);
    double totalMass=0;

    for(int i=0; i<Nbody; i++)
    {
      points[i]=particle(pos[i*3], pos[i*3+1], pos[i*3+2], mass[i], (usedens ? dens[i] : -1));
      totalMass += mass[i];
    }

    if(usedens)   // center on density maximum
    {
      std::sort(points.begin(), points.end());
    }

    gsl_vector *eval = gsl_vector_alloc (3);
    gsl_matrix *evec = gsl_matrix_alloc (3, 3);
    gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (3);

    double b_over_a=1, c_over_a=1, old_b_over_a=1, old_c_over_a=1;
    double innerMass=0, avgrad, binMass;
    int ind=0;
    double Ixx,Iyy,Izz,Ixy,Ixz,Iyz;
    if(compact) std::cout << currenttime;
    for(int bin=0; bin<nbins; bin++)
    {
      double diff=0, prevInnerMass=innerMass;
      int niter=0, prevind=ind;
      double ex0=1, ex1=0, ex2=0; // components of the x eigenvector (initialized to match cartesian axes at start of the iterations)
      double ey0=0, ey1=1, ey2=0; // components of the y eigenvector
      double ez0=0, ez1=0, ez2=1; // components of the z eigenvector
      do{  // iterations
        if(!usedens) 
        {
          old_b_over_a=b_over_a;
          old_c_over_a=c_over_a;
          // init elliptic radius according to current estimate of b/a, c/a
          for(int i=0; i<Nbody; i++) 
          { // rotate elliptic radius according to current-iteration estimate of angles
            double xp=ex0*points[i].x + ey0*points[i].y + ez0*points[i].z;
            double yp=ex1*points[i].x + ey1*points[i].y + ez1*points[i].z;
            double zp=ex2*points[i].x + ey2*points[i].y + ez2*points[i].z;
            points[i].erad = static_cast<float>(sqrt(gsl_pow_2(xp) + gsl_pow_2(yp/b_over_a) + gsl_pow_2(zp/c_over_a)));
          }
          std::sort(points.begin(), points.end());
          innerMass=prevInnerMass;
          ind=prevind;
        }
        if(cumul) { ind=0; innerMass=0; }  // restart from the center
        avgrad=0; double avgradmass=0; binMass=0;
        Ixx=Iyy=Izz=Ixy=Ixz=Iyz=0.0;
        do{
          Ixx += points[ind].mass*(points[ind].y*points[ind].y + points[ind].z*points[ind].z);
          Iyy += points[ind].mass*(points[ind].z*points[ind].z + points[ind].x*points[ind].x);
          Izz += points[ind].mass*(points[ind].x*points[ind].x + points[ind].y*points[ind].y);
          Ixy -= points[ind].mass*(points[ind].x*points[ind].y);
          Ixz -= points[ind].mass*(points[ind].x*points[ind].z);
          Iyz -= points[ind].mass*(points[ind].y*points[ind].z);
          innerMass += points[ind].mass;
          binMass += points[ind].mass;
          if(bin==0 || innerMass>=grid[bin-1]*totalMass) {
            avgradmass+=points[ind].mass;
            avgrad += points[ind].mass*points[ind].rad;
          }
          ind++;
        } while(innerMass < totalMass * grid[bin]);
        avgrad/=avgradmass;
        double matrdata[] = 
          {Ixx , Ixy , Ixz,
           Ixy , Iyy , Iyz,
           Ixz , Iyz , Izz};
        gsl_matrix_view Inertia = gsl_matrix_view_array (matrdata, 3, 3);
        gsl_eigen_symmv (&Inertia.matrix, eval, evec, w);
        gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_ASC);
        double hx,hy,hz; // eigenvalues
        hx = gsl_vector_get (eval, 0);
        hy = gsl_vector_get (eval, 1);
        hz = gsl_vector_get (eval, 2);
        // get eigenvectors altogether
        ex0 = gsl_matrix_get(evec,0,0);
        ex1 = gsl_matrix_get(evec,0,1);
        ex2 = gsl_matrix_get(evec,0,2);
        ey0 = gsl_matrix_get(evec,1,0);
        ey1 = gsl_matrix_get(evec,1,1);
        ey2 = gsl_matrix_get(evec,1,2);
        ez0 = gsl_matrix_get(evec,2,0);
        ez1 = gsl_matrix_get(evec,2,1);
        ez2 = gsl_matrix_get(evec,2,2);
        b_over_a = sqrt((hx-hy+hz)/(-hx+hy+hz));
        c_over_a = sqrt((hx+hy-hz)/(-hx+hy+hz));
        if(!usedens)
        {
          diff = fabs(b_over_a-old_b_over_a)+fabs(c_over_a-old_c_over_a);
          niter++;
        //std::cerr << niter << ": " << innerMass << ", " << (avgrad/innerMass) << ", " << b_over_a <<" - " << c_over_a << " :: " << Ixx <<" "<<Iyy<<" "<<Izz<<" ; "<<Ixy<<" "<<Iyz<<" "<<Ixz<< "\n";
          if(gsl_isnan(diff)) {diff=0; b_over_a=c_over_a=1;}  // restart next bin from spherical if this one did not converge
        }
        if(out[0] && cumul && bin==nbins-1 && (diff<=1e-3 || niter>=20)) {  // output rectified snapshot
          for(int i=0; i<Nbody; i++) 
          { // rotate elliptic radius according to current-iteration estimate of angles
            pos[i*3  ]=ex0*points[i].x + ey0*points[i].y + ez0*points[i].z;
            pos[i*3+1]=ex1*points[i].x + ey1*points[i].y + ez1*points[i].z;
            pos[i*3+2]=ex2*points[i].x + ey2*points[i].y + ez2*points[i].z;
          }
          if(io_nemo(out,const_cast<char*>("float,n,t,x,m,save"),&nbody,&time,&pos,&mass,&dens,&bits)!=1) {
            std::cerr << "Error saving file "<<out;
            out[0]=0;
          }
        }
      } while(diff > 1e-3 && niter < 20);  // in case of usedens=1 only one iteration, since diff=0
      // output
      if(compact)
      {
        std::cout << "\t" << smile::pp(b_over_a,7) << "\t" << smile::pp(c_over_a,7);
      }
      else
      {
        std::cout << currenttime << "\t" << smile::pp(avgrad,7) 
            << "\t" << smile::pp(innerMass,7) 
            << "\t" << smile::pp(b_over_a,7) 
            << "\t" << smile::pp(c_over_a,7);
        if(angles)
        {
          double anglex,anglez;
          // first ensure that ex,ey,ez are right-handed basis with ez pointing up the xy plane and ex to the right of yz plane
          if(ez2<0) { ex2*=-1; ey2*=-1; ez2*=-1; }
          if(ex0<0) { ex0*=-1; ey0*=-1; ez0*=-1; }
          if(ex0*(ey1*ez2-ey2*ez1)+ex1*(ey2*ez0-ey0*ez2)+ex2*(ey1*ez2-ey2*ez1) < 0) { ex1*=-1; ey1*=-1; ez1*=-1; }
          // angle between the z eigenvector and the z axis
          anglez = -acos(ez2);
          // angle between the (xyplane component of the) x eigenvector and the x axis 
          anglex = -asin(ex1/(sqrt(ex0*ex0+ex1*ex1)));
          // print the angles
          std::cout <<"\t"<< smile::pp((180.0/M_PI) * anglex, 7) <<"\t"<< smile::pp((180.0/M_PI) *anglez, 7);
        }
        std::cout << std::endl << std::flush;
      }
    }
    std::cout << std::endl << std::flush;
    gsl_vector_free (eval);
    gsl_matrix_free (evec);
    gsl_eigen_symmv_free (w);

  }
  finiparam();
  return 0;
}
