/** \file   renderdensity.cpp
    \brief  Tool: Create a visual representation of an arbitrary density model
    \author EV
    \date   2011-2014

Construct an Nbody snapshot with particles distributed according to the density profile,
constructed either from parameters passed as command-line arguments, 
or as an INI file with all required parameters, 
or as a text file with potential coefficients (*.coef_bse, *.coef_bsec, *.coef_spl, *.coef_cyl
or *.mass for BSE, BSECompact, Spline, CylSpline and Spherical, correspondingly).
Particles are given zero velocity and thus the snapshot is not an equilibrium model.
*/

#include "configfile.h"
#include "icgenerator.h"
#include "iosnapshot.h"
#include "managepotential.h"
#include "potential.h"
#include "utils.h"

/// default values of command-line parameters
char const* description =
    "renderdensity:  create an nbody snapshot with rendering of a given density profile with a set of particles at rest.\n"
    "The density is specified either as command-line arguments, or as an input file.\n"
    "  In the first case the following arguments are possible:\n"
    "density=???     name of the density model\n"
    "mass=1          total mass\n"
    "q=1             axis ratio y/x\n"
    "p=1             axis ratio z/x\n"
    "scalerad=1      scale radius of the model (another spelling is rscale)\n"
    "scalerad2=1     second scale radius (if applicable)\n"
    "gamma=1         power-law index of the density cusp\n"
    "sersicindex=4   shape parameter of the Sersic profile\n"
    "  In the second case, only one argument is necessary:\n"
    "file=???        input text-file with potential parameters (*.ini), coefficients (*.coef_*), or spherical mass profile (*.mass)\n"
    "  Output the rendered density profile into the file specified by the following parameters:\n"
    "out=???         output snapshot\n"
    "nbody=???       number of points in snapshot\n"
    "outformat=Text  output format (Text/Nemo/Gadget)\n"
    "VERSION=2.5     Jan 2015 EV";

int main(int argc, char* argv[])
{
    gsl_set_error_handler(&smile::gsl_verbose_error_handler);
    smile::CConfigPotential cfg=smile::defaultConfigPotential;
    // parse command-line parameters as if they were provided in the [Potential] section of a hypothetical ini file
    smile::CConfigFile cfgFile("Potential", argc, argv);
    std::string outfilename = cfgFile.getValueString("Potential/out");
    std::string outformat   = cfgFile.getValueString("Potential/outformat", "Text");
    int nbody = cfgFile.getValueInt("Potential/nbody");
    if(outfilename.empty() || nbody<=0) 
    {
        smile::my_message("", description);
        return 0;
    }
    smile::readConfigPotential(cfgFile, &cfg);  // parse the rest of parameters
    if(!cfg.NbodyFile.empty()) {  // file name was provided
        if(smile::ends_with_str(cfg.NbodyFile, ".ini")) {  // name of ini file passed as file=... parameter
            smile::CConfigFile cfgFile1(cfg.NbodyFile);
            smile::readConfigPotential(cfgFile1, &cfg);
        } else {
            if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN)   // try to deduce potential type from file extension
                cfg.PotentialType=smile::coefFileType(cfg.NbodyFile);
            if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN) { // none of the known extensions
                smile::my_message("", "Unknown potential type for the given file extension");
                return 0;
            } else cfg.DensityType=smile::CDensity::PT_COEFS;
        }
    }
    // attempt to create a density model
    const smile::CDensity* model = NULL;
    if(cfg.PotentialType!=smile::CDensity::PT_UNKNOWN)
        model=smile::createPotential(&cfg);
    if(model==NULL) {  // config does not specify a correct potential
        if(cfg.DensityType==smile::CDensity::PT_UNKNOWN && cfg.PotentialType!=smile::CDensity::PT_UNKNOWN)
            cfg.DensityType=cfg.PotentialType;  // name provided as potential=... parameter instead of density=...
        model=smile::createDensity(&cfg);
    }
    if(model==NULL) {
        smile::my_message("", "No valid density model was specified");
        return 0;
    }
    smile::randomize();
    smile::CPointMassSet<float> points;
    smile::CICGeneratorDensitySampler icg(model);
    icg.createInitialConditions(nbody, &points);
    std::string cmdlinefull="renderdensity";
    char** argv0=argv+1;
    for(char *arg=*argv0; (arg = *argv0); argv0 ++ ){
        cmdlinefull += " " + std::string(arg);
    }
    smile::CBasicIOSnapshot* snap_out = smile::createIOSnapshotWrite(outformat, outfilename, cmdlinefull);
    if(snap_out!=NULL) {
        if(!snap_out->writeSnapshot(&points)) 
            smile::my_error("", "Can't write output snapshot: "+outfilename);
    } else {
        smile::my_error("", "Unknown output format: "+outformat);
    }
    delete snap_out;
    delete model;
    return 0;
}
