/** \file    testaccuracy.cpp
    \brief   Tool: check accuracy of approximation of an N-body snapshot by a potential expansion
    \author  EV
    \date    2014-2015

    This program checks the accuracy of potential approximation constructed from an N-body snapshot
    or from an analytic density profile, using one of the potential expansions available in SMILE
    (BSE, Spline, CylSpline, etc.)
    It doesn't compare potential or forces, but instead focuses on density (as this quantity
    is more sensitive to systematic biases if the potential parameters are chosen inappropriately).
    It creates a spatial grid, similar to the one used in Schwarzschild models,
    and computes the mass in each grid cell twice: from the smooth potential expansion, and
    summing up point masses that fall in the given cell (in the case of initialization from 
    an N-body snapshot), or from the original smooth density profile. 
    In the former case, the numbers are not expected to agree exactly, but the relative variation
    should be of order 1/sqrt{N_points_in_cell}, and that's what this program actually checks and
    reports as the chi^2/d.o.f. statistical measure, which should be of order unity for a good
    approximation.
    In the latter case, the Mean Integrated Square Error (MISE) is reported as the density-weighted
    measure of approximation error; the smaller is this number, the better.
*/
#include <fstream>
#include "configfile.h"
#include "iosnapshot.h"
#include "managepotential.h"
#include "massmodelaxi.h"
#include "schwarzschild.h"
#include "utils.h"

int main(int nparam, char** param)
{
    gsl_set_error_handler(&smile::gsl_verbose_error_handler);
    if(nparam<3) {
        smile::my_error("", "Need to provide at least potential type= and either file= or density=");
        return 0;
    }
    smile::CConfigPotential cfg=smile::defaultConfigPotential;
    smile::CConfigFile cfgFile("Potential", nparam, param);
    smile::readConfigPotential(cfgFile, &cfg);
    if( cfg.PotentialType!=smile::CDensity::PT_BSE &&
        cfg.PotentialType!=smile::CDensity::PT_BSECOMPACT &&
        cfg.PotentialType!=smile::CDensity::PT_SPLINE &&
        cfg.PotentialType!=smile::CDensity::PT_CYLSPLINE) {
        smile::my_error("", "Specified potential type=... is not a potential expansion");
        return 0;
    }
    smile::CPointMassSet<double> *points=NULL;  // either use an N-body snapshot...
    const smile::CDensity* density=NULL;        // or an analytic density profile,
    const smile::CPotential* potential=NULL;    // and compare them to the potential approximation
    if(cfg.NbodyFile.empty()) {
        density=smile::createDensity(&cfg);
        if(!density) {
            smile::my_error("", "Specified density type is incorrect");
            return 0;
        }
        cfg.NbodyFile=cfgFile.getValueString("Potential/out");  // if specified, will use the file name for writing grid and coefs
    } else {
        cfg.DensityType=smile::CPotential::PT_NB;
        smile::CBasicIOSnapshot* snapRead=smile::createIOSnapshotRead(cfg.NbodyFile);
        if(snapRead!=NULL)
            snapRead->readSnapshot(points);
        delete snapRead;
        if(points==NULL || points->size()==0) { 
            smile::my_error("", "Can't read from file "+cfg.NbodyFile); 
            return 0; 
        }
    }
    smile::my_message("", "Creating potential");
    if(points) {
        potential=smile::createPotentialFromPoints(&cfg, points);  // create a potential approximation of a point mass set
    } else if (density) {
        potential=smile::createPotential(&cfg);  // create a potential approximation for the given smooth density profile
    } else return 0;
    if(!potential) {
        smile::my_error("","Can't create potential");
        return 0;
    }
    if(!cfg.NbodyFile.empty())
        smile::writePotential(cfg.NbodyFile+smile::coefFileExtension(potential->PotentialType()), potential);
    smile::my_message("", "Creating grid");
    int numRadialCoefs  =cfgFile.getValueInt("Potential/numRadialCoefs",   20);
    int numVerticalCoefs=cfgFile.getValueInt("Potential/numVerticalCoefs", 15);
    int numAngularCoefs =cfgFile.getValueInt("Potential/numAngularCoefs",  cfg.Ncoefs_angular/2+1);
    bool cylgrid = cfgFile.getValueBool("Potential/cylGrid", false);
    smile::CSchwDataDensGrid* grid = NULL;
    if(cylgrid)
        grid=new smile::CSchwDataDensGridCylindrical(potential, numRadialCoefs, numVerticalCoefs, numAngularCoefs);
    else
        grid=new smile::CSchwDataDensGridClassic(potential, numRadialCoefs, 0., 0., numAngularCoefs, 1., 1.);
    std::vector<double> cellmassgrid(grid->numConstraints()-1);
    std::vector<double> cellmasses(grid->numConstraints()-1);
    std::vector<double> cellmasses2(grid->numConstraints()-1);
    std::vector<size_t> cellcounts(grid->numConstraints()-1);
    for(size_t c=0; c<grid->numConstraints()-1; c++) {
        cellmassgrid[c]=grid->getConstraint(c)*grid->getConstraintNormFactor(c);
    }
    if(points) {
        smile::my_message("", "Counting particles");
        for(size_t i=0; i<points->size(); i++) {
            size_t c=grid->whichCell(points->at(i).first.Pos[0], points->at(i).first.Pos[1], points->at(i).first.Pos[2]);
            cellcounts[c]++;
            cellmasses[c]+=points->at(i).second;
            cellmasses2[c]+=smile::pow_2(points->at(i).second);
        }
        double chi2=0;
        size_t numBadCells=0;
        for(size_t c=0; c<grid->numConstraints()-1; c++)
            if(cellcounts[c]>1)
                chi2+=smile::pow_2( (cellmassgrid[c]-cellmasses[c])/(cellmassgrid[c]+cellmasses[c])*2)*cellcounts[c];
            else
                numBadCells++;
        smile::my_message("", "chi2/dof= "+smile::convertToString(chi2/(grid->numConstraints()-numBadCells-1)) +
            (numBadCells>0?(", "+smile::convertToString(numBadCells)+" cells have <=1 points"):"") );
    } else {
        smile::my_message("", "Computing mass in original density model");
        double MISE=0, totalmassgrid=0;
        for(size_t c=0; c<grid->numConstraints()-1; c++) {
            cellmasses[c]=grid->initCellMass(c, density);
            MISE += cellmassgrid[c] * smile::pow_2( (cellmassgrid[c]-cellmasses[c])/(cellmassgrid[c]+cellmasses[c])*2);
            totalmassgrid += cellmassgrid[c];
        }
        smile::my_message("", "MISE="+smile::convertToString(MISE/totalmassgrid));
    }
    delete potential;
    delete points;
    delete density;
    if(!cfg.NbodyFile.empty()) {
        std::ofstream strm((cfg.NbodyFile+".grid").c_str());
        for(size_t c=0; c<grid->numConstraints()-1; c++) {
            double X,Y,Z;
            grid->cellCenter(c, &X,&Y,&Z);
            strm << X<<"\t"<<Y<<"\t"<<Z<<"\t"<<
                cellmassgrid[c] <<"\t"<< cellmasses[c];
            if(points) strm <<"\t"<<cellcounts[c];
            strm<<"\n";
        }
    }
}